/*
 * Copyright (c) 1995-1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Definitions used by a parent to start nested sub-processes.
 */
#ifndef _FLUKE_CP_STARTUP_H_
#define _FLUKE_CP_STARTUP_H_

#include <fluke/thread.h>
#include <fluke/mapping.h>
#include <fluke/ref.h>
#include <fluke/sac/types.h>


struct cp_startup
{
	/*
	 * Thread object for first thread started in the program.
	 * The first thread can simply call fluke_thread_self()
	 * to find the address of this entire structure.
	 */
	struct fluke_thread	thread;

	/*
	 * Reference to the child's own task and scheduler,
	 * for starting other threads.
	 */
	struct fluke_ref	task_ref;
	struct fluke_ref	scheduler_ref;

	/*
	 * System call port reference used to call the parent process.
	 * The referenced server must provide the fluke::Parent interface.
	 */
	struct fluke_ref	parent_ref;

	/*
	 * List of initial segments and mappings representing the
	 * program's text, data, bss, stack, etc., that were created
	 * by the parent before starting the child's first thread.
	 * See below.
	 */
	struct cp_startup_seg	*segments;
	unsigned		segment_count;

	/*
	 * Location of the arguments and environment strings.
	 */
	int			argc;
	char			**argv;
	char			**environ;
};

/*
 * There is one of these structures for each contiguous mapping in the
 * child task that was made by the parent before startup.  The parent is
 * responsible for keeping the segment MOM objects alive during
 * bootstrapping, as if the first child thread starting were the completion
 * of the reply message IPC for an RPC the child made to its parent.  A
 * flags word in each structure goes with that structure's segment (port)
 * ref just like the word passed in IPC for each MOM ref in a client RPC's
 * `out' parameters.  If the collection of flags for all the segment refs
 * indicate a "timely ack or disconnect" would be required of a client RPC
 * stub, then the child is responsible for sending a Parent::ack_startup()
 * RPC on its parent port in a timely manner after it has registered its
 * own refs to keep the segment objects live.
 */
struct cp_startup_seg
{
	/*
	 * Mapping of segment into task.
	 * The program can examine and manipulate this mappings if desired,
	 * for example to handle mprotect()s.
	 * Comes first since it has most restrictive alignment.
	 */
	struct fluke_mapping	mapping;

	/*
	 * Reference to the port representing the segment object;
	 * the region_ref in this mapping came from a RandomAccessIO::map()
	 * RPC to this port.  The flags word has the same meaning as
	 * the word passed in IPC for each MOM ref.
	 */
	struct fluke_ref	segment_ref;
	mom_ref_ipc_flags_t	segment_ref_ipc_flags;

	/*
	 * The `offset' argument used in that RandomAccessIO::map() call.
	 */
	fluke_off_t		segment_offset;

	/*
	 * Bitmask of flags associated with this segment.
	 */
	unsigned int		flags;
#define SEG_VAR		1	/* variable size segment */
#define SEG_GROWSDOWN	2	/* grows toward lower addresses */
#define SEG_AUTOGROW	4	/* should grow automatically */
#define SEG_CLEARED	8	/* memory should be zeroed when grown */
};

#endif _FLUKE_CP_STARTUP_H_
