/*
 * Copyright (c) 1995, 1996 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#ifndef _FLUKE_IPC_H_
#define _FLUKE_IPC_H_

#include <oskit/machine/types.h>
#include <fluke/ob.h>

#if FLASK
#include <flask/security.h>
#endif

struct fluke_ipc_params;
struct fluke_ref;
struct fluke_port;
struct fluke_pset;
struct fluke_ipc_params;
struct fluke_data_desc;


/*
 * Minimum size of messages sent or received across Fluke IPC connections.
 * The minimum size is always a multiple of the machine register size,
 * which is the size of a fluke_ipc_word_t.
 * The machine-dependent header defines FLUKE_MIN_MSG_WORDS
 * and fluke_ipc_word_t.
 */
#define FLUKE_MIN_MSG_SIZE	(FLUKE_MIN_MSG_WORDS * sizeof(fluke_ipc_word_t))

/*
 * Definitions for message buffer alignment and size granularity requirements.
 * The machine-dependent header defines FLUKE_MSG_ALIGN_SHIFT
 * and FLUKE_MSG_GRAN_SHIFT, which define the alignment and granularity
 * in terms of bits shifted (i.e. powers of two).
 * On any architecture, the alignment and granularity
 * will be no more than the size of type fluke_ipc_word_t.
 */
#define FLUKE_MSG_ALIGN		(1 << FLUKE_MSG_ALIGN_SHIFT)
#define FLUKE_MSG_GRAN		(1 << FLUKE_MSG_GRAN_SHIFT)
#define FLUKE_MSG_ALIGN_MASK	(FLUKE_MSG_ALIGN - 1)
#define FLUKE_MSG_GRAN_MASK	(FLUKE_MSG_GRAN - 1)

/*
 * Each port has an associated server-defined alias,
 * which the server uses to identify the source of incoming messages.
 */
typedef oskit_addr_t fluke_alias_t;

/*
 * Entry describing one scatter/gather buffer,
 * in a scatter/gather buffer array.
 */
struct fluke_ipc_buf
{
	oskit_size_t		words;
	void			*buf;
};
typedef struct fluke_ipc_buf fluke_ipc_buf_t;

/*
 * Structure containing message send and receive information for IPC.
 * The structure is actually defined in <fluke/machine/ipc.h>,
 * since its exact organization is architecture-specific.
 */
typedef struct fluke_ipc_params fluke_ipc_params_t;


/*** One-way IPC ***/

fluke_error_t fluke_ipc_send(
	struct fluke_ref	*destination_port_ref,
	struct fluke_ipc_params	*ipc_params);

#if FLASK
fluke_error_t fluke_ipc_send_secure(
	struct fluke_ref	*destination_port_ref,
	struct fluke_ipc_params	*ipc_params,
        security_id_t		client_sid);  /* effective client SID */
#endif


/*** Idempotent IPC ***/

fluke_error_t fluke_ipc_call(
	struct fluke_ref	*destination_port_ref,
	struct fluke_ipc_params	*ipc_params);

#if FLASK
fluke_error_t fluke_ipc_call_secure(
	struct fluke_ref	*destination_port_ref,
	struct fluke_ipc_params	*ipc_params,
        security_id_t		client_sid);  /* effective client SID */
#endif

fluke_error_t fluke_ipc_reply(
	struct fluke_ipc_params	*ipc_params);

/*
 * Reply to the currently outstanding idempotent request
 * and wait for another request to come in.
 * In the ipc_params, the send_buf_tab is ignored,
 * because idempotent communication doesn't support scatter/gather.
 * The recv_* fields determine where to receive the next incoming message.
 */
fluke_error_t fluke_ipc_reply_wait_receive(
	struct fluke_ipc_params	*ipc_params,
	fluke_alias_t		*out_port_alias);

#if FLASK
fluke_error_t fluke_ipc_reply_wait_receive_secure(
	struct fluke_ipc_params	*ipc_params,
	fluke_alias_t		*out_port_alias,
        security_id_t		*out_client_sid);  /* effective client SID */
#endif

/*
 * Reply to the thread that invoked us and donate our CPU time to it.
 * At the same time, go to sleep waiting for another message to arrive.
 */
void fluke_ipc_schedule_wait_receive(
	struct fluke_ipc_params	*ipc_params,
	struct fluke_port	**out_port);

#if FLASK
void fluke_ipc_schedule_wait_receive_secure(
	struct fluke_ipc_params	*ipc_params,
	struct fluke_port	**out_port,
        security_id_t		*out_client_sid);  /* effective client SID */
#endif


/*** Reliable IPC ***/

/* Client-side reliable IPC routines */

fluke_error_t fluke_ipc_client_connect_send(
	struct fluke_ref	*destination_port_ref,
	struct fluke_ipc_params	*ipc_params);

#if FLASK
fluke_error_t fluke_ipc_client_connect_send_secure(
	struct fluke_ref	*destination_port_ref,
	struct fluke_ipc_params	*ipc_params,
        security_id_t		client_sid);
#endif

fluke_error_t fluke_ipc_client_ack_send(
	struct fluke_ipc_params	*ipc_params);

fluke_error_t fluke_ipc_client_send(
	struct fluke_ipc_params	*ipc_params);

fluke_error_t fluke_ipc_client_connect_send_over_receive(
	struct fluke_ref	*destination_port_ref,
	struct fluke_ipc_params	*ipc_params);

#if FLASK
fluke_error_t fluke_ipc_client_connect_send_over_receive_secure(
	struct fluke_ref	*destination_port_ref,
	struct fluke_ipc_params	*ipc_params,
        security_id_t		client_sid);     /* effective client SID */
#endif

fluke_error_t fluke_ipc_client_ack_send_over_receive(
	struct fluke_ipc_params	*ipc_params);

fluke_error_t fluke_ipc_client_send_over_receive(
	struct fluke_ipc_params	*ipc_params);

fluke_error_t fluke_ipc_client_over_receive(
	struct fluke_ipc_params	*ipc_params);

fluke_error_t fluke_ipc_client_receive(
	struct fluke_ipc_params	*ipc_params);

void fluke_ipc_client_disconnect(void);

fluke_error_t fluke_ipc_client_alert(void);


/* Server-side reliable IPC routines */

fluke_error_t fluke_ipc_server_ack_send(
	struct fluke_ipc_params	*ipc_params);

fluke_error_t fluke_ipc_server_send(
	struct fluke_ipc_params	*ipc_params);

fluke_error_t fluke_ipc_server_ack_send_wait_receive(
	struct fluke_ipc_params	*ipc_params,
	fluke_alias_t		*out_port_alias);

#if FLASK
fluke_error_t fluke_ipc_server_ack_send_wait_receive_secure(
	struct fluke_ipc_params	*ipc_params,
	fluke_alias_t		*out_port_alias,
        security_id_t		*out_client_sid);  /* effective client SID */
#endif

fluke_error_t fluke_ipc_server_ack_send_over_receive(
	struct fluke_ipc_params	*ipc_params);

fluke_error_t fluke_ipc_server_send_wait_receive(
	struct fluke_ipc_params	*ipc_params,
	fluke_alias_t		*out_port_alias);

#if FLASK
fluke_error_t fluke_ipc_server_send_wait_receive_secure(
	struct fluke_ipc_params	*ipc_params,
	fluke_alias_t		*out_port_alias,
        security_id_t		*out_client_sid);  /* effective client SID */
#endif

fluke_error_t fluke_ipc_server_send_over_receive(
	struct fluke_ipc_params	*ipc_params);

fluke_error_t fluke_ipc_server_over_receive(
	struct fluke_ipc_params	*ipc_params);

fluke_error_t fluke_ipc_server_receive(
	struct fluke_ipc_params	*ipc_params);

void fluke_ipc_server_disconnect(void);

fluke_error_t fluke_ipc_server_alert(void);


/*** General server-side routines ***/

fluke_error_t fluke_ipc_setup_wait_receive(
	struct fluke_pset	*server_pset,
	struct fluke_ipc_params	*ipc_params,
	fluke_alias_t		*out_port_alias);

#if FLASK
fluke_error_t fluke_ipc_setup_wait_receive_secure(
	struct fluke_pset	*server_pset,
	struct fluke_ipc_params	*ipc_params,
	fluke_alias_t		*out_port_alias,
        security_id_t		*out_client_sid);  /* effective client SID */
#endif

fluke_error_t fluke_ipc_wait_receive(
	struct fluke_ipc_params	*ipc_params,
	fluke_alias_t		*out_port_alias);

#if FLASK
fluke_error_t fluke_ipc_wait_receive_secure(
	struct fluke_ipc_params	*ipc_params,
	fluke_alias_t		*out_port_alias,
        security_id_t		*out_client_sid);  /* effective client SID */
#endif


/*** Routines that deal with both sides at once.  ***/

fluke_error_t fluke_ipc_client_server_passthrough(
	oskit_size_t		data_words,
	oskit_size_t		refs);

fluke_error_t fluke_ipc_server_client_passthrough(
	oskit_size_t		data_words,
	oskit_size_t		refs);


#include <fluke/machine/ipc.h>

#endif /* _FLUKE_IPC_H_ */
