/*
 * Copyright (c) 1997 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#ifndef _NEST_SERVER_MEMPOOL_H_
#define _NEST_SERVER_MEMPOOL_H_

#if FLASK
#include <flask/security.h>
#endif

struct mm_segment;

/*
 * Page replacement policy.
 */
struct mm_policy {
	void (*init)(void);
	int (*pagein)(struct mm_segment *, oskit_addr_t, size_t);
	int (*pageout)(struct mm_segment *, oskit_addr_t, size_t);
	void *(*alloc)(struct mm_segment *, oskit_addr_t, oskit_size_t, void *);
	void (*free)(struct mm_segment *, oskit_addr_t, oskit_size_t, void *);
	void (*exit)(void);
};
extern struct mm_policy mm_null_policy;
extern struct mm_policy mm_fifo_policy;
extern struct mm_policy mm_minfund_policy;

struct mm_info {
	char *mm_ident;			/* string identifying manager */
	void *mm_membase;		/* VA of physical memory region */
	int mm_memsize;			/* size of physical memory region */
	int mm_vmsize;			/* size of VM (physmem + swap) */
	int mm_vminuse;			/* size of VM in use */
	int mm_flags;			/* tailor the behavior of the library */
	struct mm_policy *mm_policy;	/* page replacement policy */
};
extern struct mm_info mminfo;

/* flags */
#define MM_EAGERALLOC		0x01	/* allocate memory at set_size time */
#define MM_PREALLOCMAXVM	0x02	/* preallocate VM resources */
#define MM_DELAYFLOBDESTROY	0x04	/* don't destroy segment flobs til
					   containing pool is destroyed */
#define MM_NOCLEANUPONEXIT	0x08	/* XXX hack for Flask kserver */

/*
 * Initialize the MM library.
 *
 * The mminfo structure must be initialized before calling.
 * In particular, membase and memsize must be set to define the ``physical''
 * memory area.
 */
void mm_setup(void);


/*
 * Memory allocation callbacks.
 *	lalloc/lfree are for memory used by the memory manager,
 *	malloc/mfree are for managed memory.
 */

/*
 * Return a pointer to new memory of the indicated size aligned to the
 * given size.  Returns zero if no memory is available.
 */
void *mm_lalloc(unsigned align, size_t size);
#if FLASK
void *mm_lalloc_secure(unsigned align, size_t size, security_id_t sid);
#endif

/*
 * Free memory allocated with mm_lalloc.
 */
void mm_lfree(void *mem, size_t size);
#if FLASK
void mm_lfree_secure(void *mem, size_t size, security_id_t sid);
#endif

/*
 * Returns page-aligned memory of the indicated size.
 * If possible, return memory starting at the desired address (daddr).
 * Returns zero if no memory is available.
 *
 * For Flask configuration, we request memory with a specific SID
 * (memory_sid).  If no memory is available for that SID, mm_malloc will 
 * return zero and may also set *out_replacement_sid to the desired SID 
 * to target for page replacement.  If there is no preferred SID for
 * page replacement, then *out_replacement_sid should be set to zero.
 */
#if FLASK
void *mm_malloc(size_t size, void *daddr, 
		security_id_t memory_sid,
		security_id_t *out_replacement_sid);
#else
void *mm_malloc(size_t size, void *daddr);
#endif

/*
 * Free memory allocated with mm_malloc.
 */
#if FLASK
void mm_mfree(void *mem, size_t size, security_id_t sid);
#else
void mm_mfree(void *mem, size_t size);
#endif


/*
 * Backing store (swap) allocation callbacks.
 *	alloc/free are for allocating and freeing swap blocks,
 *	read/write are for IO.
 */

/*
 * Allocate a piece of swap space of the indicated size.
 * Returns a swap block index (used to identify the chunk in future calls)
 * or -1 on an error.
 */
int mm_swap_alloc(size_t size);

/*
 * Free a previousdly allocated swap block of the indicated size.
 */ 
void mm_swap_free(int bn, size_t size);

/*
 * Read a chunk of memory of the indicated size from the specified swap block.
 * Returns zero on success, an error code otherwise.
 */
int mm_swap_read(void *mem, size_t size, int bn);

/*
 * Write a chunk of memory of the indicated size to the specified swap block.
 * Returns zero on success, an error code otherwise.
 */
int mm_swap_write(void *mem, size_t size, int bn);

#if FLASK
/*
 * XXX interface for creating Fluke task implementing a segment.
 * This interface is only overridden in the kernel kserver implementation.
 */
struct fluke_task;
struct fluke_ref;
int fluke_segment_task_create(struct fluke_task *task, security_id_t sid);
#endif

#endif _NEST_SERVER_MEMPOOL_H_
