/*
 * Copyright (c) 1997-1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * Contributed by the Computer Security Research division,
 * INFOSEC Research and Technology Office, NSA.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#if FLASK

#ifndef _FLUKE_NEST_SERVER_AVC_H_
#define _FLUKE_NEST_SERVER_AVC_H_

#include <oskit/types.h>
#include <oskit/boolean.h>
#include <flask/security.h>
#include <flask/av_permissions.h>
#include <mom/object.h>

/*
 * Access vector cache implementation for policy-enforcing servers.
 *
 * These routines are used internally by policy-enforcing servers.
 *
 * The naming convention (nest_server_avc) distinguishes
 * these routines from the server work functions for the
 * Avc interface, which have the nest_avc prefix. 
 * 
 * The nest_server_avc_has_permission routine is used by 
 * the Flick-generated dispatcher to perform permission checks,
 * and may be used by other server code.
 *
 * The nest_server_avc_add and nest_server_avc_remove routines
 * may be used by the server nest_avc_add and nest_avc_remove
 * work functions.  Additionally, they may be used locally
 * by the server for management of the cache (eg to load
 * initial state into the cache). 
 */


/*
 * Create and initialize the access vector cache.
 *
 * Returns FALSE if insufficient memory.
 */
oskit_bool_t nest_server_avc_init(void);


#include <oskit/avc/avc.h>
oskit_error_t nest_server_avc_com_init(oskit_avc_t **out_avc);


/*
 * Register the access vector cache with the
 * security policy server, using the specified
 * module to service requests from the security	
 * policy server.
 *  
 * Returns FALSE if insufficient memory.
 */

typedef struct security_class_pair {
  security_class_t sclass;
  security_class_t tclass;
} security_class_pair_t;

oskit_bool_t nest_server_avc_register(mom_module_t m,
				      int npairs,
				      security_class_pair_t *pairs);


/*
 * Determine whether the requested permissions are
 * granted between (sclass,ssid) and (tclass,tsid);
 * if a valid access ruling is not present in the access
 * vector cache, then obtain one from the security
 * policy server for the access query.
 *
 *
 * Returns OSKIT_ENOMEM if there is insufficient memory,
 *	   OSKIT_EAGAIN if client must retry the operation,  
 *	   OSKIT_EACCES if permission is denied, or
 *	   0 if permission is granted.
 */
oskit_error_t nest_server_avc_has_permission(security_class_t sclass,
					     security_id_t ssid,
					     security_class_t tclass,
					     security_id_t tsid,
					     access_vector_t requested);


/*
 * Add or replace the entry in the access vector cache
 * for (sclass, ssid, tclass, tsid).  
 *
 * Returns OSKIT_ENOMEM if there is insufficient memory,
 *         OSKIT_EEXIST if 'replace' is FALSE and an entry already exists, or
 *	   0 on success.
 */
oskit_error_t nest_server_avc_add(security_class_t sclass,
				  security_id_t ssid,
				  security_class_t tclass,
				  security_id_t tsid,
				  access_vector_t allowed,
				  access_vector_t decided,
				  oskit_u32_t expiration,
				  oskit_bool_t wired,
				  oskit_bool_t replace,
				  oskit_u32_t seqno);


/*
 * Remove one or more entries from the access vector
 * cache.
 *
 * Any one of the four attributes (sclass, ssid, tclass, tsid)
 * may be wildcarded; SECCLASS_WILD may be used to wildcard
 * sclass or tclass; SECSID_WILD may be used to wildcard ssid
 * or tsid.  
 *
 * A wired entry will only be removed if removewired is TRUE.
 *
 * Returns OSKIT_ENOENT if no wildcards were used and no entry
 * existed for the specified (sclass,ssid,tclass,tsid) quadruple, or
 * 0 otherwise.
 */
oskit_error_t nest_server_avc_remove(security_class_t sclass,
				     security_id_t ssid,
				     security_class_t tclass,
				     security_id_t tsid,
				     oskit_bool_t removewired);


/*
 * Register the specified callback for reset requests.
 *
 * All registered callbacks will be invoked on a reset
 * request after the cache has been cleared.
 */ 
oskit_error_t 
nest_server_avc_register_reset_callback(void (*callback)(void));

/*
 * Reset the access vector cache.
 */
oskit_error_t nest_server_avc_reset(oskit_u32_t seqno);


/*
 * Register the specified callback for revocation requests
 * which match the specified sclass, ssid, tclass, tsid
 * and revoke parameters.  The class and SID parameters
 * to this function may be wildcarded, in which case
 * they will match any value in a revocation request.
 * The revoke value for the callback matches the revoke value
 * in a revocation request if they have a nonnull intersection.
 *
 * All matching callbacks will be invoked on a revocation
 * request, and will be provided with the actual class
 * and SID parameters to the revocation request.  The
 * revoke parameter to the callback will be the intersection
 * of the actual revoke parameter to the request and the
 * registered revoke vector for the callback.
 * These callbacks are expected to deal with migrated permissions.
 */
oskit_error_t 
nest_server_avc_register_revoke_callback(security_class_t sclass,
					 security_id_t ssid, 
					 security_class_t tclass,
					 security_id_t tsid, 
					 access_vector_t revoke,
					 void (*callback)(
						   security_class_t sclass,
						   security_id_t ssid, 
						   security_class_t tclass,
						   security_id_t tsid, 
						   access_vector_t revoke));


/*
 * Revoke the specified permissions from the specified entry
 * in the access vector cache and invoke the appropriate
 * callbacks.  Wildcard values may not be used.
 *
 * A wired entry will only be updated if revokewired is TRUE.
 */
oskit_error_t nest_server_avc_revoke(security_class_t sclass,
				     security_id_t ssid,
				     security_class_t tclass,
				     security_id_t tsid,
				     access_vector_t revoke,
				     oskit_bool_t revokewired,
				     oskit_u32_t seqno);


/*
 * Register the specified callback for try_remove requests.
 */
oskit_error_t
nest_server_avc_register_try_remove_callback(security_class_t sclass,
                                         security_id_t ssid,
                                         security_class_t tclass,
                                         security_id_t tsid,
                                         access_vector_t permissions,
                                         void (*callback)( 
                                                   security_class_t sclass,
                                                   security_id_t ssid,
                                                   security_class_t tclass,
                                                   security_id_t tsid,
                                                   access_vector_t permissions,
                                                   access_vector_t *out_retained));

/*
 * Try to remove the specified permissions from the specified entry,
 * but only if they are not retained in the object manager's state.
 * Return the subset of permissions which are retained.
 * If removewired is FALSE and the entry is wired, then any
 * permissions granted in the wired node are also considered
 * to be retained.
 */
oskit_error_t nest_server_avc_try_remove(security_class_t sclass,
                                     security_id_t ssid,
                                     security_class_t tclass,
                                     security_id_t tsid,
                                     access_vector_t permissions,
                                     oskit_bool_t removewired,
                                     oskit_u32_t seqno,
                                     access_vector_t *out_retained);


#ifdef DEBUG
/*
 * Display the contents of an access vector, interpreted
 * with respect to a given (sclass,tclass) pair.
 */
void nest_server_avc_dump_av(security_class_t sclass,
			     security_class_t tclass, 
			     access_vector_t av);

/*
 * Display the contents of the access vector cache.
 */
void nest_server_avc_dump(char *tag);
#endif

void nest_server_avc_exit(void);


#endif _FLUKE_NEST_SERVER_AVC_H_

#endif

