/*
 * Copyright (c) 1995, 1996, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#ifndef _FLUKE_X86_IPC_H_
#define _FLUKE_X86_IPC_H_

typedef oskit_addr_t fluke_ipc_word_t;

#define FLUKE_MIN_MSG_WORDS	2	/* minimum message size in ipc_words */

#define FLUKE_MSG_ALIGN_SHIFT	0	/* any byte alignment works */
#define FLUKE_MSG_GRAN_SHIFT	2	/* size must be multiple of a word */

typedef unsigned char fluke_ipc_buf_count_t;
typedef unsigned char fluke_ipc_ref_count_t;

#define FLUKE_MAX_BUFS		255	/* max buffers per IPC operation */
#define FLUKE_MAX_REFS		255	/* max refs per IPC operation */

/*
 * Although the contents of this structure are machine-independent,
 * its specific layout is architecture-specific.
 */
struct fluke_ipc_params
{
	fluke_ipc_buf_count_t	send_buf_count;
	fluke_ipc_ref_count_t	send_ref_count;
	fluke_ipc_buf_count_t	recv_buf_count;
	fluke_ipc_ref_count_t	recv_ref_count;

	struct fluke_ipc_buf	send_buf;
	struct fluke_ipc_buf	*send_buf_tab;
	struct fluke_ref	**send_ref_tab;

	struct fluke_ipc_buf	recv_buf;
	struct fluke_ipc_buf	*recv_buf_tab;
	struct fluke_ref	**recv_ref_tab;
};

#if defined(__GNUC__) && defined(FLUKE_INLINING)

#include <oskit/compiler.h>
#ifndef FLUKE_ENTRY_INLINE
#define FLUKE_ENTRY_INLINE OSKIT_INLINE
#endif
#ifdef GPROF
#define e(name) extern char _fluke_##name##_[];
#include <fluke/x86/syscall_list.h>
#undef e
#include <oskit/x86/proc_reg.h>
#include <fluke/profile.h>
#endif

/*** Idempotent IPC ***/

FLUKE_ENTRY_INLINE fluke_error_t fluke_ipc_send(
	struct fluke_ref	*destination_port_ref,
	struct fluke_ipc_params	*ipc_params)
{
	unsigned eax;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_send_);
#endif
	asm volatile("
		pushl	%%ebp
		call	_fluke_ipc_send
		popl	%%ebp
	" : "=a" (eax)
	  : "a" (destination_port_ref),
	    "b" (ipc_params)
	  : "ebx", "ecx", "edx", "esi", "edi", "memory");
	return (unsigned char)(eax >> 16);
}

#if FLASK
FLUKE_ENTRY_INLINE fluke_error_t fluke_ipc_send_secure(
	struct fluke_ref	*destination_port_ref,
	struct fluke_ipc_params	*ipc_params,
        security_id_t 		client_sid)
{
	unsigned eax;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_send_secure_);
#endif
	asm volatile("
		pushl	%%ebp
		call	_fluke_ipc_send_secure
		popl	%%ebp
	" : "=a" (eax)
	  : "a" (destination_port_ref),
	    "b" (ipc_params),
	    "d" (client_sid)
	  : "ebx", "ecx", "edx", "esi", "edi", "memory");
	return (unsigned char)(eax >> 16);
}
#endif

FLUKE_ENTRY_INLINE fluke_error_t fluke_ipc_call(
	struct fluke_ref	*destination_port_ref,
	struct fluke_ipc_params	*ipc_params)
{
	unsigned eax;
	unsigned char rc;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_call_);
#endif
	asm volatile("
		pushl	%%ebp
		call	_fluke_ipc_call
		popl	%%ebp
	" : "=a" (eax),
	    "=c" (ipc_params->recv_buf.words),
	    "=S" (((fluke_ipc_word_t*)ipc_params->recv_buf.buf)[0]),
	    "=D" (((fluke_ipc_word_t*)ipc_params->recv_buf.buf)[1])
	  : "a" (destination_port_ref),
	    "b" (ipc_params)
	  : "ebx", "edx", "memory");
	rc = eax >> 16;

	/* sets both recv_buf_count and recv_ref_count. */
	*(unsigned short*)&ipc_params->recv_buf_count = eax;

	return rc;
}

#if FLASK
FLUKE_ENTRY_INLINE fluke_error_t fluke_ipc_call_secure(
	struct fluke_ref	*destination_port_ref,
	struct fluke_ipc_params	*ipc_params,
        security_id_t 		client_sid)     
{
	unsigned eax;
	unsigned char rc;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_call_secure_);
#endif
	asm volatile("
		pushl	%%ebp
		call	_fluke_ipc_call_secure
		popl	%%ebp
	" : "=a" (eax),
	    "=c" (ipc_params->recv_buf.words),
	    "=S" (((fluke_ipc_word_t*)ipc_params->recv_buf.buf)[0]),
	    "=D" (((fluke_ipc_word_t*)ipc_params->recv_buf.buf)[1])
	  : "a" (destination_port_ref),
	    "b" (ipc_params),
	    "d" (client_sid)
	  : "ebx", "edx", "memory");
	rc = eax >> 16;
	/* sets both recv_buf_count and recv_ref_count */
	*(unsigned short*)&ipc_params->recv_buf_count = eax;
	return rc;
}
#endif 

FLUKE_ENTRY_INLINE fluke_error_t fluke_ipc_reply(
	struct fluke_ipc_params	*ipc_params)
{
	unsigned eax;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_reply_);
#endif
	asm volatile("
		pushl	%%ebp
		call	_fluke_ipc_reply
		popl	%%ebp
	" : "=a" (eax)
	  : "a" (*(unsigned short*)&ipc_params->send_buf_count),
		     /* eax gets both buf_count and ref_count */
	    "c" (ipc_params->send_buf.words),
	    "d" (ipc_params->send_ref_tab),
	    "S" (ipc_params->send_buf.buf),
	    "D" (ipc_params->send_buf_tab)
	  : "ebx", "ecx", "edx", "esi", "edi");
	return (unsigned char)(eax >> 16);
}

FLUKE_ENTRY_INLINE fluke_error_t fluke_ipc_reply_wait_receive(
	struct fluke_ipc_params	*ipc_params,
	fluke_alias_t		*out_port_alias)
{
	unsigned eax, ecx;
	unsigned char rc;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_reply_wait_receive_);
#endif
	asm volatile("
		pushl	%%ebp
		movl	4(%%ebx),%%ecx
		call	_fluke_ipc_reply_wait_receive
		popl	%%ebp
	" : "=a" (eax),
	    "=b" (*out_port_alias),
	    "=c" (ecx),
	    "=S" (((fluke_ipc_word_t*)ipc_params->recv_buf.buf)[0]),
	    "=D" (((fluke_ipc_word_t*)ipc_params->recv_buf.buf)[1])
	  : "a" (*(unsigned short*)&ipc_params->send_buf_count),
		     /* eax gets both buf_count and ref_count */
	    "b" (ipc_params),
	    /*"c" (ipc_params->send_buf.words),*/
	    "d" (ipc_params->send_ref_tab),
	    "S" (ipc_params->send_buf.buf),
	    "D" (ipc_params->send_buf_tab)
	  : "edx", "memory");
	rc = eax >> 16;
	if ((rc & FLUKE_OP_MASK) == FLUKE_OP_IPC_RECEIVE ||
	    (rc & FLUKE_OP_MASK) == FLUKE_OP_IPC_WAIT)
	{
		/* sets both recv_buf_count and recv_ref_count */
		*(unsigned short*)&ipc_params->recv_buf_count = eax;
		ipc_params->recv_buf.words = ecx;
	}
	return rc;
}

#if FLASK
FLUKE_ENTRY_INLINE fluke_error_t fluke_ipc_reply_wait_receive_secure(
	struct fluke_ipc_params	*ipc_params,
	fluke_alias_t		*out_port_alias,
        security_id_t		*out_client_sid) 
{
	unsigned eax, ecx;
	unsigned char rc;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_reply_wait_receive_secure_);
#endif

	asm volatile("
		pushl	%%ebp
		movl	4(%%ebx),%%ecx
		call	_fluke_ipc_reply_wait_receive_secure
		popl	%%ebp
	" : "=a" (eax),
	    "=b" (*out_port_alias),
	    "=c" (ecx),
	    "=S" (*out_client_sid)
	  : "a" (*(unsigned short*)&ipc_params->send_buf_count),
		     /* eax gets both buf_count and ref_count */
	    "b" (ipc_params),
	    /*"c" (ipc_params->send_buf.words),*/
	    "d" (ipc_params->send_ref_tab),
	    "S" (ipc_params->send_buf.buf),
	    "D" (ipc_params->send_buf_tab)
	  : "edx", "memory");
	rc = eax >> 16;
	if ((rc & FLUKE_OP_MASK) == FLUKE_OP_IPC_RECEIVE ||
	    (rc & FLUKE_OP_MASK) == FLUKE_OP_IPC_WAIT)
	{
		/* sets both recv_buf_count and recv_ref_count */
		*(unsigned short*)&ipc_params->recv_buf_count = eax;
		ipc_params->recv_buf.words = ecx;
	}
	return rc;
}
#endif 


/*** Reliable IPC ***/

FLUKE_ENTRY_INLINE fluke_error_t fluke_ipc_client_connect_send(
	struct fluke_ref	*destination_port_ref,
	struct fluke_ipc_params	*ipc_params)
{
	unsigned eax;
#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_client_connect_send_);
#endif
	asm volatile("
		pushl	%%ebp
		call	_fluke_ipc_client_connect_send
		popl	%%ebp
	" : "=a" (eax)
	  : "a" (destination_port_ref),
	    "b" (ipc_params)
	  : "ebx", "ecx", "edx", "esi", "edi");
	return (unsigned char)(eax >> 16);
}

#if FLASK
FLUKE_ENTRY_INLINE fluke_error_t fluke_ipc_client_connect_send_secure(
	struct fluke_ref	*destination_port_ref,
	struct fluke_ipc_params	*ipc_params,
        security_id_t 		client_sid)      
{
	unsigned eax;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_client_connect_send_secure_);
#endif
	asm volatile("
		pushl	%%ebp
		call	_fluke_ipc_client_connect_send_secure
		popl	%%ebp
	" : "=a" (eax)
	  : "a" (destination_port_ref),
	    "b" (ipc_params),
            "d" (client_sid)
	  : "ebx", "ecx", "edx", "esi", "edi");
	return (unsigned char)(eax >> 16);
}
#endif

FLUKE_ENTRY_INLINE fluke_error_t fluke_ipc_client_ack_send(
	struct fluke_ipc_params	*ipc_params)
{
	unsigned eax;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_client_ack_send_);
#endif
	asm volatile("
		pushl	%%ebp
		call	_fluke_ipc_client_ack_send
		popl	%%ebp
	" : "=a" (eax)
	  : "a" (*(unsigned short*)&ipc_params->send_buf_count),
		     /* eax gets both send_buf_count and send_ref_count */
	    "c" (ipc_params->send_buf.words),
	    "d" (ipc_params->send_ref_tab),
	    "S" (ipc_params->send_buf.buf),
	    "D" (ipc_params->send_buf_tab)
	  : "ebx", "ecx", "edx", "esi", "edi");
	return (unsigned char)(eax >> 16);
}

FLUKE_ENTRY_INLINE fluke_error_t fluke_ipc_client_send(
	struct fluke_ipc_params	*ipc_params)
{
	unsigned eax;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_client_send_);
#endif
	asm volatile("
		pushl	%%ebp
		call	_fluke_ipc_client_send
		popl	%%ebp
	" : "=a" (eax)
	  : "a" (*(unsigned short*)&ipc_params->send_buf_count),
		     /* eax gets both buf_count and ref_count */
	    "c" (ipc_params->send_buf.words),
	    "d" (ipc_params->send_ref_tab),
	    "S" (ipc_params->send_buf.buf),
	    "D" (ipc_params->send_buf_tab)
	  : "ebx", "ecx", "edx", "esi", "edi");
	return (unsigned char)(eax >> 16);
}

FLUKE_ENTRY_INLINE fluke_error_t fluke_ipc_client_connect_send_over_receive(
	struct fluke_ref	*destination_port_ref,
	struct fluke_ipc_params	*ipc_params)
{
	unsigned eax, ecx;
	unsigned char rc;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_client_connect_send_over_receive_);
#endif
	asm volatile("
		pushl	%%ebp
		call	_fluke_ipc_client_connect_send_over_receive
		popl	%%ebp
	" : "=a" (eax),
	    "=c" (ecx),
	    "=S" (((fluke_ipc_word_t*)ipc_params->recv_buf.buf)[0]),
	    "=D" (((fluke_ipc_word_t*)ipc_params->recv_buf.buf)[1])
	  : "a" (destination_port_ref),
	    "b" (ipc_params)
	  : "ebx", "edx", "memory");
	rc = eax >> 16;
	if ((rc & FLUKE_OP_MASK) == FLUKE_OP_IPC_RECEIVE)
	{
		/* sets both recv_buf_count and recv_ref_count */
		*(unsigned short*)&ipc_params->recv_buf_count = eax;
		ipc_params->recv_buf.words = ecx;
	}
	return rc;
}

#if FLASK
FLUKE_ENTRY_INLINE fluke_error_t fluke_ipc_client_connect_send_over_receive_secure(
	struct fluke_ref	*destination_port_ref,
	struct fluke_ipc_params	*ipc_params,
        security_id_t 		client_sid)   
{
	unsigned eax, ecx;
	unsigned char rc;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_client_connect_send_over_receive_secure_);
#endif
	asm volatile("
		pushl	%%ebp
		call	_fluke_ipc_client_connect_send_over_receive_secure
		popl	%%ebp
	" : "=a" (eax),
	    "=c" (ecx),
	    "=S" (((fluke_ipc_word_t*)ipc_params->recv_buf.buf)[0]),
	    "=D" (((fluke_ipc_word_t*)ipc_params->recv_buf.buf)[1])
	  : "a" (destination_port_ref),
	    "b" (ipc_params),
            "d" (client_sid)
	  : "ebx", "edx", "memory");
	rc = eax >> 16;
	if ((rc & FLUKE_OP_MASK) == FLUKE_OP_IPC_RECEIVE)
	{
		/* sets both recv_buf_count and recv_ref_count */
		*(unsigned short*)&ipc_params->recv_buf_count = eax;
		ipc_params->recv_buf.words = ecx;
	}
	return rc;
}
#endif

FLUKE_ENTRY_INLINE fluke_error_t fluke_ipc_client_ack_send_over_receive(
	struct fluke_ipc_params	*ipc_params)
{
	unsigned eax, ecx;
	unsigned char rc;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_client_ack_send_over_receive_);
#endif
	asm volatile("
		pushl	%%ebp
		movl	4(%%ebx),%%ecx
		call	_fluke_ipc_client_ack_send_over_receive
		popl	%%ebp
	" : "=a" (eax),
	    "=c" (ecx),
	    "=S" (((fluke_ipc_word_t*)ipc_params->recv_buf.buf)[0]),
	    "=D" (((fluke_ipc_word_t*)ipc_params->recv_buf.buf)[1])
	  : "a" (*(unsigned short*)&ipc_params->send_buf_count),
		     /* eax gets both send_buf_count and send_ref_count */
	    "b" (ipc_params),
	    /*"c" (ipc_params->send_buf.words),*/
	    "d" (ipc_params->send_ref_tab),
	    "S" (ipc_params->send_buf.buf),
	    "D" (ipc_params->send_buf_tab)
	  : "ebx", "edx", "memory");
	rc = eax >> 16;
	if ((rc & FLUKE_OP_MASK) == FLUKE_OP_IPC_RECEIVE)
	{
		/* sets both recv_buf_count and recv_ref_count */
		*(unsigned short*)&ipc_params->recv_buf_count = eax;
		ipc_params->recv_buf.words = ecx;
	}
	return rc;
}

FLUKE_ENTRY_INLINE fluke_error_t fluke_ipc_client_send_over_receive(
	struct fluke_ipc_params	*ipc_params)
{
	unsigned eax, ecx;
	unsigned char rc;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_client_send_over_receive_);
#endif
	asm volatile("
		pushl	%%ebp
		movl	4(%%ebx),%%ecx
		call	_fluke_ipc_client_send_over_receive
		popl	%%ebp
	" : "=a" (eax),
	    "=c" (ecx),
	    "=S" (((fluke_ipc_word_t*)ipc_params->recv_buf.buf)[0]),
	    "=D" (((fluke_ipc_word_t*)ipc_params->recv_buf.buf)[1])
	  : "a" (*(unsigned short*)&ipc_params->send_buf_count),
		     /* eax gets both buf_count and ref_count */
	    "b" (ipc_params),
	    /*"c" (ipc_params->send_buf.words),*/
	    "d" (ipc_params->send_ref_tab),
	    "S" (ipc_params->send_buf.buf),
	    "D" (ipc_params->send_buf_tab)
	  : "ebx", "edx", "memory");
	rc = eax >> 16;
	if ((rc & FLUKE_OP_MASK) == FLUKE_OP_IPC_RECEIVE)
	{
		/* sets both recv_buf_count and recv_ref_count */
		*(unsigned short*)&ipc_params->recv_buf_count = eax;
		ipc_params->recv_buf.words = ecx;
	}
	return rc;
}

FLUKE_ENTRY_INLINE fluke_error_t fluke_ipc_client_over_receive(
	struct fluke_ipc_params	*ipc_params)
{
	unsigned eax, ecx;
	unsigned char rc;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_client_send_over_receive_);
#endif
	asm volatile("
		pushl	%%ebp
		call	_fluke_ipc_client_send_over_receive
		popl	%%ebp
	" : "=a" (eax),
	    "=c" (ecx),
	    "=S" (((fluke_ipc_word_t*)ipc_params->recv_buf.buf)[0]),
	    "=D" (((fluke_ipc_word_t*)ipc_params->recv_buf.buf)[1])
	  : "a" (0), /* send_buf_count and send_ref_count */
	    "b" (ipc_params),
	    "c" (0) /* send_buf_words */
	  : "ebx", "edx", "memory");
	rc = eax >> 16;
	if ((rc & FLUKE_OP_MASK) == FLUKE_OP_IPC_RECEIVE)
	{
		/* sets both recv_buf_count and recv_ref_count */
		*(unsigned short*)&ipc_params->recv_buf_count = eax;
		ipc_params->recv_buf.words = ecx;
	}
	return rc;
}

FLUKE_ENTRY_INLINE fluke_error_t fluke_ipc_client_receive(
	struct fluke_ipc_params	*ipc_params)
{
	unsigned eax, ecx;
	unsigned char rc;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_client_receive_);
#endif
	asm volatile("
		pushl	%%ebp
		call	_fluke_ipc_client_receive
		popl	%%ebp
	" : "=a" (eax),
	    "=c" (ecx)
	  : "a" (*(unsigned short*)&ipc_params->recv_buf_count),
		     /* eax gets both buf_count and ref_count */
	    "c" (ipc_params->recv_buf.words),
	    "d" (ipc_params->recv_ref_tab),
	    "S" (ipc_params->recv_buf_tab),
	    "D" (((fluke_ipc_word_t*)ipc_params->recv_buf.buf))
	  : "ebx", "edx", "esi", "edi", "memory");
	rc = eax >> 16;

	/* sets both recv_buf_count and recv_ref_count */
	*(unsigned short*)&ipc_params->recv_buf_count = eax;
	ipc_params->recv_buf.words = ecx;
	return rc;
}

FLUKE_ENTRY_INLINE void fluke_ipc_client_disconnect(void)
{
#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_client_disconnect_);
#endif
	asm volatile("
		call	_fluke_ipc_client_disconnect
	" : : : "eax", "ecx", "edx");
}

FLUKE_ENTRY_INLINE fluke_error_t fluke_ipc_server_ack_send(
	struct fluke_ipc_params	*ipc_params)
{
	unsigned eax;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_server_ack_send_);
#endif
	asm volatile("
		pushl	%%ebp
		call	_fluke_ipc_server_ack_send
		popl	%%ebp
	" : "=a" (eax)
	  : "a" (*(unsigned short*)&ipc_params->send_buf_count),
		     /* eax gets both send_buf_count and send_ref_count */
	    "c" (ipc_params->send_buf.words),
	    "d" (ipc_params->send_ref_tab),
	    "S" (ipc_params->send_buf.buf),
	    "D" (ipc_params->send_buf_tab)
	  : "ebx", "ecx", "edx", "esi", "edi");
	return (unsigned char)(eax >> 16);
}

FLUKE_ENTRY_INLINE fluke_error_t fluke_ipc_server_send(
	struct fluke_ipc_params	*ipc_params)
{
	unsigned eax;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_server_send_);
#endif
	asm volatile("
		pushl	%%ebp
		call	_fluke_ipc_server_send
		popl	%%ebp
	" : "=a" (eax)
	  : "a" (*(unsigned short*)&ipc_params->send_buf_count),
		     /* eax gets both send_buf_count and send_ref_count */
	    "c" (ipc_params->send_buf.words),
	    "d" (ipc_params->send_ref_tab),
	    "S" (ipc_params->send_buf.buf),
	    "D" (ipc_params->send_buf_tab)
	  : "ebx", "ecx", "edx", "esi", "edi");
	return (unsigned char)(eax >> 16);
}

FLUKE_ENTRY_INLINE fluke_error_t fluke_ipc_server_ack_send_wait_receive(
	struct fluke_ipc_params	*ipc_params,
	fluke_alias_t		*out_port_alias)
{
	unsigned eax, ecx;
	unsigned char rc;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_server_ack_send_wait_receive_);
#endif
	asm volatile("
		pushl	%%ebp
		movl	4(%%ebx),%%ecx
		call	_fluke_ipc_server_ack_send_wait_receive
		popl	%%ebp
	" : "=a" (eax),
	    "=b" (*out_port_alias),
	    "=c" (ecx),
	    "=S" (((fluke_ipc_word_t*)ipc_params->recv_buf.buf)[0]),
	    "=D" (((fluke_ipc_word_t*)ipc_params->recv_buf.buf)[1])
	  : "a" (*(unsigned short*)&ipc_params->send_buf_count),
		     /* eax gets both send_buf_count and send_ref_count */
	    "b" (ipc_params),
	    /*"c" (ipc_params->send_buf.words),*/
	    "d" (ipc_params->send_ref_tab),
	    "S" (ipc_params->send_buf.buf),
	    "D" (ipc_params->send_buf_tab)
	  : "edx", "esi", "edi", "memory");
	rc = eax >> 16;
	if ((rc & FLUKE_OP_MASK) == FLUKE_OP_IPC_RECEIVE ||
	    (rc & FLUKE_OP_MASK) == FLUKE_OP_IPC_WAIT)
	{
		/* sets both recv_buf_count and recv_ref_count */
		*(unsigned short*)&ipc_params->recv_buf_count = eax;
		ipc_params->recv_buf.words = ecx;
	}
	return rc;
}

#if FLASK
FLUKE_ENTRY_INLINE fluke_error_t fluke_ipc_server_ack_send_wait_receive_secure(
	struct fluke_ipc_params	*ipc_params,
	fluke_alias_t		*out_port_alias,
        security_id_t		*out_client_sid)
{
	unsigned eax, ecx;
	unsigned char rc;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_server_ack_send_wait_receive_secure_);
#endif

	asm volatile("
		pushl	%%ebp
		movl	4(%%ebx),%%ecx
		call	_fluke_ipc_server_ack_send_wait_receive_secure
		popl	%%ebp
	" : "=a" (eax),
	    "=b" (*out_port_alias),
	    "=c" (ecx),
	    "=S" (*out_client_sid)
	  : "a" (*(unsigned short*)&ipc_params->send_buf_count),
		     /* eax gets both send_buf_count and send_ref_count */
	    "b" (ipc_params),
	    /*"c" (ipc_params->send_buf.words),*/
	    "d" (ipc_params->send_ref_tab),
	    "S" (ipc_params->send_buf.buf),
	    "D" (ipc_params->send_buf_tab)
	  : "edx", "esi", "edi", "memory");
	rc = eax >> 16;
	if ((rc & FLUKE_OP_MASK) == FLUKE_OP_IPC_RECEIVE ||
	    (rc & FLUKE_OP_MASK) == FLUKE_OP_IPC_WAIT)
	{
		/* sets both recv_buf_count and recv_ref_count */
		*(unsigned short*)&ipc_params->recv_buf_count = eax;
		ipc_params->recv_buf.words = ecx;
	}
	return rc;
}
#endif

FLUKE_ENTRY_INLINE fluke_error_t fluke_ipc_server_ack_send_over_receive(
	struct fluke_ipc_params	*ipc_params)
{
	unsigned eax, ecx;
	unsigned char rc;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_server_ack_send_over_receive_);
#endif
	asm volatile("
		pushl	%%ebp
		movl	4(%%ebx),%%ecx
		call	_fluke_ipc_server_ack_send_over_receive
		popl	%%ebp
	" : "=a" (eax),
	    "=c" (ecx),
	    "=S" (((fluke_ipc_word_t*)ipc_params->recv_buf.buf)[0]),
	    "=D" (((fluke_ipc_word_t*)ipc_params->recv_buf.buf)[1])
	  : "a" (*(unsigned short*)&ipc_params->send_buf_count),
		     /* eax gets both send_buf_count and send_ref_count */
	    "b" (ipc_params),
	    /*"c" (ipc_params->send_buf.words),*/
	    "d" (ipc_params->send_ref_tab),
	    "S" (ipc_params->send_buf.buf),
	    "D" (ipc_params->send_buf_tab)
	  : "ebx", "edx", "memory");
	rc = eax >> 16;
	if ((rc & FLUKE_OP_MASK) == FLUKE_OP_IPC_RECEIVE)
	{
		/* sets both recv_buf_count and recv_ref_count */
		*(unsigned short*)&ipc_params->recv_buf_count = eax;
		ipc_params->recv_buf.words = ecx;
	}
	return rc;
}

FLUKE_ENTRY_INLINE fluke_error_t fluke_ipc_server_send_wait_receive(
	struct fluke_ipc_params	*ipc_params,
	fluke_alias_t		*out_port_alias)
{
	unsigned eax, ecx;
	unsigned char rc;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_server_send_wait_receive_);
#endif
	asm volatile("
		pushl	%%ebp
		movl	4(%%ebx),%%ecx
		call	_fluke_ipc_server_send_wait_receive
		popl	%%ebp
	" : "=a" (eax),
	    "=b" (*out_port_alias),
	    "=c" (ecx),
	    "=S" (((fluke_ipc_word_t*)ipc_params->recv_buf.buf)[0]),
	    "=D" (((fluke_ipc_word_t*)ipc_params->recv_buf.buf)[1])
	  : "a" (*(unsigned short*)&ipc_params->send_buf_count),
		     /* eax gets both send_buf_count and send_ref_count */
	    "b" (ipc_params),
	    /*"c" (ipc_params->send_buf.words),*/
	    "d" (ipc_params->send_ref_tab),
	    "S" (ipc_params->send_buf.buf),
	    "D" (ipc_params->send_buf_tab)
	  : "edx", "memory");
	rc = eax >> 16;
	if ((rc & FLUKE_OP_MASK) == FLUKE_OP_IPC_RECEIVE ||
	    (rc & FLUKE_OP_MASK) == FLUKE_OP_IPC_WAIT)
	{
		/* sets both recv_buf_count and recv_ref_count */
		*(unsigned short*)&ipc_params->recv_buf_count = eax;
		ipc_params->recv_buf.words = ecx;
	}
	return rc;
}

#if FLASK
FLUKE_ENTRY_INLINE fluke_error_t fluke_ipc_server_send_wait_receive_secure(
	struct fluke_ipc_params	*ipc_params,
	fluke_alias_t		*out_port_alias,
        security_id_t		*out_client_sid)
{
	unsigned eax, ecx;
	unsigned char rc;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_server_send_wait_receive_secure_);
#endif

	asm volatile("
		pushl	%%ebp
		movl	4(%%ebx),%%ecx
		call	_fluke_ipc_server_send_wait_receive_secure
		popl	%%ebp
	" : "=a" (eax),
	    "=b" (*out_port_alias),
	    "=c" (ecx),
	    "=S" (*out_client_sid)
	  : "a" (*(unsigned short*)&ipc_params->send_buf_count),
		     /* eax gets both send_buf_count and send_ref_count */
	    "b" (ipc_params),
	    /*"c" (ipc_params->send_buf.words),*/
	    "d" (ipc_params->send_ref_tab),
	    "S" (ipc_params->send_buf.buf),
	    "D" (ipc_params->send_buf_tab)
	  : "edx", "memory");
	rc = eax >> 16;
	if ((rc & FLUKE_OP_MASK) == FLUKE_OP_IPC_RECEIVE ||
	    (rc & FLUKE_OP_MASK) == FLUKE_OP_IPC_WAIT)
	{
		/* sets both recv_buf_count and recv_ref_count */
		*(unsigned short*)&ipc_params->recv_buf_count = eax;
		ipc_params->recv_buf.words = ecx;
	}
	return rc;
}
#endif

FLUKE_ENTRY_INLINE fluke_error_t fluke_ipc_server_send_over_receive(
	struct fluke_ipc_params	*ipc_params)
{
	unsigned eax, ecx;
	unsigned char rc;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_server_send_over_receive_);
#endif
	asm volatile("
		pushl	%%ebp
		movl	4(%%ebx),%%ecx
		call	_fluke_ipc_server_send_over_receive
		popl	%%ebp
	" : "=a" (eax),
	    "=c" (ecx),
	    "=S" (((fluke_ipc_word_t*)ipc_params->recv_buf.buf)[0]),
	    "=D" (((fluke_ipc_word_t*)ipc_params->recv_buf.buf)[1])
	  : "a" (*(unsigned short*)&ipc_params->send_buf_count),
		     /* eax gets both send_buf_count and send_ref_count */
	    "b" (ipc_params),
	    /*"c" (ipc_params->send_buf.words),*/
	    "d" (ipc_params->send_ref_tab),
	    "S" (ipc_params->send_buf.buf),
	    "D" (ipc_params->send_buf_tab)
	  : "ebx", "edx", "memory");
	rc = eax >> 16;
	if ((rc & FLUKE_OP_MASK) == FLUKE_OP_IPC_RECEIVE)
	{
		/* sets both recv_buf_count and recv_ref_count */
		*(unsigned short*)&ipc_params->recv_buf_count = eax;
		ipc_params->recv_buf.words = ecx;
	}
	return rc;
}

FLUKE_ENTRY_INLINE fluke_error_t fluke_ipc_server_over_receive(
	struct fluke_ipc_params	*ipc_params)
{
	unsigned eax, ecx;
	unsigned char rc;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_server_send_over_receive_);
#endif
	asm volatile("
		pushl	%%ebp
		call	_fluke_ipc_server_send_over_receive
		popl	%%ebp
	" : "=a" (eax),
	    "=c" (ecx),
	    "=S" (((fluke_ipc_word_t*)ipc_params->recv_buf.buf)[0]),
	    "=D" (((fluke_ipc_word_t*)ipc_params->recv_buf.buf)[1])
	  : "a" (0), /* send_buf_count and send_ref_count */
	    "b" (ipc_params),
	    "c" (0) /* send_buf_words */
	  : "ebx", "edx", "memory");
	rc = eax >> 16;
	if ((rc & FLUKE_OP_MASK) == FLUKE_OP_IPC_RECEIVE)
	{
		/* sets both recv_buf_count and recv_ref_count */
		*(unsigned short*)&ipc_params->recv_buf_count = eax;
		ipc_params->recv_buf.words = ecx;
	}
	return rc;
}

FLUKE_ENTRY_INLINE fluke_error_t fluke_ipc_server_receive(
	struct fluke_ipc_params	*ipc_params)
{
	unsigned eax, ecx;
	unsigned char rc;

	/*
	 * If you are astute, you will notice that the MIN_MSG_WORDs
	 * that the other *_receive_* functions deal with are not
	 * included here.  This is because this is an operation that
	 * you can sit in a loop performing.  For example, if someone
	 * sends a message of N words and the receiver's buffer is
	 * only N-1 words, the receiver will get the MORE_DATA
	 * indication.  The receiver will only get 1 word the next
	 * time it goes back into server_receive().  We cannot require
	 * that it has to get MIN_MSG_WORDS in this case.
	 */
#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_server_receive_);
#endif
	asm volatile("
		pushl	%%ebp
		call	_fluke_ipc_server_receive
		popl	%%ebp
	" : "=a" (eax),
	    "=c" (ecx)
	  : "a" (*(unsigned short*)&ipc_params->recv_buf_count),
		     /* eax gets both buf_count and ref_count */
	    "c" (ipc_params->recv_buf.words),
	    "d" (ipc_params->recv_ref_tab),
	    "S" (ipc_params->recv_buf_tab),
	    "D" (((fluke_ipc_word_t*)ipc_params->recv_buf.buf))
	  : "ebx", "edx", "esi", "edi", "memory");
	rc = eax >> 16;
	/* sets both recv_buf_count and recv_ref_count */
	*(unsigned short*)&ipc_params->recv_buf_count = eax;
	ipc_params->recv_buf.words = ecx;
	return rc;
}

FLUKE_ENTRY_INLINE void fluke_ipc_server_disconnect(void)
{
#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_server_disconnect_);
#endif
	asm volatile("
		call	_fluke_ipc_server_disconnect
	" : : : "eax", "ecx", "edx");
}

/* XXX this probably should be a macro,
   since it'll probably malfunction if not inlined. */
FLUKE_ENTRY_INLINE fluke_error_t fluke_ipc_setup_wait_receive(
	struct fluke_pset	*server_pset,
	struct fluke_ipc_params	*ipc_params,
	fluke_alias_t		*out_port_alias)
{
	unsigned eax;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_setup_);
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_wait_receive_);
#endif
	asm volatile("
		pushl	%%ebp
		movl	%%esp,%%ecx
		movl	$1f,%%edx
		movl	$2f,%%esi
		movl	$3f,%%edi
		call	_fluke_ipc_setup
		movzwl	2(%%ebx),%%eax
		movl	20(%%ebx),%%ecx
		movl	24(%%ebx),%%edi
		movl	28(%%ebx),%%esi
		movl	32(%%ebx),%%edx
		call	_fluke_ipc_wait_receive
		jmp	3f
	1:	orl	%7,%%eax
		jmp	3f
	2:	orl	%8,%%eax
	3:	popl	%%ebp
	" : "=a" (eax),
	    "=b" (*out_port_alias),
	    "=c" (ipc_params->recv_buf.words),
	    "=S" (((fluke_ipc_word_t*)ipc_params->recv_buf.buf)[0]),
	    "=D" (((fluke_ipc_word_t*)ipc_params->recv_buf.buf)[1])
	  : "a" (server_pset),
	    "b" (ipc_params),
	    "i" (FLUKE_IPC_WAIT_ONEWAY << 16),
	    "i" (FLUKE_IPC_WAIT_IDEMPOTENT << 16),
	    /*
	     * XXX
	     * We have to add a garbage in-value to edx because the compiler
	     * was not recognizing that edx was being trashed.
	     * This adds a movl instruction, but at least it works right.
	     */
	    "d" (0xf1ac)
	  : "edx", "memory");

	/* sets both recv_buf_count and recv_ref_count */
	*(unsigned short*)&ipc_params->recv_buf_count = eax;
	return (unsigned char)(eax >> 16);
}

#if FLASK
FLUKE_ENTRY_INLINE fluke_error_t fluke_ipc_setup_wait_receive_secure(
	struct fluke_pset	*server_pset,
	struct fluke_ipc_params	*ipc_params,
	fluke_alias_t		*out_port_alias,
        security_id_t		*out_client_sid)
{
	unsigned eax;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_setup_);
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_wait_receive_secure_);
#endif

	asm volatile("
		pushl	%%ebp
		movl	%%esp,%%ecx
		movl	$1f,%%edx
		movl	$2f,%%esi
		movl	$3f,%%edi
		call	_fluke_ipc_setup
		movzwl	2(%%ebx),%%eax
		movl	20(%%ebx),%%ecx
		movl	24(%%ebx),%%edi
		movl	28(%%ebx),%%esi
		movl	32(%%ebx),%%edx
		call	_fluke_ipc_wait_receive_secure
		jmp	3f
	1:	orl	%6,%%eax
		jmp	3f
	2:	orl	%7,%%eax
	3:	popl	%%ebp
	" : "=a" (eax),
	    "=b" (*out_port_alias),
	    "=c" (ipc_params->recv_buf.words),
	    "=S" (*out_client_sid)
	  : "a" (server_pset),
	    "b" (ipc_params),
	    "i" (FLUKE_IPC_WAIT_ONEWAY << 16),
	    "i" (FLUKE_IPC_WAIT_IDEMPOTENT << 16),
	    /*
	     * XXX
	     * We have to add a garbage in-value to edx because the compiler
	     * was not recognizing that edx was being trashed.
	     * This adds a movl instruction, but at least it works right.
	     */
	    "d" (0xf1ac)
	  : "edx", "edi", "memory");

	/* sets both recv_buf_count and recv_ref_count */
	*(unsigned short*)&ipc_params->recv_buf_count = eax;
	return (unsigned char)(eax >> 16);
}
#endif

FLUKE_ENTRY_INLINE fluke_error_t fluke_ipc_wait_receive(
	struct fluke_ipc_params	*ipc_params,
	fluke_alias_t		*out_port_alias)
{
	unsigned eax, ecx;
	unsigned char rc;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_wait_receive_);
#endif
	asm volatile("
		pushl	%%ebp
		call	_fluke_ipc_wait_receive
		popl	%%ebp
	" : "=a" (eax),
	    "=b" (*out_port_alias),
	    "=c" (ecx),
	    "=S" (((fluke_ipc_word_t*)ipc_params->recv_buf.buf)[0]),
	    "=D" (((fluke_ipc_word_t*)ipc_params->recv_buf.buf)[1])
	  : "a" (*(unsigned short*)&ipc_params->recv_buf_count),
		     /* eax gets both recv_buf_count and recv_ref_count */
	    "c" (ipc_params->recv_buf.words),
	    "d" (ipc_params->recv_ref_tab),
	    "S" (ipc_params->recv_buf_tab),
	    "D" (ipc_params->recv_buf.buf)
	  : "edx", "memory");
	rc = eax >> 16;
	/* sets both recv_buf_count and recv_ref_count */
	*(unsigned short*)&ipc_params->recv_buf_count = eax;
	ipc_params->recv_buf.words = ecx;
	return rc;
}

#if FLASK
FLUKE_ENTRY_INLINE fluke_error_t fluke_ipc_wait_receive_secure(
	struct fluke_ipc_params	*ipc_params,
	fluke_alias_t		*out_port_alias,
        security_id_t		*out_client_sid)
{
	unsigned eax, ecx;
	unsigned char rc;

#ifdef GPROF
	_mcount(get_eip(), (oskit_addr_t)_fluke_ipc_wait_receive_secure_);
#endif

	asm volatile("
		pushl	%%ebp
		call	_fluke_ipc_wait_receive_secure
		popl	%%ebp
	" : "=a" (eax),
	    "=b" (*out_port_alias),
	    "=c" (ecx),
	    "=S" (*out_client_sid)
	  : "a" (*(unsigned short*)&ipc_params->recv_buf_count),
		     /* eax gets both buf_count and ref_count */
	    "c" (ipc_params->recv_buf.words),
	    "d" (ipc_params->recv_ref_tab),
	    "S" (ipc_params->recv_buf_tab),
	    "D" (ipc_params->recv_buf.buf)
	  : "edx", "memory");
	rc = eax >> 16;
	/* sets both recv_buf_count and recv_ref_count */
	*(unsigned short*)&ipc_params->recv_buf_count = eax;
	ipc_params->recv_buf.words = ecx;
	return rc;
}
#endif

#endif /* __GNUC__ */

#endif /* _FLUKE_X86_IPC_H_ */
