/*
 * Copyright (c) 1996-1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include "s_thread.h"
#include "preempt.h"
#include "mutex.h"
#include "wait.h"
#include "cond.h"
#include "debug.h"
#include "config_instrument.h"
#include "s_instrument.h"

void cond_init(cond_t *c)
{
	wait_queue_init(&c->queue);
}

void cond_destroy(cond_t *c)
{
	cond_broadcast(c);
	wait_queue_destroy(&c->queue);
}

#include <fluke/thread.h>
int cond_wait(cond_t *c, mutex_t *m, link_t *inheritor_link)
{
	s_thread_t *cur_th = current_thread();

	assert_preemption_enabled();

	disable_preemption();
	spin_lock(&cur_th->wait_state.lock);
	assert(cur_th->wait_state.val == WAIT_NONE);

	/*
	 * If a cancel was already pending,
	 * return an error immediately without switching threads.
	 *
	 * If it is a thread-interrupt, and interrupts are
	 * currently disabled, then ignore the cancel.
	 * Otherwise a thread can't make system calls from inside
	 * the interrupt handler.
	 */
	if ((cur_th->wait_state.cancel_pending) &&
	    (cur_th->th_lflags & FLUKE_THREAD_EXCEPTION_ENABLE)) {
		/*
		 * Release the mutex.  It will be reacquired when
		 * we retry.
		 */
		pd_mutex_unlock(cur_th, m);
		spin_unlock(&cur_th->wait_state.lock);
		enable_preemption();
		INSTRUMENT_EVENT(cur_th, KR_CANCEL, 0);
		return KR_CANCEL;
	}

	/* By default, when we wake up, just restart from user space. */
	cur_th->wait_state.resume_rc = KR_RESTART;

	cur_th->wait_state.val = WAIT_ON_COND;
	cur_th->wait_state.waiting_on = c;
	wait_queue_add(&c->queue, cur_th);

	/* We can release the mutex now that we're on the queue. */
	pd_mutex_unlock(cur_th, m);

	cur_th = pd_thread_dispatch(cur_th, &cur_th->wait_state.lock,
				    0, inheritor_link);
	enable_preemption();
	INSTRUMENT_EVENT(cur_th, cur_th->wait_state.resume_rc, 0);
	return cur_th->wait_state.resume_rc;
}

#if 0
#ifdef CONFIG_INLINING
/* this is a critical path, so it helps to inline this into cond_wait_other */
#include "config_user_sched.h"
#ifndef CONFIG_USER_SCHED
static
#endif
inline
#endif
#endif
void pd_cond_wait_other_locked(cond_t *c, s_thread_t *thread)
{
	assert(thread->wait_state.val == WAIT_NONE);
	if (thread->wait_state.cancel_pending) {
		thread->wait_state.resume_rc = KR_CANCEL;
		thread_ready(thread);
	} else {
		thread->wait_state.resume_rc = KR_RESTART;
		thread->wait_state.val = WAIT_ON_COND;
		thread->wait_state.waiting_on = c;
		wait_queue_add(&c->queue, thread);
	}
}


void cond_wait_other(cond_t *c, s_thread_t *thread)
{
	disable_preemption();
	spin_lock(&thread->wait_state.lock);
	pd_cond_wait_other_locked(c, thread);
	spin_unlock(&thread->wait_state.lock);
	enable_preemption();
}

s_thread_t *pd_cond_wakeup(
	cond_t			*c,
	wait_val_t		*out_waitval)
{
	s_thread_t *th;

	th = wait_queue_remove(&c->queue);
	if (th != 0)
	{
		/*
		 * Clear the thread's wait state.
		 * Note that someone else may try to cancel the thread here,
		 * after we've removed it from the queue
		 * but before we have a chance to clear the wait state.
		 * In that case, they'll just notice that it's not on the queue
		 * and do nothing, knowing that we will (see thread_cancel).
		 */
		spin_lock(&th->wait_state.lock);
		assert((th->wait_state.val & WAIT_QUEUED_ON_COND)
		       == WAIT_QUEUED_ON_COND);
		assert(th->wait_state.waiting_on == c);
		if (out_waitval)
			*out_waitval = th->wait_state.val;
		th->wait_state.val = WAIT_NONE;
		do_debug(th->wait_state.waiting_on = 0);
		spin_unlock(&th->wait_state.lock);
	}

	return th;
}

s_thread_t *cond_wakeup(
	cond_t			*c)
{
	s_thread_t *th;

	disable_preemption();
	th = pd_cond_wakeup(c, 0);
	enable_preemption();

	return th;
}

void cond_signal(
	cond_t			*c)
{
	s_thread_t *th;

	disable_preemption();

	if ((th = pd_cond_wakeup(c, 0)) != 0)
		thread_ready(th);

	enable_preemption();
}

void cond_broadcast(
	cond_t			*c)
{
	s_thread_t *th;

	disable_preemption();

	while ((th = pd_cond_wakeup(c, 0)) != 0)
		thread_ready(th);

	enable_preemption();
}
