/*
 * Copyright (c) 1996, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * POSIX threads-like condition variable implementation
 * for use by higher-level code in the Fluke kernel.
 */
#ifndef _FLUKE_KERNEL_COND_H_
#define _FLUKE_KERNEL_COND_H_

#include "dispatch.h"
#include "wait_queue.h"
#include "wait.h"

struct mutex;
struct link;

struct cond
{
	struct wait_queue	queue;
};
typedef struct cond cond_t;


void cond_init(struct cond *c);

/*
 * Destroy a condition variable.
 * Any threads waiting on the condition variable will be woken.
 */
void cond_destroy(struct cond *c);

/*
 * Wait on the specified condition variable,
 * atomically unlocking the associated mutex
 * and donating the CPU time to 'inheritor_link' (if non-NULL).
 * Consumes the link.
 * Does _not_ re-lock the mutex before returning (unlike POSIX conds).
 */
int cond_wait(
	struct cond		*c,
	struct mutex		*m,
	struct link		*inheritor_link);

/*
 * Place another (captured) thread asleep on a condition variable.
 */
void cond_wait_other(struct cond *c, struct s_thread *thread);

/*
 * Same, but preemption is disabled and thread->wait_state.lock
 * is held throughout.
 */
void pd_cond_wait_other_locked(cond_t *c, struct s_thread *thread);

/*
 * Wakeup a thread on a condition variable but hold it hostage.
 * The caller must call thread_ready() on it later to release it.
 * Returns NULL if no threads are waiting on the condition.
 */
struct s_thread *cond_wakeup(struct cond *c);

/*
 * Same as cond_wakeup, but preemption is disabled.
 * If out_waitval is non-null, *out_waitval gets the
 * wait state the returned thread was in before wakeup.
 */
struct s_thread *pd_cond_wakeup(
	struct cond		*c,
	wait_val_t		*out_waitval);


/*
 * Signal a condition variable,
 * waking up at least one thread if any threads are waiting on it.
 */
void cond_signal(struct cond *c);

/*
 * Wake up all threads that are waiting on a condition variable.
 */
void cond_broadcast(struct cond *c);


#endif /* _FLUKE_KERNEL_COND_H_ */
