/*
 * Copyright (c) 1996, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * This file declares kernel return codes used within the kernel
 * to flag exceptional conditions that require interrupting a kernel operation.
 * These return codes generally (though not always)
 * cause the kernel operation in progress to be rolled back
 * all the way to the entrypoint from user mode,
 * at which point the kernel entry code takes some appropriate action.
 * The exception_state stored in a particular thread
 * is sort of the internal Fluke kernel equivalent of 'errno';
 * it provides supplemental information about exceptions that occur.
 *
 * Note that some routines (particularly IPC routines)
 * may return a fluke_error_t status code instead of a KR return code.
 * Such a status code will be passed back to the user verbatim
 * instead of causing a special action at the kernel/user interface layer.
 */
#ifndef _FLUKE_KERNEL_EXCEPTION_H_
#define _FLUKE_KERNEL_EXCEPTION_H_

#include <fluke/vm.h>

#include "config_nonblocking.h"

/*
 * Minimum kernel-internal return code value.
 * For kernel routines that may return
 * either a kernel-internal return code or a user-visible return code,
 * any value greater than or equal to this is a kernel-internal code.
 */
#define KR_MIN			0xf000

/*
 * A processor exception occurred which should be blamed on the user
 * (e.g. because the exception was generated while accessing user space).
 * The current thread's exception_state contains the details.
 */
#define KR_USER_EXCEPTION	0xf000

/*
 * Page fault occurred.
 * This gets turned into a real KR_USER_EXCEPTION by the kentry layer
 * if the page fault cannot be resolved in the kernel
 * and no appropriate region keeper can be found to handle the fault
 * (or if a region keeper was called but returned an error).
 */
#define KR_PAGE_FAULT		0xf100

/*
 * Another thread is trying to manipulate us and has asynchronously canceled us,
 * e.g. due to thread_interrupt(), thread_get_state(), thread_set_state().
 */
#define KR_CANCEL		0xf200

/*
 * Ran out of kernel memory while trying to do something.
 */
#define KR_NO_MEMORY		0xf300

/*
 * This return code indicates that we have context switched due to a wait,
 * and we need to restart execution in user mode before doing anything else,
 * since in a nonblocking (stack-per-processor) kernel
 * any variables still on the kernel stack will still refer to the old thread.
 */
#define KR_RESTART		0xf400

#if FLASK
/*
 * Security fault occurred.
 * An upcall to the security policy server is required.
 */
#define KR_SECURITY_FAULT	0xf500
#endif


#ifdef DEBUG
/*
 * Some handy asserts for double checking rc values
 */
#include "debug.h"
#define assert_rc_is(rc, isWhat)					\
	if (rc != isWhat) {						\
		panic("%s:%d assert_rc_%s() failed; rc is %#x\n",	\
		      __FILE__, __LINE__, #isWhat, rc);			\
	}
#define assert_rc_0(rc)          assert_rc_is(rc, 0)
#define assert_rc_KR_RESTART(rc) assert_rc_is(rc, KR_RESTART)
#define assert_rc_KR_CANCEL(rc)  assert_rc_is(rc, KR_CANCEL)
#else /* not DEBUG */
#define assert_rc_is(x, y)      ((void)0)
#define assert_rc_0(x)	   	((void)0)
#define assert_rc_KR_RESTART(x)	((void)0)
#define assert_rc_KR_CANCEL(x)	((void)0)
#endif DEBUG

struct s_thread;

/*
 * Set up the current thread's exception state
 * to generate an insanity exception with the specified subcode.
 * The caller should then return KR_USER_EXCEPTION
 * to cause the exception to be taken by the indicated thread.
 */
void exc_set_insanity(struct s_thread *th, int code);

/*
 * Set up the current thread's exception state
 * to reflect a page fault.
 */
void exc_set_pagefault(struct s_thread *th, int fva, fluke_prot_t prot);

/*
 * Convenience macro to flag an insanity condition in the current thread
 * and then return from the current function with an error code.
 */
#define EXC_RETURN_INSANITY(c)					\
({								\
	exc_set_insanity(current_thread(), FLUKE_INSANITY_##c);	\
	return KR_USER_EXCEPTION;				\
})

struct fluke_thread_state;

void exc_state_dump(struct exc_state *st);

void thread_exc_state_init(struct s_thread *th);
void thread_exc_get_state(struct s_thread *th,
			  struct fluke_thread_state *state);
void thread_exc_set_state(struct s_thread *th,
			  struct fluke_thread_state *state);

#endif /* _FLUKE_KERNEL_EXCEPTION_H_ */
