/*
 * Copyright (c) 1998, 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * XXX this is a copy of the generic libnest/server/com/clock.c adapter,
 * hacked up for the in-kernel clock.  The reason for this is the need
 * to receive the timer callbacks at interrupt level in kernel mode.
 */

#include "kserver.h"
#include "s_cond.h"
#include "s_task.h"
#include "s_thread.h"
#include "prog.h"		/* do_in_supervisor_mode */
#include "preempt.h"
#include "exception.h"
#include "debug.h"

#include <fluke/nest/server/com.h>
#include <oskit/dev/clock.h>
#include <oskit/dev/timer.h>
#include <oskit/fs/openfile.h>
#include <oskit/fs/file.h>
#include <oskit/com/listener.h>
#if FLASK
#include <fluke/nest/server/com-clock_secure.h>
#include <fluke/nest/server/com-timer_secure.h>
#else
#include <fluke/nest/server/com-clock.h>
#include <fluke/nest/server/com-timer.h>
#endif

#include <pthread.h>
#include <fluke/sac/listener.h>



/*
 * This struct describes the MOM object wrapping a oskit_timer COM object.
 * The first three members are those defined by COMSTRUCT for all
 * MOM->COM wrapper objects (see <fluke/nest/server/com.h>).
 */
struct nest_com_Timer_struct {
	mom_obj_header_t header;
	oskit_timer_t *com;
	struct nest_com_lock_data lock_data;

	/*** Extra data portion follows.  ***/

	fluke_cond_t wakeup;
	fluke_mutex_t lock;
	pthread_t thread;
	int pending;		/* XXX word for atomic store */
	s_cond_t *s_wakeup;	/* XXX */

	/*
	 * This is the listener callback object we pass to
	 * `oskit_timer_setlistener'.  The ref is released
	 * in the listener_release method (below); we Know
	 * that releasing the last ref to the timer COM object
	 * makes it release our listener COM object.
	 */
	oskit_listener_t li;
	unsigned int li_refs;	/* ref count for COM listener object */
	mom_ref_t listener;	/* where callback gets sent */
};
#define EXTRABYTES(Timer) (sizeof(struct nest_com_##Timer##_struct) - \
			   sizeof(struct nest_com_obj_struct))


void
nest_com_Timer_setlistener(nest_com_Timer_t *_obj,
			   mom_ref_t listener,
#if FLASK
			   const security_id_t _csid,
			   FLASK_SSID_SARG
#endif
			   mom_exc_t *_ev)
{
#if FLASK
	CCONSET(_csid,_ev,);
#endif

	LOCK;
	if (!mom_ref_equiv(listener, _obj->listener)) {
		if (_obj->listener == MOM_REF_NULL) {
			int rc;
			static void *do_callback(void *);
			/*
			 * We had no listener before, so we need to
			 * install our the COM listener object to get
			 * notifications.
			 */
			oskit_timer_setlistener(_obj->com, &_obj->li);

			/*
			 * XXX must start thread eagerly to keep
			 * interrupt-level code simple.
			 * Start up a thread to do the callback,
			 * which might block arbitrarily on the client.
			 */
			rc = fluke_cond_create(&_obj->wakeup);
			assert(rc == 0);
			rc = fluke_mutex_create(&_obj->lock);
			assert(rc == 0);
			rc = pthread_create(&_obj->thread, NULL,
					    &do_callback, _obj);
			assert(rc == 0);
			{ /* XXX kernel-mode pointer to the cond */
				int foo(void *arg) {
					nest_com_Timer_t *const t = arg;
					int rc = task_ob_lookup_type
						(current_thread(),
						 (oskit_addr_t) &t->wakeup,
						 FLUKE_PROT_WRITE,
						 FLUKE_COND,
						 (s_ob_t **)&t->s_wakeup);
					assert_rc_0(rc);
					s_cond_unlock(t->s_wakeup);
					return rc;
				}
				do_in_supervisor_mode(&foo, _obj);
			}
		}
		else if (listener == MOM_REF_NULL) {
			/*
			 * The listener is being cleared, so we will
			 * have no place to send our `notify' RPCs.
			 * Deinstall our COM listener object so there
			 * are no superfluous wakeups with nothing to do.
			 */
			oskit_timer_setlistener(_obj->com, NULL);
			/*
			 * That released the old _obj->listener.
			 */
			assert(_obj->listener == MOM_REF_NULL);
		}
		mom_ref_acquire(listener);
		assert(_obj->listener == MOM_REF_NULL);
		_obj->listener = listener;
	}
	UNLOCK;
}

void
nest_com_Timer_gettime(nest_com_Timer_t *_obj,
		       fluke_itimerspec_t *value,
#if FLASK
		       const security_id_t _csid,
		       FLASK_SSID_SARG
#endif
		       mom_exc_t *_ev)
{
#if FLASK
	CCONSET(_csid,_ev,);
#endif

	COMCALL_V(timer_gettime, value);
}

void
nest_com_Timer_settime(nest_com_Timer_t *_obj,
		       const oskit_bool_t absolute_time,
		       const fluke_itimerspec_t *value,
#if FLASK
		       const security_id_t _csid,
		       FLASK_SSID_SARG
#endif
		       mom_exc_t *_ev)
{
#if FLASK
	CCONSET(_csid,_ev,);
#endif

	COMCALL(timer_settime, absolute_time, value);
}

oskit_u32_t
nest_com_Timer_getoverrun(nest_com_Timer_t *_obj,
#if FLASK
			  const security_id_t _csid,
			  FLASK_SSID_SARG
#endif
			  mom_exc_t *_ev)
{
#if FLASK
	CCONSET(_csid,_ev,0);
#endif

	return COMCALL(timer_getoverrun);
}



#define listener2timer(l) \
	((nest_com_Timer_t *) ((char *) (l) - offsetof(nest_com_Timer_t, li)))

static OSKIT_COMDECL
listener_query(oskit_listener_t *io, const oskit_iid_t *iid,
	       void **out_ihandle)
{
        nest_com_Timer_t *t = listener2timer(io);

        if (memcmp(iid, &oskit_iunknown_iid, sizeof(*iid)) == 0 ||
            memcmp(iid, &oskit_listener_iid, sizeof(*iid)) == 0) {
                *out_ihandle = &t->li;
                ++t->li_refs;
                return 0;
        }

        *out_ihandle = NULL;
        return OSKIT_E_NOINTERFACE;
}

static OSKIT_COMDECL_U
listener_addref(oskit_listener_t *io)
{
        nest_com_Timer_t *t = listener2timer(io);

        return ++t->li_refs;
}

static OSKIT_COMDECL_U
listener_release(oskit_listener_t *io)
{
        nest_com_Timer_t *t = listener2timer(io);
        unsigned newcount;

        newcount = --t->li_refs;
        if (newcount == 0) {
		if (t->thread) {
			mom_ref_release(t->listener);
			t->listener = MOM_REF_NULL;
		}
		else {
			/*
			 * There is a callback thread sitting around,
			 * so wake it up and tell it to die.
			 * XXX should be just pthread_cancel+pthread_join
			 */
			pthread_t thread = t->thread;
			fluke_mutex_lock(&t->lock);
			t->pending = 1;
			t->thread = 0;
			mom_ref_release(t->listener);
			t->listener = MOM_REF_NULL;
			fluke_cond_signal(&t->wakeup);
			fluke_mutex_unlock(&t->lock);
			/*
			 * The callback thread will now wake up,
			 * destroy the cond and mutex, and return.
			 * We must wait for it to complete before
			 * we can return and allow the timer object
			 * to potentially die.
			 */
			pthread_join(thread, (void **) &thread);
		}
        }

        return newcount;
}

static void *
do_callback(void *arg)
{
	nest_com_Timer_t *t = arg;

	while (1) {
		mom_exc_t exc;
		mom_ref_t listener;

		fluke_mutex_lock(&t->lock);
		while (!t->pending)
			fluke_cond_wait(&t->wakeup, &t->lock);

		if (!t->thread) { /* Time to die.  */
			fluke_cond_destroy(&t->wakeup);
			fluke_mutex_destroy(&t->lock);
			return 0;
		}

		t->pending = 0;
		listener = t->listener;
		mom_ref_acquire(listener);
		fluke_mutex_unlock(&t->lock);

		cp_listener_notify(listener, &exc);
		mom_ref_release(listener);
		if (exc._type == exc_mom_system_exception)
			switch ((int) exc._except) {
			case EBADF: /* listener died */
				nest_com_lock(&t->lock_data);
				oskit_timer_setlistener(t->com, NULL);
				nest_com_unlock(&t->lock_data);
				break;
			default:
				/*
				 * Ignore other transport errors.
				 */
				break;
			}
		else
			assert(exc._type == exc_mom_no_exception);
	}
}

/* ARGSUSED */
static OSKIT_COMDECL
listener_notify(oskit_listener_t *io, oskit_iunknown_t *obj)
{
        nest_com_Timer_t *const t = listener2timer(io);


	/*
	 * Now we just wake up Bubba.
	 */

	t->pending = 1;		/* atomic store */
	assert_preemption_enabled();
	s_cond_lock(t->s_wakeup);
	s_cond_signal(t->s_wakeup);
	s_cond_unlock(t->s_wakeup);

	return 0;
}

static struct oskit_listener_ops listener_ops = {
	listener_query,
	listener_addref,
	listener_release,
	listener_notify
};



COMSTRUCT(Clock, clock);

mom_class_t nest_com_clock_class, nest_com_timer_class;
mom_class_t nest_com_file_clock_class, nest_com_openfile_clock_class;


int
nest_com_clock_init()
{
	COMCLASSINIT(Clock, clock);
	COMCLASSINIT(Timer, timer);
	COMCLASSINIT(ClockDeviceFile, file_clock);
	COMCLASSINIT(OpenClockDeviceFile, openfile_clock);

	return 0;
}


static inline oskit_clock_t *
coerce_clock(void *arg, mom_exc_t *_ev)
{
	oskit_iunknown_t *const obj = arg;
	oskit_error_t rc;
	oskit_clock_t *clock;
	oskit_openfile_t *of;

	rc = oskit_iunknown_query(obj, &oskit_clock_iid, (void **) &clock);
	if (! rc)
		return clock;
	rc = oskit_iunknown_query(obj, &oskit_openfile_iid, (void **) &of);
	if (! rc) {
		oskit_file_t *f;
		rc = oskit_openfile_getfile(of, &f);
		oskit_openfile_release(of);
		if (! rc) {
			rc = oskit_file_query(f, &oskit_clock_iid,
					     (void **) &clock);
			oskit_file_release(f);
			if (! rc)
				return clock;
		}
	}

	err2exc(rc, _ev);
	return NULL;
}


#define CLOCKCALL_1(decl, rc, finish, call, args...) {			      \
	oskit_clock_t *clock;						      \
	LOCK;								      \
	clock = coerce_clock(_obj->com, _ev);		      		      \
	if (clock) {							      \
		decl;					      		      \
		rc oskit_clock_##call(clock ,##args);			      \
		oskit_clock_release(clock);				      \
		UNLOCK;							      \
		finish;							      \
	}								      \
}
#define CLOCKCALL(call, args...) CLOCKCALL_1(oskit_error_t rc,		      \
					     rc =, err2exc(rc, _ev),	      \
					     call ,##args)
#define CLOCKCALL_V(call, args...) CLOCKCALL_1(,,, call ,##args)

void
nest_com_Clock_gettime(nest_com_Clock_t *_obj,
		       fluke_timespec_t *out_time,
#if FLASK
		       const security_id_t _csid,
		       FLASK_SSID_SARG
#endif
		       mom_exc_t *_ev)
{
#if FLASK
	CCONSET(_csid,_ev,);
#endif

	CLOCKCALL_V(gettime, out_time);
}


void
nest_com_Clock_getclockinfo(nest_com_Clock_t *_obj,
			    fluke_Clock_info_t *info,
#if FLASK
			    const security_id_t _csid,
			    FLASK_SSID_SARG
#endif
			    mom_exc_t *_ev)
{
#if FLASK
	CCONSET(_csid,_ev,);
#endif

	CLOCKCALL_V(getclockinfo, info);
}


void
nest_com_Clock_sleep(nest_com_Clock_t *_obj,
		     const fluke_timespec_t *duration,
#if FLASK
		     const security_id_t _csid,
		     FLASK_SSID_SARG
#endif
		     mom_exc_t *_ev)
{
#if FLASK
	CCONSET(_csid,_ev,);
#endif

	CLOCKCALL_V(sleep, duration);
}

void
nest_com_Clock_spin(nest_com_Clock_t *_obj,
		    const fluke_timespec_t *duration,
#if FLASK
		    const security_id_t _csid,
		    FLASK_SSID_SARG
#endif
		    mom_exc_t *_ev)
{
#if FLASK
	CCONSET(_csid,_ev,);
#endif

	CLOCKCALL_V(spin, duration);
}


void
nest_com_Clock_settime(nest_com_Clock_t *_obj,
		       const fluke_timespec_t *new_time,
#if FLASK
		       const security_id_t _csid,
		       FLASK_SSID_SARG
#endif
		       mom_exc_t *_ev)
{
#if FLASK
	CCONSET(_csid,_ev,);
#endif

	CLOCKCALL(settime, new_time);
}

void
nest_com_Clock_adjtime(nest_com_Clock_t *_obj,
		       const fluke_timespec_t *new_time,
#if FLASK
		       const security_id_t _csid,
		       FLASK_SSID_SARG
#endif
		       mom_exc_t *_ev)
{
#if FLASK
	CCONSET(_csid,_ev,);
#endif

	CLOCKCALL(adjtime, new_time);
}

void
nest_com_Clock_setperiod(nest_com_Clock_t *_obj,
			 const fluke_timespec_t *new_time,
#if FLASK
			 const security_id_t _csid,
			 FLASK_SSID_SARG
#endif
			 mom_exc_t *_ev)
{
#if FLASK
	CCONSET(_csid,_ev,);
#endif

	CLOCKCALL(setperiod, new_time);
}


void
nest_com_Clock_open(nest_com_Clock_t *_obj,
		    mom_ref_t *opendev,
#if FLASK
		    const security_id_t _csid,
		    FLASK_SSID_SARG
#endif
		    mom_exc_t *_ev)
{
	oskit_clock_t *new;

#if FLASK
	CCONSET(_csid,_ev,);
#endif

	CLOCKCALL_1(oskit_error_t rc, rc =,
		    {
			    if (rc)
				    err2exc(rc, _ev);
			    else
				    *opendev = nest_com_instantiate_ref(
					    nest_com_clock_class, new,
					    &_obj->lock_data, _ev);
		    },
			    open, &new);
}


mom_ref_t
nest_com_Clock_createtimer(nest_com_Clock_t *_obj,
#if FLASK
			   const security_id_t _csid,
			   FLASK_SSID_SARG
#endif
			   mom_exc_t *_ev)
{
	oskit_clock_t *clock;
	oskit_timer_t *timer;
	oskit_error_t rc;
	nest_com_Timer_t *new;
	mom_ref_t ret;

#if FLASK
	CCONSET(_csid,_ev,MOM_REF_NULL);
#endif

	LOCK;
	clock = coerce_clock(_obj->com, _ev);
	if (!clock) {
		UNLOCK;
		return NULL;
	}

	rc = oskit_clock_createtimer(clock, &timer);
	oskit_clock_release(clock);
	UNLOCK;
	if (rc) {
		err2exc(rc, _ev);
		return MOM_REF_NULL;
	}

	new = nest_com_instantiate_object(nest_com_timer_class,
					  timer, &_obj->lock_data,
					  EXTRABYTES(Timer), NULL);
	LOCK;
	oskit_timer_release(timer);
	UNLOCK;
	if (!new) {
	nomem:
		mom_set_exc(_ev, ENOMEM);
		return MOM_REF_NULL;
	}

	new->listener = MOM_REF_NULL;
	new->li.ops = &listener_ops;

	rc = mom_obj_reference(new, &ret);
	mom_obj_release_ref(new);
	if (rc)
		goto nomem;
	return ret;
}

#include <fluke/mutex.h>
#include <oskit/fs/soa.h>
#include "../fdev/fluke_fdev.h"
#include <fluke/sac/clock.h>
#include <fcntl.h>
#include <sys/time.h>

static fluke_mutex_t ks_clock_lock;

void
kserver_clock_init()
{
	int fd, rc;
	mom_ref_t node;
	oskit_clock_t *clock;
	oskit_file_t *clockfile;
	static const struct nest_com_lock_data fdevlock = {
		fdev_process_lock, fdev_process_unlock, NULL
	};
	fluke_timespec_t ts;
	mom_exc_t exc;

	nest_com_clock_init();

	rc = fluke_mutex_create(&ks_clock_lock);
	assert_rc_0(rc);

	fd = open("/dev/clock", O_RDWR|O_CREAT, 0644);
	assert(fd >= 0);
	rc = mom_ref_from_fd(fd, &node);
	assert_rc_0(rc);
	rc = close(fd);
	assert_rc_0(rc);

	clock = oskit_clock_init();
	assert(clock);
#if 0				/* XXX libfdev overrides oskit_rtc_get */
	rc = oskit_rtc_get((oskit_timespec_t *) &ts);
	assert_rc_0(rc);
	/*
	 * Convert local to univeral time.
	 */
	ts.tv_sec += 60*60* 6;	/* XXX Mountain Standard Time */
#else
	{
		extern struct timeval time;
		TIMEVAL_TO_TIMESPEC(&time, &ts);
	}
#endif
	rc = oskit_clock_settime(clock, (oskit_timespec_t *) &ts);
	assert_rc_0(rc);

	rc = oskit_soa_file_from_clock(clock, &clockfile);
	oskit_clock_release(clock);
	assert_rc_0(rc);

	mom_com_mount(node, clockfile, &fdevlock);
	mom_ref_release(node);

	fd = open("/dev/clock", O_RDONLY, 0);
	assert(fd >= 0);
	rc = mom_ref_from_fd(fd, &node);
	assert_rc_0(rc);
	rc = close(fd);
	assert_rc_0(rc);
	cp_clock_gettime(node, &ts, &exc);
	assert(exc._type == exc_mom_no_exception);
	mom_ref_release(node);

	printf("Kernel server clock reads %ld.%.9ld seconds since 1970\n",
	       (long int) ts.tv_sec, (long int) ts.tv_nsec);
}
