/*
 * Copyright (c) 1997-1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * This file contains the code for the kserver `heartbeat' thread,
 * which executes every clock tick for a brief period.
 */


#include "kserver/kserver.h"
#include "prog.h" /* XXX: do_in_supervisor_mode */
#include <oskit/dev/dev.h> /* osenv_timer_register */
#include <fluke/debug.h> /* fluke_debug_label_flob */

/*
 * Nearly all of this code executes in supervisor mode.
 */

#ifdef HEARTBEAT_THREAD

#include <pthread.h>
#include "preempt.h"
#include "dispatch.h"
#include "s_thread.h"

struct {
	spin_lock_t lock;
	int foo; /* XXX if spin locks are compiled out something must be left */
} hb_lock;

struct s_thread *captured_hbthread;
struct preempter heartbeat_preempter;
#ifdef PREEMPTSTATS
unsigned int hbruns, hbcycles, hbnotready, hbrunning, hbstarted;
unsigned int hbstartt, hbendt;
unsigned int hbqueued, hbqtime, hbdcycles;
s_thread_t *hbthread;
#endif

void
heartbeat_func(void)
{
	static int myticks;

	assert_preemption_disabled();

	if (captured_hbthread == NULL || myticks-- > 0) {
#ifdef PREEMPTSTATS
		if (hbstarted)
			hbrunning++;
		else
			hbnotready++;
#endif
		return;
	}
#if defined(PREEMPTSTATS) || defined(SLEEP_DEVICE)
	/* every clock tick */
	myticks = 0;
#else
	/* every second */
	myticks = 100;
#endif

	thread_ready(captured_hbthread);
	captured_hbthread = NULL;
}


int
do_heartbeat_start(void *arg)
{
	spin_lock_init(&hb_lock.lock);
#ifdef PREEMPTSTATS
	disable_preemption();
	hbthread = current_thread();
	asm volatile("rdtsc" : "=a" (hbstartt) :: "eax", "edx");
	hbstarted = 1;
	enable_preemption();
#else
	osenv_timer_register(heartbeat_func, 100);
#endif
	return 0;
}


int
do_heartbeat_wait(void *arg)
{
	struct s_thread *cur_thread = current_thread();

#if !defined(PREEMPTSTATS) && !defined(SLEEP_DEVICE)
	printf("%p: (thubump)\n", cur_thread);
#endif

	cur_thread->wait_state.resume_rc = KR_RESTART;

	disable_preemption();
	spin_lock(&hb_lock.lock);
	captured_hbthread = cur_thread;
	cur_thread = pd_thread_dispatch(cur_thread, &hb_lock.lock, 0, 0);
	enable_preemption();

#ifdef SLEEP_DEVICE
	/*
	 * make sure that beast actually returns to user mode
	 * and doesn't start over every time
	 */
	return 0;
#else
	return cur_thread->wait_state.resume_rc;
#endif
}

void *
heartbeat_thread(void *foo)
{
	printf("Heartbeat thread starting...\n");
#ifdef DEBUG
	fluke_debug_label_flob(fluke_thread_self(), "HEARTBEAT");
#endif

	do_in_supervisor_mode(do_heartbeat_start, NULL);
	while (1) {
		do_in_supervisor_mode(do_heartbeat_wait, NULL);
#ifdef SLEEP_DEVICE
		{
			extern fluke_cond_t *sleep_dev_cond;
			extern fluke_mutex_t *sleep_dev_mutex;
			/* only signal if the device is up */
			if (sleep_dev_cond) {
				fluke_mutex_lock(sleep_dev_mutex);
				fluke_cond_broadcast(sleep_dev_cond);
				fluke_mutex_unlock(sleep_dev_mutex);
			}
		}
#endif /* SLEEP_DEVICE */
	}

	return NULL; /* XXX */
}
#endif /* HEARTBEAT_THREAD */
