/*
 * Copyright (c) 1996, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include <oskit/anno.h>

#include "dispatch.h"
#include "intr.h"
#include "kalloc.h"
#include "pmap.h"
#include "s_ob.h"
#include "s_task.h"
#include "s_mapping.h"
#include "s_thread.h"
#include "kentry.h"
#include "clock.h"
#include "sched.h"

#include "prog.h"
#include "debug.h"
#include "config_smp.h"
#include "config_profiling.h"

#ifdef CONFIG_PROFILING
#include "profiling.h"
#endif

#if FLASK
#include "avc.h"
#endif 

extern char version[];
extern char build_host[];
extern char build_date[];
extern char config_id[];

#include <oskit/gdb.h>

/* XXX: base_multiboot_main() knows about profiling oskit kernels */
int enable_gprof = 0;
void base_gprof_init() {}

void smp_boot();
void interrupt_boot(void);

void main(int argc, char **argv, char **envp)
{
#if FLASK
        int rc;
#endif
	anno_init();

	/* Print identifying information. */
	printf("%s\nBuilt on %s on %s\nConfiguration: %s\n",
		version, build_host, build_date, config_id);

	/*
	 * Initialize all the layers that need initialization,
	 * in order from lowest layer to highest.
	 */
	csw_boot();
	disp_boot();
	interrupt_boot();
	kalloc_boot();
	sched_boot();
#ifdef CONFIG_SMP
	smp_boot();	/* XXX needs to go before paging is enabled */
#endif

	pmap_boot();
#ifdef CONFIG_KPREEMPT
	/*
	 * XXX mega layering problems.
	 * original: XXX needs to happen after sched_boot
	 * new: XXX needs to happen after SMP and pmap boot since
	 * preemption is per-cpu and identity of current CPU might be
	 * found by looking in the magic thread GS segment (x86) setup
	 * in pmap.
	 */
	preempt_boot();
#endif

	s_ob_boot();
	s_task_boot();
	s_mapping_boot();
	s_thread_boot();

	clock_boot();		/* machine dep stuff here too? */
#ifdef CONFIG_PROFILING
	/* XXX Don't know just where this should be in the layering
	 * structure.  In BSD this code is run just before initalizing
	 * the scheduler, mounting root, adding swap, starting up init
	 * and the pagedaemon, etc.
	 */
	profiling_boot();
#endif

#if FLASK
	/*
	 * Initialize the kernel access vector cache.
	 * This is shared by the kernel and kserver, and
	 * must be initialized prior to any permission
	 * checks by either.
	 */
	avc_init();

	/* reset of all migrated permissions */
	rc = avc_register_reset_callback(s_ob_reset);
	assert(rc == 0);

	/* revocation of migrated subject->subject permissions */
	rc = avc_register_revoke_callback(SECCLASS_SUBJECT,
					  SECSID_WILD,
					  SECCLASS_SUBJECT,
					  SECSID_WILD,
					  FLUKE_PROT_ALL | 
					  SUBJECT_SUBJECT__CONNECT |
					  SUBJECT_SUBJECT__SPECIFY_CLIENT |
					  SUBJECT_SUBJECT__SPECIFY_SERVER |
					  SUBJECT_SUBJECT__TASK_KEEPER |
					  SUBJECT_SUBJECT__MAP |
					  SUBJECT_SUBJECT__THREAD_SCHEDULER,
					  s_ob_revoke);
	assert(rc == 0);

	/* relinquishment of migrated subject->subject permissions */
        rc = avc_register_try_remove_callback(SECCLASS_SUBJECT,
                                          SECSID_WILD,
                                          SECCLASS_SUBJECT,
                                          SECSID_WILD,
                                          FLUKE_PROT_ALL |
                                          SUBJECT_SUBJECT__CONNECT |
                                          SUBJECT_SUBJECT__SPECIFY_CLIENT |
                                          SUBJECT_SUBJECT__SPECIFY_SERVER |
                                          SUBJECT_SUBJECT__TASK_KEEPER |
                                          SUBJECT_SUBJECT__MAP |
                                          SUBJECT_SUBJECT__THREAD_SCHEDULER,
                                          s_ob_try_remove);
        assert(rc == 0);
#endif 

	/*
	 * This sets up a thread that, when activated, will
	 * create an idle thread for each processor and
	 * start up any processors that need to be started.
	 * It also sets up the kserver stuff.
	 */
	prog_boot(argv, envp);	/* never returns... */
}
