/*
 * Copyright (c) 1996-1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Versions of a few malloc-like functions
 * for use by the user-mode root task code
 * (the initial program loader and the kernel server).
 * These functions are callable from either user or supervisor mode.
 */

#include <malloc.h>
#include <strings.h>

#include "kalloc.h"
#include "prog.h"		/* XXX for do_in_supervisor_mode */

#include "debug.h"

struct memalign_args
{
	size_t alignment;
	size_t size;
	lmm_flags_t flags;
	void **ptr;
};

#ifdef MEMSTATS
extern unsigned long mstats_ks_bytes, mstats_ks_maxbytes;
#endif

static int do_smemalign(void *args)
{
	struct memalign_args *a = args;
	unsigned shift;
	void **ptr = a->ptr;

	for (shift = 0; (1 << shift) < a->alignment; shift++);

	*ptr = kalloc_gen(a->size, shift, a->flags);

#ifdef MEMSTATS
	if (*ptr) {
		mstats_ks_bytes += a->size;
		if (mstats_ks_bytes > mstats_ks_maxbytes)
			mstats_ks_maxbytes = mstats_ks_bytes;
	}
#endif
	return 0;
}

void *smemalign(size_t alignment, size_t size)
{
        void *ptr = 0;
	struct memalign_args args = { alignment, size, 0, &ptr } ;

	do_in_supervisor_mode(do_smemalign, &args);
	return ptr;
}

void *smemalignf(size_t alignment, size_t size, lmm_flags_t flags)
{
        void *ptr = 0;
	struct memalign_args args = { alignment, size, flags, &ptr } ;

	do_in_supervisor_mode(do_smemalign, &args);
	return ptr;
}

void *smalloc(size_t size)
{
        void *ptr = 0;
	struct memalign_args args = { 0, size, 0, &ptr } ;

	do_in_supervisor_mode(do_smemalign, &args);
	return ptr;
}

void *smallocf(size_t size, lmm_flags_t flags)
{
        void *ptr = 0;
	struct memalign_args args = { 0, size, flags, &ptr } ;

	do_in_supervisor_mode(do_smemalign, &args);
	return ptr;
}


struct sfree_args
{
	void *mem;
	size_t size;
};

static int do_sfree(void *args)
{
	struct sfree_args *a = args;

	kfree(a->mem, a->size);
#ifdef MEMSTATS
	mstats_ks_bytes -= a->size;
#endif
	return 0;
}

void sfree(void *mem, size_t size)
{
	struct sfree_args args = { mem, size };

	do_in_supervisor_mode(do_sfree, &args);
}

/* XXX libnest's object.c uses these for MOM support
   fdev_mem_* uses smalloc
*/
/*
 * These routines can only be used with each other: can't mix
 * realloc and smalloc, for example.
 *
 * XXX: What we really need to do is make sure that the LMM
 * used by libc is empty, so any allocations using the wrong
 * memory allocators will fail.
 */
void
free(void *ptr)
{
	void *newptr;
	if (!ptr)
		return;

	newptr = ptr - sizeof(size_t);
	sfree(newptr, *(size_t *)newptr);
}

/*
 * This allocates memopy, but tracks the size in a word before
 * the start of the memory address passed back, for use by free()
 */
void *
mallocf(size_t size, lmm_flags_t flags)

{
	size_t newsize;
	void *ptr;
	newsize = size + sizeof(size_t);

	ptr = smallocf(newsize, flags);
	if (!ptr)
		return ptr;

	*(size_t *)ptr = newsize;
	return (ptr + sizeof(size_t));
}

/*
 * XXX This is needed by osenv_mem_alloc (oskit/fdev/mem.c) to link;
 * XXX hopefully it isn't really used by drivers we need right now.
 */
void *
memalignf(size_t align, size_t size, lmm_flags_t flags)
{
	assert(align <= 4);	/* XXX */
	return mallocf(size, flags);
}

void *
malloc(size_t size)
{
	return mallocf(size, 0);
}

void *
calloc(size_t num, size_t size)
{
	char *tmp;
	size_t new_size = num * size;

	tmp = malloc(new_size);
	if (tmp)
		bzero(tmp, new_size);

	return tmp;
}


#include <strings.h>	/* XXX bcopy */
void *
realloc(void *buf, oskit_size_t new_size)
{
	void *newbuf;
	void *tmp;

	/*
	 * Pretty dumb: allocate a new region, copy the data, and free
	 * the old memory allocated.
	 */
	newbuf = malloc(new_size);
	if (!buf)
		return newbuf;

	tmp = buf - sizeof(size_t);
	bcopy(buf, newbuf, *(size_t *)tmp);
	free(buf);
	return newbuf;
}
