/*
 * Copyright (c) 1996, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Default implementation of blocking mutexes
 * used in fully preemptible versions of the Fluke kernel.
 */

#include "s_thread.h"
#include "preempt.h"
#include "wait_queue.h"
#include "mutex.h"
#include "debug.h"

#ifdef CONFIG_KPREEMPT
void mutex_init(mutex_t *m)
{
	wait_queue_init(&m->queue);
	m->owner = 0;
}

void mutex_destroy(mutex_t *m)
{
	assert(m->owner == 0);
	wait_queue_destroy(&m->queue);
}

inline void pd_mutex_lock(
	struct s_thread		*cur_thread,
	mutex_t			*m)
{
	assert(cur_thread == current_thread());
	assert(m->owner != cur_thread);

	wait_queue_lock(&m->queue);
	if (m->owner == 0) {
		/* Grab the mutex. */
		m->owner = cur_thread;
		wait_queue_unlock(&m->queue);
	} else {
		/* Mutex is already locked - wait for someone to free it. */
		wait_queue_add_locked(&m->queue, cur_thread);
		pd_thread_dispatch(cur_thread, &m->queue.lock, m->owner, 0);
		assert(cur_thread == current_thread());
		assert(m->owner == cur_thread);
	}
}

#ifndef MD_MUTEX_LOCK
void mutex_lock(mutex_t *m)
{
	assert_preemption_enabled();

	disable_preemption();
	pd_mutex_lock(current_thread(), m);
	enable_preemption();
}
#endif /* MD_MUTEX_LOCK */

int mutex_trylock(mutex_t *m)
{
	int rc = 0;

	assert_preemption_enabled();

	disable_preemption();
	wait_queue_lock(&m->queue);

	if (m->owner == 0) {
		m->owner = current_thread();
		rc = 1;
	}

	wait_queue_unlock(&m->queue);
	enable_preemption();

	return rc;
}

inline void pd_mutex_unlock(
	struct s_thread		*cur_thread,
	mutex_t			*m)
{
	s_thread_t *next_owner;

	wait_queue_lock(&m->queue);

	assert(m->owner == cur_thread);
	next_owner = wait_queue_remove_locked(&m->queue);
	if (next_owner != 0) {
		/* Hand the mutex directly to the next thread. */
		m->owner = next_owner;
		thread_ready(next_owner);
	} else
		m->owner = 0;

	wait_queue_unlock(&m->queue);
}

#ifndef MD_MUTEX_UNLOCK
void mutex_unlock(mutex_t *m)
{
	assert_preemption_enabled();

	disable_preemption();
	pd_mutex_unlock(current_thread(), m);
	enable_preemption();
}
#endif /* MD_MUTEX_UNLOCK */
#endif /* CONFIG_KPREEMPT */
