/*
 * Copyright (c) 1997 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Simple priority queue.
 * Three atomic operations are supported:
 *   Add an element to a queue using priorities.
 *   Remove the first element of highest priority from the queue.
 *   Remove a specific element from a queue.
 * The user is responsable for locking and unlocking.
 * The data structre should be improved for efficiency
 *   (ex: tree, or binary search).
 */


#ifndef _FLUKE_KERNEL_PRI_QUEUE_H_
#define _FLUKE_KERNEL_PRI_QUEUE_H_

/* 
 * if NO_kernel is defined
 * it dosn't use fluke features as
 * spin_locks. You can use this generic
 * queue on any platform.
 */

#ifndef NO_KERNEL
#include "spin_lock.h"
#endif

typedef struct prio_queue_node {
	unsigned		prio;
	unsigned		prio2; /* Two use a double priority queue */
	struct prio_queue_node	*next;
} prio_queue_node_t;


typedef struct prio_queue {
#ifndef NO_KERNEL
	spin_lock_t		lock;
#endif
	prio_queue_node_t	*head;
	prio_queue_node_t	ghost;
} prio_queue_t;


/* Initialize a queue to the empty, unlocked state.  */
void prio_queue_init(prio_queue_t *queue);

/* Destroy queue. */
void prio_queue_destroy(prio_queue_t *queue);

#ifndef NO_KERNEL
/* Lock a priority queue.
   Caller must worry about locking order dependencies,
   and ensure that interrupts are disabled while any queue is locked.  */
#define prio_queue_lock(queue)		spin_lock(&(queue)->lock)

/* Unlock a priority queue.  */
#define prio_queue_unlock(queue)	spin_unlock(&(queue)->lock)
#endif

/* Enqueue an element */
void prio_queue_insert(prio_queue_t *q, prio_queue_node_t *elem,
		       unsigned prio);

/* Enqueue an element using a two priority scheme */
void prio_queue_insert_2p(prio_queue_t *q, prio_queue_node_t *elem,
			  unsigned prio, unsigned prio2);

/* Remove specified element from queue. If not found nothing is done */
void prio_queue_remove(prio_queue_t *q, prio_queue_node_t *elem);

/* Remove element from top of the queue whith smaller number as priority */
prio_queue_node_t *prio_queue_remove_head(prio_queue_t *q);

/* Remove element from top of the queue whith smaller number as priority */
prio_queue_node_t *prio_queue_remove_head_prio(prio_queue_t *q, unsigned prio);

/* Check if element in the queue */
int prio_queue_find(prio_queue_t *q, prio_queue_node_t *elem);

/* Check if empty */
int prio_queue_empty(prio_queue_t *q);

/* Return the first element of the queue */
prio_queue_node_t *prio_queue_head(prio_queue_t *q);

/* Execute the given function on every element of the queue */
void prio_queue_map(prio_queue_t *q, void (*f)(prio_queue_node_t *));

#endif /* _FLUKE_KERNEL_PRIO_QUEUE_H_ */








