/*
 * Copyright (c) 1997 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include "config_stride_sched.h"
#ifndef CONFIG_STRIDE_SCHED

#include "s_thread.h"
#include "sched.h"
#include "round_robin.h"
#include "debug.h"
#include <stddef.h>
#include "config_smp.h"
#include "prio_queue.h"


#ifdef _RR_DEB
#define RR_DEB(fmt, args...) printf("DEB_RR:" fmt , ##args );printf("\n")
#define RR_DEBUG(x) x
#else
#define RR_DEB(args...)
#define RR_DEBUG(x)
#endif

/*Lock that must be set when using this interface*/
spin_lock_t sched_lock;


/* Queue of clients */
static prio_queue_t queue;


/*
 * Debugging function
 */
#ifdef _RR_DEB
static void
sched_dump_thread(prio_queue_node_t *s)
{
	s_thread_t *t = SCHED_LINK_TO_THREAD(s);
	printf("\tthread=%p, prio=%u\n", t, t->disp_state.pri);
}
#endif

/*
 * Set and get thread state
 */
void
thread_sched_set_state(struct s_thread *target, fluke_thread_state_t *new_state)
{
	RR_DEB("thread_sched_set_state");
}

void
thread_sched_get_state(struct s_thread *target, fluke_thread_state_t *state)
{
	RR_DEB("thread_sched_get_state");
}

void
thread_sched_state_init(struct s_thread *th)
{
	RR_DEB("thread_sched_state_init");
	th->sched_state.pri = SCHED_PRIUSER;
}

void
thread_sched_state_destroy(struct s_thread *th)
{
	/* We don't have to do anything */
	RR_DEB("thread_sched_state_destroy");
}

void
thread_sched_canceled(struct s_thread *th)
{
	/* We don't have to do anything */
	RR_DEB("thread_sched_canceled");
}


/*
 * Glue functions for the dispatcher
 */

/* Boot time initialization */
void
sched_boot(void)
{
	RR_DEB("sched_boot");
	prio_queue_init(&queue);
	spin_lock_init(&sched_lock);
}

/*
 * Marks the specified thread as runnable and returns non-zero if that
 * thread is higher priority than the current thread. 
 */
int
sched_thread_ready(struct s_thread *t)
{
	prio_queue_insert(&queue, &t->sched_state.link, t->sched_state.pri);
	return (t->sched_state.pri < current_thread()->sched_state.pri);
}

/*
 * Returns non-zero if there is a thread with higher priority than the
 * current thread.  The current thread is put on the run queue in that case.
 */
int
sched_thread_yield(void)
{
	s_thread_t *cur_thread = current_thread();

	RR_DEB("sched_thread_yield");
	RR_DEBUG(prio_queue_map(&queue, sched_dump_thread));
	if (!prio_queue_empty(&queue)) {
		s_thread_t *t = SCHED_LINK_TO_THREAD(prio_queue_head(&queue));

		if (t->sched_state.pri <= cur_thread->sched_state.pri) {
			prio_queue_insert(&queue, &cur_thread->sched_state.link,
					  cur_thread->sched_state.pri);
			return 1;
		}
	}
	return 0;
}

/*
 * Function called from dispatch to get the new thread to run
 */
struct s_thread *
sched_thread_next(void)
{
	struct s_thread *t;
	RR_DEB("stride_thread_next");
	RR_DEBUG(prio_queue_map(&queue, sched_dump_thread););
	assert(!prio_queue_empty(&queue));
	t = SCHED_LINK_TO_THREAD(prio_queue_remove_head(&queue));
#if 0
	if (t->sched_state.pri == SCHED_PRIIDLE)
		panic("thread %p: empty ready queue in disp_yield\n", t);
#endif
	return t;
}

struct s_thread *
sched_thread_idle(void)
{
	struct s_thread *current;

	current = SCHED_LINK_TO_THREAD(
			prio_queue_remove_head_prio(&queue, SCHED_PRIIDLE));
	assert(current);
	return current;
}

int
sched_ready_queue_empty(void)
{
	RR_DEB("sched_ready_queue_empty");
	return prio_queue_empty(&queue);
}

void
sched_first_thread(struct s_thread *t)
{
	RR_DEB("sched_first_thread. thread=%p", t);
}

#endif
