/*
 * Copyright (c) 1996 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Fluke mapping object implementation.
 * Mappings make memory exported by regions accessible in a task.
 */

#include <fluke/ref.h>
#include <fluke/mapping.h>
#include <fluke/exception.h>
#include <oskit/page.h>
#include <oskit/x86/paging.h>

#include "s_ref.h"
#include "s_region.h"
#include "s_mapping.h"
#include "frag.h"
#include "s_task.h"
#include "s_thread.h"
#include "debug.h"

#if FLASK
#include "avc.h"
#endif 

/* XXX move these elsewhere */
#define min(a, b) ((a) <= (b) ? (a) : (b))
#define max(a, b) ((a) < (b) ? (b) : (a))

static fluke_mapping_t kernel_mapping;

int s_mapping_create_locked(oskit_addr_t uva, fluke_hash_t hash,
			    s_mapping_t **out_mapping);

/*
 * Create an initial mapping and single fragment for physical memory,
 * link it to the kernel task.
 */
void s_mapping_boot(void)
{
	fluke_mapping_state_t mstate;
	s_mapping_t *mapping;
	int rc;
	extern oskit_addr_t phys_mem_max;
	extern link_t kernel_task_link;

	rc = s_mapping_create((oskit_addr_t)&kernel_mapping, 0, &mapping);
	assert(rc == 0);
#ifdef DEBUG
	s_ob_label(&mapping->sob, "kernel mapping");
#endif

	mstate.region_off = 0;
	mstate.start_va = 0;
	mstate.size = round_superpage(phys_mem_max);
	mstate.prot = FLUKE_PROT_ALL;
	rc = s_mapping_set_state(mapping, &mstate, 0, &kernel_task_link);
	assert(rc == 0);

#if FLASK
	mapping_frag_boot(mapping, mstate.start_va, mstate.start_va, 
			  mstate.size, mstate.prot, 
			  SECINITSID_KERNEL);
#else
	mapping_frag_boot(mapping, mstate.start_va, mstate.start_va, 
			  mstate.size, mstate.prot);
#endif 
	s_ob_unlock(&mapping->sob);
}

int s_mapping_create(oskit_addr_t uva, fluke_hash_t hash, s_mapping_t **out_mapping)
{
	s_mapping_t *mapping;
	int err;

	err = task_ob_create(current_thread(), FLUKE_MAPPING, uva, hash,
			     (s_ob_t **)&mapping);
	if (err)
		return (err);

	link_init(&mapping->task);
	mapping->region_off = 0;
	mapping->start_va = 0;
	mapping->size = 0;
	mapping->prot = FLUKE_PROT_NONE;

	mutex_init(&mapping->flock);
	link_init(&mapping->region);

	mutex_init(&mapping->ulock);
	mapping->taskp = 0;
	mapping_frag_init(mapping);

	*out_mapping = mapping;
	return 0;
}


/*
 * Remove the indicated mapping.
 * Mapping object should be locked.
 *
 * This has the additional effect of invalidating any VtoP translations
 * covered by this mapping or descendent mappings.
 */
void s_mapping_destroy(s_mapping_t *mapping)
{
	s_region_t *region;
	s_task_t *task;

	assert(mapping && mapping->sob.active);
	assert(mapping->sob.type == FLUKE_MAPPING);
	assert_s_ob_locked(&mapping->sob);

	/*
	 * Remove mapping from the task list.
	 */
	task = (s_task_t *)link_consume(&mapping->task);
	if (task) {
		s_mapping_remove_task(mapping, task);
		s_task_unlock(task);
	}

	/*
	 * Now remove from the region list.
	 */
	s_mapping_f_lock(mapping);
	region = (s_region_t *)link_consume(&mapping->region);
	if (region) {
		s_region_u_lock(region);
		queue_remove(&region->mappings, mapping, s_mapping_t *, rlink);
		s_region_u_unlock(region);
		s_region_unlock(region);
	}
	s_mapping_f_unlock(mapping);

	s_ob_destroy(&mapping->sob);
}

/*
 * Return the state of a mapping.
 * Mapping object should be locked on call and will be locked on return.
 */
int s_mapping_get_state(s_mapping_t *mapping, fluke_mapping_state_t *state,
			link_t *region_link, link_t *task_link)
{
	assert(mapping && mapping->sob.active);
	assert(mapping->sob.type == FLUKE_MAPPING);
	assert_s_ob_locked(&mapping->sob);

	if (region_link) {
		s_mapping_f_lock(mapping);
		link_create_copy(&mapping->region, region_link);
		s_mapping_f_unlock(mapping);
	}

	if (task_link)
		link_create_copy(&mapping->task, task_link);

	if (state) {
		state->hash = s_ob_gethash(&mapping->sob);
		state->region_off = mapping->region_off;
		state->start_va = mapping->start_va;
		state->size = mapping->size;
		state->prot = mapping->prot;
	}

	return (0);
}

/*
 * Set the state of a mapping.
 * Mapping object must be locked and is returned locked.
 */
int s_mapping_set_state(s_mapping_t *mapping, fluke_mapping_state_t *state,
			link_t *region_link, link_t *task_link)
{
	s_region_t *region;
	s_task_t *task = 0;
	oskit_bool_t sametask, newrange, sameregion;
	oskit_bool_t flushed = FALSE;
	int err;

	assert(mapping && mapping->sob.active);
	assert(mapping->sob.type == FLUKE_MAPPING);
	assert_s_ob_locked(&mapping->sob);

	/*
	 * Trivial state checks.
	 * Address must be page aligned and size a page multiple and
	 * covered range must be within the AS bounds of a task.
	 */
	if (state) {
		if (!page_aligned(state->start_va) ||
		    trunc_page(state->size) != state->size)
			EXC_RETURN_INSANITY(NOT_ALIGNED);

		if (state->start_va < FLUKE_VM_MIN ||
		    state->start_va + state->size > FLUKE_VM_MAX)
			EXC_RETURN_INSANITY(INVALID_ADDRESS);
	}

	sametask = (task_link == 0 ||
		    link_compare(&mapping->task, task_link));
	newrange = (state && (mapping->start_va != state->start_va ||
			      mapping->size != state->size));
	sameregion = (region_link == 0 ||
		      link_compare(&mapping->region, region_link));

	/*
	 * Common case: existing mapping is growing forward.
	 * No need to flush any translations, just make sure the expanding
	 * mapping doesn't overlap with the following one.
	 * XXX is this common enough to be worth the trouble?
	 */
	if (sametask && newrange && mapping->start_va == state->start_va &&
	    mapping->size < state->size &&
	    sameregion && mapping->region_off == state->region_off &&
	    mapping->prot == state->prot) {
		task = (s_task_t *)link_follow(&mapping->task);
		if (task) {
			s_mapping_t *nmapping;

			nmapping = (s_mapping_t *)queue_next(&mapping->tlink);
			if (queue_end(&task->mappings, (queue_t)nmapping) ||
			    nmapping->start_va >= state->start_va+state->size) {
				s_task_unlock(task);
				mapping->size = state->size;
				return (0);
			}
			s_task_unlock(task);
		}
	}
   
	/*
	 * If mapping is changing tasks or has changed locations, remove it
	 * from its old list now as it will need to be inserted into the
	 * new task (or the old task at a new location) later.
	 *
	 * N.B. we do not have to hold the fault or unmap lock when changing
	 * the task link since neither path accesses it.
	 */
	if (!sametask || newrange) {
		task = (s_task_t *)link_consume(&mapping->task);
		link_init(&mapping->task);
		if (task)
			s_mapping_remove_task(mapping, task);
		/*
		 * In either case (no former task or just flushed from task)
		 * we don't have to flush mapping later.
		 */
		flushed = TRUE;
	}

	/*
	 * If we are moving it in the same task, check for overlap and
	 * reinsert in map while we still have the task locked.
	 */
	if (sametask && task && newrange) {
		oskit_addr_t ova;
		oskit_size_t osz;

		/*
		 * XXX Note that the API doesn't require us to leave the
		 * mapping untouched if the insert fails, but we make an
		 * attempt.  We preserve the old state of the mapping but
		 * leave the mapping detached from any task.  It is left
		 * detached to be consistant with the new task case below.
		 */
		ova = mapping->start_va;
		osz = mapping->size;
		s_mapping_f_lock(mapping);
		s_mapping_u_lock(mapping);
		mapping->start_va = state->start_va;
		mapping->size = state->size;

		err = s_mapping_insert_task(mapping, task);
		if (err) {
			mapping->start_va = ova;
			mapping->size = osz;
			s_mapping_u_unlock(mapping);
			s_mapping_f_unlock(mapping);
			s_task_unlock(task);
			return (err);
		}
		s_mapping_u_unlock(mapping);
		s_mapping_f_unlock(mapping);
		link_produce(&mapping->task, &task->sob);
	}

	/*
	 * If mapping is going into a new task, drop the old task lock
	 * and acquire the new task lock.
	 */
	else if (!sametask) {
		if (task)
			s_task_unlock(task);
		task = (s_task_t *)link_follow(task_link);
		if (task) {
			oskit_addr_t ova;
			oskit_size_t osz;

			/*
			 * XXX Again we preserve the state of the mapping but
			 * don't reinsert it on the old task list.  Reinserting
			 * the mapping might fail (since the old task was
			 * unlocked and may have been destroyed or another
			 * mapping inserted into it) so we don't even try.
			 */
			ova = mapping->start_va;
			osz = mapping->size;
			s_mapping_f_lock(mapping);
			s_mapping_u_lock(mapping);
			if (state) {
				mapping->start_va = state->start_va;
				mapping->size = state->size;
			}
			err = s_mapping_insert_task(mapping, task);
			if (err) {
				if (state) {
					mapping->start_va = ova;
					mapping->size = osz;
				}
				s_mapping_u_unlock(mapping);
				s_mapping_f_unlock(mapping);
				s_task_unlock(task);
				return (err);
			}
			s_mapping_u_unlock(mapping);
			s_mapping_f_unlock(mapping);
			link_produce(&mapping->task, &task->sob);
		}
	}
    
	/*
	 * At this point, the mapping is associated with the correct task.
	 * We are done with that task, so unlock it.
	 */
	if (task)
		s_task_unlock(task);

	/*
	 * Handle region reference.
	 * If we are moving to a new region or removing from any region
	 * handle that now.  Unlink from the old region, invalidate all
	 * derived translations, and insert on the new list.
	 * Fault lock must be held before using or changing the region
	 * reference.
	 */
	if (!sameregion) {
		s_mapping_f_lock(mapping);
		region = (s_region_t *)link_consume(&mapping->region);
		link_init(&mapping->region);
		if (region) {
			queue_remove(&region->mappings, mapping,
				     s_mapping_t *, rlink);
			if (!flushed) {
				s_mapping_u_lock(mapping);
				s_mapping_restrict(mapping, 0, mapping->size,
						   FLUKE_PROT_NONE);
				s_mapping_u_unlock(mapping);
			}
			s_region_unlock(region);
		}

		/*
		 * Again, if either we just flushed mappings or there was
		 * no region, we don't have to flush mapping again.
		 */
		flushed = TRUE;

		if (region_link) {
			region = (s_region_t *)link_follow(region_link);
			if (region) {
				queue_enter(&region->mappings, mapping,
					    s_mapping_t *, rlink);
				link_produce(&mapping->region, &region->sob);
				s_region_unlock(region);
			}
		}
		s_mapping_f_unlock(mapping);
	}

	/*
	 * Changing region offset, invalidate translations if not already done.
	 * Need fault and unmap lock for updating region offset field,
	 * unmap lock for flushing.
	 */
	if (state && mapping->region_off != state->region_off) {
		s_mapping_f_lock(mapping);
		s_mapping_u_lock(mapping);
		if (!flushed) {
			s_mapping_restrict(mapping, 0, mapping->size,
					   FLUKE_PROT_NONE);
			flushed = TRUE;
		}
		mapping->region_off = state->region_off;
		s_mapping_u_unlock(mapping);
		s_mapping_f_unlock(mapping);
	}

	/*
	 * Protection change, modify only the mappings for our task.
	 */
	if (state && mapping->prot != state->prot)
		s_mapping_protect(mapping, state->prot);

	return (0);
}

/*
 * Change the protection on a mapping.
 * This routine may either increase or decrease the protection arbitrarily.
 * The actual protection used in translations cannot exceed that of the region
 * though we don't restrict that here (in general, there is no way to ensure
 * this since the region can reduce its protection at any time).
 * Mapping object lock should be held on call and return.
 */
void s_mapping_protect(s_mapping_t *mapping, fluke_prot_t prot)
{
	assert(mapping && mapping->sob.active);
	assert(mapping->sob.type == FLUKE_MAPPING);
	assert_s_ob_locked(&mapping->sob);
	assert((prot & FLUKE_PROT_ALL) == prot);

	if (mapping->prot == prot)
		return;

	/*
	 * Changing the protection requires acquiring the fault and
	 * unmap locks, both for flushing translations and for
	 * changing the protection field.
	 */
	s_mapping_f_lock(mapping);
	s_mapping_u_lock(mapping);

	/*
	 * If we are further restricting permissions,
	 * we must do it now. 
	 */
	if ((mapping->prot & prot) != mapping->prot)
		mapping_frag_protect(mapping, 0, mapping->size, prot);
	/*
	 * If we are relaxing, let it happen on demand.
	 * In order to eagerly determine how much to relax the actual
	 * VtoP translation protections, we would have to trace down
	 * from the root to determine the maximum allowable value.
	 */
	else
		;

	mapping->prot = prot;

	s_mapping_u_unlock(mapping);
	s_mapping_f_unlock(mapping);
}

/*
 * Destroy all the fragments associated with a subset of a mapping
 * and, recursively, all fragments associated with (subsets of) mappings
 * derived (via regions) from this one.
 * Mapping unmap lock should be held on call and return.
 *
 * XXX belongs in unmap layer (combine with s_region_restrict)
 */
void s_mapping_restrict(s_mapping_t *mapping, oskit_addr_t offset,
			oskit_size_t size, fluke_prot_t prot)
{
	s_region_t *region;
	oskit_addr_t foffset, fsize, rsva, reva, msva, meva;
	int oversion; /* XXX */

	assert_mutex_locked(&mapping->ulock);

	/*
	 * Short-circuit check.  If mapping has no associated task or size is
	 * zero there should be no translations from this point on.
	 */
	if (mapping->taskp == 0 || mapping->size == 0) {
#ifdef USEAMM
		/* ??? */
#else
		assert(queue_empty(&mapping->fragments));
#endif
		return;
	}

	/*
	 * For every region derived from this mapping range,
	 * flush the relevant portion.
	 */
	msva = mapping->start_va + offset;
	meva = msva + size;
	assert(meva <= mapping->start_va + mapping->size);

	s_task_u_lock(mapping->taskp);
restart:
	queue_iterate(&mapping->taskp->regions, region, s_region_t *, tlink) {
		/*
		 * XXX we can deadlock here if a task maps a region into itself.
		 */
		s_region_u_lock(region);

		rsva = region->start_va;
		reva = rsva + region->size;

		if (rsva < meva && reva > msva) {
			foffset = max(rsva, msva);
			fsize = min(reva, meva) - foffset;
			foffset -= rsva;
			/*
			 * XXX must drop the task ulock in case the region
			 * is mapped into that task (either directly or
			 * indirectly).  Otherwise we would lock against
			 * ourselves.  This needs a better solution.
			 */
			oversion = mapping->taskp->version;
			s_task_u_unlock(mapping->taskp);

			s_region_restrict(region, foffset, fsize, prot);
			s_region_u_unlock(region);

			/*
			 * XXX reaquire the task ulock after dropping the
			 * region lock.  If the task region list has changed,
			 * we have to restart.
			 */
			s_task_u_lock(mapping->taskp);
			if (oversion != mapping->taskp->version)
				goto restart;
		} else
			s_region_u_unlock(region);
	}
	s_task_u_unlock(mapping->taskp);

	/*
	 * If the mapping allows no stronger access than indicated,
	 * then none of its frags can, so there is no need to check them.
	 *
	 * N.B. note that this check is after we have checked all descendent
	 * regions.  This is because mapping protections do not affect regions.
	 */
	if ((mapping->prot & prot) == mapping->prot)
		return;

	/*
	 * Now modify or remove any VtoP translations for this mapping range.
	 */
	mapping_frag_protect(mapping, offset, size, prot);
}

/*
 * Insert a mapping onto a particular task's mapping list.
 * Mappings in the list are ordered by start address for fast lookup.
 * Returns an insanity condition if the mapping will overlap with others,
 * otherwise the mapping is entered on the task list.
 * Task fault (object) lock should be held when called.
 *
 * XXX mapping unmap lock should also be held.  This is to avoid a race
 * in s_mapping_set_state.
 *
 * XXX belongs in unmap layer? (task layer?)
 */
int s_mapping_insert_task(s_mapping_t *mapping, s_task_t *task)
{
	oskit_addr_t sva, eva;
	s_mapping_t *mp;

	assert(mapping);
	assert_mutex_locked(&mapping->ulock);	/* XXX */
	assert(task);
	assert_s_ob_locked(&task->sob);

	/*
	 * Check address range of target task for overlap.
	 */
	sva = mapping->start_va;
	eva = sva + mapping->size;
	queue_iterate(&task->mappings, mp, s_mapping_t *, tlink) {
		if (mp == mapping)
			continue;
		if (eva <= mp->start_va)
			break;
		if (sva < mp->start_va + mp->size)
			EXC_RETURN_INSANITY(ADDRESS_OVERLAP);
	}

	/*
	 * If no overlap, insert the mapping in the task list and update
	 * the mapping's task association.
	 */
	queue_enter_before(&task->mappings, mp, mapping, s_mapping_t *, tlink);

	mapping->taskp = task;

	return(0);
}

/*
 * Remove a mapping from a particular task's mapping list.
 * Invalidates translations for the mapping and all derived mappings.
 * Task fault (object) lock should be held when called.
 *
 * XXX belongs in unmap layer? (task layer?)
 */
void s_mapping_remove_task(s_mapping_t *mapping, s_task_t *task)
{
	assert(mapping);
	assert(task);
	assert_s_ob_locked(&task->sob);

	queue_remove(&task->mappings, mapping, s_mapping_t *, tlink);
	if (task->mhint == mapping)
		task->mhint = 0;

	s_mapping_u_lock(mapping);
	s_mapping_restrict(mapping, 0, mapping->size, FLUKE_PROT_NONE);
	mapping->taskp = 0;
	s_mapping_u_unlock(mapping);
}

/*
 * Insert a mapping into a region's mapping list.
 * Region unmap lock should be held when called.
 *
 * XXX belongs in unmap layer? (region layer?)
 */
void s_mapping_insert_region(s_mapping_t *mapping, s_region_t *region)
{
	assert(mapping);
	assert(region);
	assert_mutex_locked(&region->ulock);

	queue_enter(&region->mappings, mapping, s_mapping_t *, rlink);
}

#ifdef DEBUG
void
s_mapping_dump(s_mapping_t *map, int indent)
{
	int i;

	for (i = 0; i < indent; i++) putchar(' ');
	printf("M%p: task=%p, region=%p, region_off=0x%08x, start_va=0x%08x, size=0x%08x, prot=0x%08x(%c%c%c%c)\n",
	       map,
	       map->taskp,
	       map->region.target,
	       map->region_off,
	       map->start_va,
	       map->size,
	       map->prot,
	       (map->prot&8)?'C':'-',
	       (map->prot&4)?'r':'-',
	       (map->prot&2)?'w':'-',
	       (map->prot&1)?'x':'-');
#if 0
	printf(", ");
	mapping_frag_dump(map, indent);
#endif
#if 0
	for (i = 0; i < indent; i++) putchar(' ');
	printf("region %p:\n", map->region.target);
	if (map->region.target)
		s_debug_dump_flob(map->region.target, indent+1);
#endif
}
#endif
