/*
 * Copyright (c) 1995, 1998, 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#ifndef _FLUKE_KERNEL_S_OB_H_
#define _FLUKE_KERNEL_S_OB_H_

#include <oskit/machine/types.h>
#include <fluke/ob.h>

#if FLASK
#include <flask/security.h>
#endif

#include "mutex.h"
#include "link.h"
#include "acount.h"
#include "obinfo.h"
#include "config_flob_check.h"

#ifdef DEBUG
struct s_ob_st {
	struct s_ob		*thread;	/* creating thread */
	unsigned		buf[7];		/* trace at creation time */
};
#endif

typedef struct s_ob
{
	/* Fluke object type.  */
	fluke_type_t		type;

	/* True if this object is currently active;
	   false if it has been destroyed
	   and is merely waiting for all references to it to be dropped.  */
	oskit_bool_t		active;

#ifdef CONFIG_FLOB_CHECK
	/* True if the user portion of this object should be checked
	   for corruption prior to use.  */
	oskit_bool_t		flob_debug;
#endif

	/* Count of outstanding references to this object.  */
	acount_t		refs;

	/* Mutex lock protecting this object.
	   This mutex covers all data in the basic s_ob part of the object,
	   and may also protect some or all of the type-specific data.  */
	struct mutex		lock;

	/* Physical address at which the user part of the object is located.  */
	oskit_addr_t		u_pa;

	/* Hash value.
	   Set at object creation time, exported to user via get_state.  */
	fluke_hash_t		u_hash;

	/* Static object type info.
	   Size of kernel/user parts, alignments, etc.  */
	OB_INFO_FIELDS;

#if FLASK
	/* Fluke object SID. */
	security_id_t		sid;
#else
	/* Keep Fluke/flask same size. */
	int			pad;
#endif

	/* Chain on the global object list.
	   XXX should be a hash table or binary tree or something.  */
	struct s_ob		*all_next;

#ifdef DEBUG
	/* Useful debug information.
	   A label to provide more information about what this object is,
	   and a stacktrace recording where the object was created from.  */
        char                    *label;
	struct s_ob_st		*st;
#endif
} s_ob_t;



/*
 * Create a s_ob and initialize it appropriately.
 * The object-type-specific part must be initialized elsewhere
 * (e.g. in the object-type-specific create routine that calls this).
 * Returns with the newly created object locked.
 */
int s_ob_create(
	fluke_type_t		type,
	oskit_addr_t		u_pa,
	fluke_hash_t		u_hash,
	s_ob_t			**out_ob);

/*
 * Create a s_ob in the provided memory.
 * (The object-type-specific part might already be in use, it's not clobbered.)
 * The object is locked on return.
 */
int s_ob_enter(
	fluke_type_t		type,
	oskit_addr_t		u_pa,
	fluke_hash_t		u_hash,
	s_ob_t			*ob);

#ifdef DEBUG
/*
 * Store the current user stack trace in the object
 */
void s_ob_store_trace(
	struct s_ob		*ob);

/*
 * With a debugging enabled, provide the ability to ``label'' flobs
 * with a string.  Max string length is defined in s_ob.c
 */
void s_ob_label(
	struct s_ob             *ob,
	char                    *label);

/*
 * Dump a list of all thread objects, using s_debug_dump_flob().
 */
void s_ob_dump_threads(void);

int s_debug_dump_flob(s_ob_t *sobp, int indent);

#endif

/*
 * Destroy an active s_ob.
 * The object must be locked on entry and is released on return.
 * It will be returned to its zone as soon as possible.
 */
void s_ob_destroy(
	struct s_ob		*ob);

/*
 * Lookup an object given a kernel virtual address for the user part.
 * Returns an error (and sets insanity code) if the object wasn't found.
 * If object is found, it is returned locked.
 */
int s_ob_lookup(
	oskit_addr_t		u_pa,
	struct s_ob		**obp);

/*
 * Find all objects who fall in the range [*pa - epa] and record them in
 * the provided oblist.  The global list lock is held during the search to
 * ensure that objects don't move or get destroyed.
 * Returns 0 if all objects were recorded, 1 if there are more objects in
 * the range.  For the latter, *pa is updated as a hint to show where to
 * start a new search.
 */
int s_ob_search(
	oskit_addr_t		*pa,
	oskit_addr_t		epa,
	fluke_obspec_t		*oblist,
	unsigned int		*lsizep);

/*
 * Associates the supervisor part of an object with a new kernel VA.
 * Object should be locked on call.
 */
int s_ob_move(
	struct s_ob		*ob,
	oskit_addr_t		new_pa);

#define s_ob_gethash(ob)	((ob)->u_hash)

#define s_ob_lock(ob)		mutex_lock(&(ob)->lock)
#define s_ob_trylock(ob)	mutex_trylock(&(ob)->lock)
#define s_ob_unlock(ob)		mutex_unlock(&(ob)->lock)
#define assert_s_ob_locked(ob)	assert_mutex_locked(&(ob)->lock)

#define s_ob_incrref(ob)	(void)acount_incr(&(ob)->refs)
#define s_ob_decrref(ob)	(acount_decr(&(ob)->refs) == 0 ? \
					s_ob_free(ob) : 0)

/* Private functions to be used only by the macros above.  */
void s_ob_free(
	struct s_ob		*ob);

/*
 * Initialize the supervisor object layer at boot time.
 */
void s_ob_boot(void);

/* Dump all objects */
void s_ob_dump_all(void);

/*
 * Apply f to each sob of the indicated type in the sob list.
 *
 * The specified args are supplied to f on each call.
 * If f returns a non-zero value, then s_ob_map
 * will immediately return that value.
 */ 
int s_ob_map(int obmask, int (*f)(s_ob_t *sob, void *args), void *args);


#if FLASK
int s_ob_stop_all_threads(void);
int s_ob_release_all_threads(void);

int s_ob_revoke(security_class_t sclass,
		 security_id_t ssid, 
		 security_class_t tclass,
		 security_id_t tsid, 
		 access_vector_t revoke);

int s_ob_try_remove(security_class_t sclass,
                 security_id_t ssid,
                 security_class_t tclass,
                 security_id_t tsid,
                 access_vector_t permissions,
		 access_vector_t *out_retained);

int s_ob_reset(void);
#endif


#ifdef CONFIG_INLINING
#include "debug.h"

#define LINK_INVALID	((struct s_ob*)-1)

/*
 * Destroy a link and reinitialize it to null (pointing to no object).
 */
OSKIT_INLINE void link_reinit(
	struct link		*link)
{
	struct s_ob *ob = link->target;

	assert(ob != LINK_INVALID);

	if (ob != 0)
		s_ob_decrref(ob);

	link->target = 0;
}

/*
 * Initialize a link to point to an object.
 */
OSKIT_INLINE void link_produce(
	struct link		*link,
	struct s_ob		*target)
{
	assert(link);
	assert(target && target != LINK_INVALID);

	link->target = target;
	s_ob_incrref(target);
}

/*
 * Create a copy of a link (either an object link or a number link).
 * The memory pointed to by 'dest' is assumed to be uninitialized.
 */
OSKIT_INLINE void link_create_copy(
	struct link		*src,
	struct link		*dest)
{
	struct s_ob *ob = src->target;

	assert(ob != LINK_INVALID);

	dest->target = ob;
	if (ob != 0)
		s_ob_incrref(ob);
}

/*
 * Destroy a link, leaving its contents undefined.
 */
OSKIT_INLINE void link_destroy(
	struct link		*link)
{
	struct s_ob *ob = link->target;

	assert(ob != LINK_INVALID);

	if (ob != 0)
		s_ob_decrref(ob);

	do_debug(link->target = LINK_INVALID);
}

/*
 * Given two already-initialized links,
 * change the 'dest' link to point to whatever 'src' points to.
 */
OSKIT_INLINE void link_transfer(
	struct link		*src,
	struct link		*dest)
{
	struct s_ob *ob;

	assert(src); assert(src->target != LINK_INVALID);
	assert(dest); assert(dest->target != LINK_INVALID);

	ob = src->target;
	link_destroy(dest);

	dest->target = ob;
	if (ob)
		s_ob_incrref(ob);
}

/*
 * Try to follow a link and lock the object it points to.
 * If successful, returns a pointer to the now-locked object.
 * If the link is stale (or becomes stale before the object can be locked),
 * this function returns null.
 */
OSKIT_INLINE struct s_ob *link_follow(
	struct link		*link)
{
	struct s_ob *ob;

	assert(link); assert(link->target != LINK_INVALID);

	ob = link->target;
	if (ob == 0)
		return 0;

	assert(ob->type != FLUKE_NULL);

	s_ob_lock(ob);

	if (ob->active)
		return ob;

	/* The link is stale, so clear it and drop the reference.  */
	link->target = 0;

	s_ob_unlock(ob);
	s_ob_decrref(ob);

	return 0;
}


/*
 * Try to follow a link and obtain the hash of the referenced object.
 * If successful, returns the hash.
 * If the link is stale this function returns 0.
 * The object is not locked at any point during this operation.
 */
OSKIT_INLINE fluke_hash_t link_hash(
	struct link		*link)
{
	struct s_ob *ob;

	assert(link); assert(link->target != LINK_INVALID);

	ob = link->target;
	if (ob == 0)
		return 0;

	assert(ob->type != FLUKE_NULL);

	if (ob->active)
		return ob->u_hash;

	/* The link is stale, so clear it and drop the reference.  */
	link->target = 0;
	s_ob_decrref(ob);

	return 0;
}


/*
 * Find the target of a link and consume the link itself.
 * If successful, returns a pointer to the now-locked object.
 * If the link was stale, returns null.
 * In either case, the contents of the link is left uninitialized.
 */
OSKIT_INLINE struct s_ob *link_consume(
	struct link		*link)
{
	struct s_ob *ob = link_follow(link);

	if (ob)
		s_ob_decrref(ob);

	do_debug(link->target = LINK_INVALID);

	return ob;
}

/*
 * Compare two links to see if they point to the same thing.
 * If one of the links is NULL (or stale) but the other is invalid,
 * this routine always returns false, as you would expect.
 * However, if both links are NULL (or stale),
 * then this routine may return either true or false.
 * e.g. depending on whether they _used_ to point to the same object.
 */
OSKIT_INLINE int link_compare(
	struct link		*link1,
	struct link		*link2)
{
	return (link1->target == link2->target);
}

/*
 * Compare a link with a pointer to a locked object.
 */
OSKIT_INLINE int link_compare_ptr(
	struct link		*link,
	struct s_ob		*ptr)
{
	assert(ptr != 0);
	assert(link->target != LINK_INVALID);

	return (link->target == ptr);
}

#endif CONFIG_INLINING

#endif /* _FLUKE_KERNEL_S_OB_H_ */
