/*
 * Copyright (c) 1996 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#ifndef _FLOSS_KERNEL_FLUKE_S_REGION_H_
#define _FLOSS_KERNEL_FLUKE_S_REGION_H_

#include <oskit/queue.h>
#include <fluke/vm.h>
#include "s_ob.h"

/*
 * Flux memory region implementation.
 *
 * A region defines a contiguous range of virtual address space
 * in a task that can be exported to other tasks to map into their
 * address space via mappings.
 *
 * Locking protocol:
 *
 * For regions, the object and fault layers are covered by the object lock
 * and the unmap layer has a separate lock.
 *
 * The start_va, size and prot fields are read-only for the fault and unmap
 * layers but may be written by the object layer (region_set_state).  The
 * keeper and task fields are read-only for the fault layer and read-write
 * for the object layer.  Hence, region_set_state should hold both locks
 * when modifying any of these fields.
 */
typedef struct s_region {
	/* general object layer, protected by object lock */
	struct s_ob	sob;		/* supervisor object state */
	struct link	task;		/* task over which region is defined */
	struct link	keeper;		/* port on which to send fault msg */
	oskit_addr_t	start_va;	/* start VA of the region */
	oskit_size_t	size;		/* size of the region */
	fluke_prot_t	prot;		/* protection attributes granted */

	/* fault layer, uses object lock as well */

	/* unmap layer */
	struct mutex	ulock;		/* lock */
	queue_chain_t	tlink;		/* all regions on the same task */
	queue_head_t	mappings;	/* mappings derived from this region */
} s_region_t;

/* XXX for now */
#define s_region_lock(r)	s_ob_lock(&(r)->sob)
#define s_region_unlock(r)	s_ob_unlock(&(r)->sob)
#define s_region_f_lock(r)	s_ob_lock(&(r)->sob)
#define s_region_f_unlock(r)	s_ob_unlock(&(r)->sob)
#define s_region_u_lock(r)	mutex_lock(&(r)->ulock);
#define s_region_u_unlock(r)	mutex_unlock(&(r)->ulock);

struct s_ref;
struct s_task;
struct fluke_region_state;

int s_region_create(oskit_addr_t uva, fluke_hash_t hash, s_region_t **out_region);
void s_region_reference(s_region_t *region, struct s_ref *region_ref);
void s_region_destroy(s_region_t *region);
int s_region_get_state(s_region_t *region, struct fluke_region_state *state,
		       struct link *task_link, struct link *keeper_link);
int s_region_set_state(s_region_t *region, struct fluke_region_state *state,
		       struct link *task_link, struct link *keeper_link);
void s_region_protect(s_region_t *region, fluke_prot_t prot);
void s_region_restrict(s_region_t *region, oskit_addr_t offset, oskit_size_t size,
		       fluke_prot_t prot);
void s_region_insert_task(s_region_t *region, struct s_task *task);
void s_region_remove_task(s_region_t *region, struct s_task *task);
int s_region_range(s_region_t *region, oskit_addr_t offset, oskit_size_t size,
		   oskit_addr_t *saddrp, oskit_addr_t *eaddrp);

#ifdef DEBUG
/*
 * Dump state of thread to console
 */
void s_region_dump(s_region_t *reg, int indent);
#endif

#endif /* _FLOSS_KERNEL_FLUKE_S_REGION_H_ */
