/*
 * Copyright (c) 1995, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include <fluke/ref.h>
#include <fluke/task.h>
#include <fluke/exception.h>
#include <oskit/page.h>
#include <oskit/machine/base_vm.h>

#include "s_task.h"
#include "s_mapping.h"
#include "frag.h"
#include "s_region.h"
#include "s_ref.h"
#include "s_thread.h"
#include "exception.h"
#include "debug.h"

#include "config_instrument.h"
#include "config_flob_check.h"
#include "config_opt_threads.h"
#include "config_opt_synch.h"
#include "config_user_gprof.h"
#include "s_instrument.h"

#if FLASK
#include "avc.h"
#endif 

#include "xfer.h"
#include "obinfo.h"

/*#define OBCACHE_STATS*/

#ifdef OBCACHE_STATS
/*
 * 0 == lookups
 * 1 == hit: good
 * 2 == hit: object gone
 * 3 == hit: protection bad
 * 4 == miss: capacity
 * 5 == miss: conflict
 * 6 == creates
 * 7 == create: capacity displacement
 * 8 == create: conflict displacement
 * 9 == moves
 * 10 == move: source hit
 * 11 == move: dest conflict displacement
 * 12 == hit: object corrupt
 * 13 == hit: flushed
 * 14 == nocache calls
 */
unsigned ocachestats[16];

/*
 * # of times each entry used
 */
unsigned ocacheuse[OBTAB_SIZE];

#define OBCACHE_STATS_INC(b) \
	ocachestats[b]++
#define OBCACHE_STATS_INCUSE(b) \
	ocacheuse[b]++
#define OBCACHE_STATS_MISSTYPE(ot, b) { \
	int i; \
	for (i = 0; i < OBTAB_SIZE; i++) \
		if ((ot)[i].uva == -1) break; \
	ocachestats[i < OBTAB_SIZE ? (b+1) : (b+0)]++; \
}
#else
#define OBCACHE_STATS_INC(b)
#define OBCACHE_STATS_INCUSE(b)
#define OBCACHE_STATS_MISSTYPE(ot, b)
#endif

fluke_task_t kernel_task;
link_t kernel_task_link;

#ifdef CONFIG_FLOB_CHECK
/* The signature written into the user space part of objects */
static const unsigned flob_sign[] = {
	0x645f3d5f, 0x20746e6f, 0x72657665,
	0x756f7420, 0x74206863, 0x20736968,
	0x6f6d656d, 0x005f7972
};

#define FLOB_SIGN_SIZE(usize) ( (usize) < sizeof(flob_sign) ? \
				(usize) : sizeof(flob_sign))
#endif /* CONFIG_FLOB_CHECK */

void s_task_boot(void)
{
	s_task_t *task;
	int rc;

	rc = s_task_create((oskit_addr_t)&kernel_task, 0, &task);
	assert(rc == 0);
#ifdef DEBUG
	s_ob_label(&task->sob, "kernel task");
#endif

	link_produce(&kernel_task_link, &task->sob);
	s_ob_unlock(&task->sob);
}

/*
 * Create an ``empty'' task object.
 * Keeper reference is NULL.  New task object is returned locked.
 */
int s_task_create(
	oskit_addr_t		uva,
	fluke_hash_t		hash,
	s_task_t		**out_task)
{
	s_task_t *task;
	int err, i;

	err = task_ob_create(current_thread(), FLUKE_TASK, uva, hash,
			     (s_ob_t **)&task);
	if (err) return err;

	link_init(&task->keeper);
	queue_init(&task->threads);
	cond_init(&task->cond);
#if FLASK
	task->segment_sid = SECSID_NULL;
#endif 

	task->mhint = 0;
	queue_init(&task->mappings);

	mutex_init(&task->ulock);
	task->version = 0;	/* XXX hack around race in s_mapping_restrict */
	queue_init(&task->regions);

	for (i = 0; i < OBTAB_SIZE; i++) {
		task->obtab[i].uva = -1;
		link_init(&task->obtab[i].link);
	}

	err = task_pmap_state_init(task);
	if (err) {
		s_ob_destroy(&task->sob);
		return err;
	}

#ifdef IOPL0
	/* init the permissions table */
	memset(&task->bitmask, 0xff, sizeof(task->bitmask));

	/* This port is sometimes used for timings... */
	io_alloc(task->bitmask, 0x80, 0x1);

	/* XXX: temp hack to give access to PIC */
	io_alloc(task->bitmask, 0x20, 0x2);
	io_alloc(task->bitmask, 0xa0, 0x2);
#endif IOPL0

#ifdef CONFIG_USER_GPROF
	task->p_flag = 0;
	task->p_prof = NULL;
#endif
	*out_task = task;
	return 0;
}

#if FLASK
#include "private_interfaces.h"

int do_private_task_create_secure(void *args)
{
        struct task_create_secure_args *a = args;
        s_task_t *task;
	int rc;

	rc = s_task_create(a->uva, 0, &task);
	if (rc)
		return rc;

	task->sob.sid = a->sid;
	s_ob_unlock(&task->sob);

	return 0;
}


int do_private_segment_task_create(void *args)
{
        struct segment_task_create_args *a = args;
        s_task_t *task;
	int rc;

	rc = s_task_create(a->uva, 0, &task);
	if (rc)
		return rc;

	task->segment_sid = a->sid;

	s_ob_unlock(&task->sob);

	return 0;
}
#endif 

int s_task_destroy(
	s_task_t		*task)
{
	s_mapping_t *mapping;
	s_region_t *region;

	assert(task && task->sob.active);
	assert(task->sob.type == FLUKE_TASK);
	assert_s_ob_locked(&task->sob);

	/*
	 * Disassociate task from all threads running in it.
	 * We have to kick the threads and let them take themselves
	 * off of the task's threads list:
	 * we can't take them off ourselves directly,
	 * because they might be running in user mode
	 * in this task's address space right now.
	 */
	if (!queue_empty(&task->threads))
	{
		s_thread_t *thread;
		int rc;
		link_t *linkp = 0;
#ifdef CONFIG_PRIO_INHERIT
		link_t link;
#endif

		/*
		 * Cancel all threads attached to this task.
		 */
		queue_iterate(&task->threads, thread,
			      s_thread_t *, task_state.task_chain)
		{
			assert(thread->sob.type == FLUKE_THREAD);
			thread->task_state.task_dying = 1;
			thread_cancel(thread);
#ifdef CONFIG_PRIO_INHERIT
			if (!linkp) {
				/*
				 * We will donate our priority to
				 * this thread until it finishes
				 * its cancellation and wakes us up.
				 */
				link_produce(&link, &thread->sob);
				linkp = &link;
			}
#endif
		}

		/*
		 * Wait until those threads wake up and remove themselves.
		 */
		rc = cond_wait(&task->cond, &task->sob.lock, linkp);
		if (rc)
			return rc;
		return KR_CANCEL;
	}

	/*
	 * Disassociate task from all regions.
	 * Task unmap lock covers the region list.
	 *
	 * N.B. we do this before handling mappings to minimize the amount of
	 * translation invalidation traversals we have to do.  For each
	 * region we will traverse down the hierarchy to flush all descendent
	 * translations.  Then when we disassociate from mappings we will
	 * only have to invalidate the translations associated with the
	 * mapping, since there will be no associated regions at that point.
	 * If we did mappings first we would descend the tree twice.
	 */
	s_task_u_lock(task);
	while (!queue_empty(&task->regions)) {
		region = (s_region_t *)queue_first(&task->regions);
		s_region_remove_task(region, task);
	}
	s_task_u_unlock(task);

	/*
	 * Disassociate task from all mappings.
	 * Task fault (object) lock covers the mapping list.
	 */
	while (!queue_empty(&task->mappings)) {
		mapping = (s_mapping_t *)queue_first(&task->mappings);
		s_mapping_remove_task(mapping, task);
	}

#ifdef DEBUG
	/*
	 * Now that we have flushed all the mappings,
	 * make sure no cache references are left valid.
	 */
	{
		int i;

		for (i = 0; i < OBTAB_SIZE; i++)
			assert(task->obtab[i].uva == -1);
	}
#endif

	link_destroy(&task->keeper);

	/*
	 * Destroy the PMAP layer's task-specific state for this task.
	 */
	task_pmap_state_destroy(task);

	s_ob_destroy(&task->sob);
	return 0;
}

int s_task_get_state(
	struct s_task			*task,
	fluke_task_state_t		*state,
	struct link			*keeper_link)
{
	assert(task && task->sob.active);
	assert(task->sob.type == FLUKE_TASK);
	assert_s_ob_locked(&task->sob);

	if (state)
		state->hash = s_ob_gethash(&task->sob);

	if (keeper_link)
		link_create_copy(&task->keeper, keeper_link);

	return (0);
}

int s_task_set_state(
	struct s_task			*task,
	fluke_task_state_t		*state,
	struct link			*keeper_link)
{
	assert(task && task->sob.active);
	assert(task->sob.type == FLUKE_TASK);
	assert_s_ob_locked(&task->sob);

	if (state)
		;	/* XXX no state other than the hash */

	if (keeper_link && !link_compare(&task->keeper, keeper_link))
		link_transfer(keeper_link, &task->keeper);

	return (0);
}

/*
 * Initialize the per-thread state maintained by the task layer for a thread.
 */
void thread_task_state_init(
	struct s_thread			*thread)
{
	thread->task_state.task = 0;
	thread->task_state.task_dying = 0;
	link_init(&thread->task_state.task_link);
}

void thread_task_state_start(
	struct s_thread			*thread)
{
	if (thread->task_state.task == 0)
	{
		s_task_t *task;

		task = (s_task_t *)link_follow(&thread->task_state.task_link);
		if (task == 0)
			return;	/* We really have no task! */
		if (task->sob.type != FLUKE_TASK) {
			/* XXX sys_thread_set_state should detect this
			 * and throw an insanity trap? */
			s_task_unlock(task);
			return;
		}

		/* Register this thread as an active user of the task. */
		queue_enter(&task->threads, thread,
			    s_thread_t *, task_state.task_chain);
		thread->task_state.task = task;
		pmap_activate(thread, task);

		s_task_unlock(task);
	}
}

void thread_task_state_stop(
	struct s_thread			*cur_thread)
{
	s_task_t *task = cur_thread->task_state.task;
	if (task != 0)
	{
		s_task_lock(task);

		pmap_deactivate(cur_thread, task);

		assert(!queue_empty(&task->threads));
		queue_remove(&task->threads, cur_thread,
			     s_thread_t *, task_state.task_chain);

		/*
		 * If we're the last thread off the list,
		 * signal the task's condition variable
		 * in case someone is trying to destroy the task.
		 */
		if (queue_empty(&task->threads))
			cond_broadcast(&task->cond);

		s_task_unlock(task);

		cur_thread->task_state.task = 0;
	}
}

/*
 * Destroy the per-thread state maintained by the task layer for a thread.
 */
void thread_task_state_destroy(
	struct s_thread			*thread)
{
	/* We should already have been detached from our task.  */
	assert(thread->task_state.task == 0);
	link_destroy(&thread->task_state.task_link);
}

#if 0
/*
 * Get the current thread's task association.
 */
void thread_task_state_get(
	struct s_thread			*cur_thread,
	link_t				*task_link)
{
	assert(task_link);
	assert(cur_thread->task_state.task == 0);
	link_transfer(&cur_thread->task_state.task_link, task_link);
}

/*
 * Set the current thread's task association
 * to refer to the specified task.
 */
int thread_task_state_set(
	struct s_thread			*cur_thread,
	link_t				*task_link)
{
	assert(task_link);
	assert(cur_thread->task_state.task == 0);
}
#endif

void thread_task_canceled(
	struct s_thread			*cur_thread)
{
	if (cur_thread->task_state.task_dying)
	{
		thread_task_state_stop(cur_thread);
		assert(cur_thread->task_state.task == 0);

		/*
		 * Null out our link to the old task
		 * so we won't immediately try to attach to it again.
		 */
		link_destroy(&cur_thread->task_state.task_link);
		link_init(&cur_thread->task_state.task_link);
		cur_thread->task_state.task_dying = 0;
	}
}

#ifdef CONFIG_FLOB_CHECK

/*
 * Setup the user space part of an object.
 * This writes a signature into userspace.
 */
static void
task_ob_user_set(s_ob_t *ob, oskit_addr_t uva)
{
	switch (ob->type) {
#ifdef CONFIG_OPT_THREADS
	case FLUKE_THREAD:
#endif
#ifdef CONFIG_OPT_SYNCH
	case FLUKE_MUTEX:
#endif
#ifdef CONFIG_OPT_REF
	case FLUKE_REF:
#endif
	case FLUKE_NULL:
		ob->flob_debug = 0;
		return;
	default:
		break;
	}

	memcpy((void *) phystokv(ob->u_pa),
	       flob_sign, FLOB_SIGN_SIZE(OB_USIZE(ob)));
	ob->flob_debug = 1;
}

/*
 * Check the signature written in the userspace part
 * return: 1 == User space part corrupt, 0 otherwise
 *
 * The uva argument is purely for informational purposes.
 * We check the user-space contents by accessing the physical
 * address stored in the flob.
 */
int
task_ob_user_check(s_ob_t *ob, oskit_addr_t uva)
{
	if (ob->flob_debug &&
	    memcmp(flob_sign, (void *) phystokv(ob->u_pa),
		   FLOB_SIGN_SIZE(OB_USIZE(ob)))) {
#ifdef DEBUG
		printf("%s: corrupt flob: type=%d, uva=%08x, eip=%08x\n",
		       "FLOB_CHECK",
		       ob->type, uva, EXC_GET_IP(current_thread()));
		s_debug_dump_flob(ob, 0);
#endif
		return 1;
	}
	return 0;
}
#endif /* CONFIG_FLOB_CHECK */


/*
 * ``Creates'' a fluke object in the specified task at the given address.
 * If successful, returns the (locked) supervisor part of the object in the
 * provided pointer.
 */
int task_ob_create(s_thread_t *thread, fluke_type_t type, oskit_addr_t uva,
		   fluke_hash_t hash, s_ob_t **sobp)
{
	struct obtab_entry *obtab;
	oskit_addr_t u_pa;
	s_task_t *task;
	frag_t *frag;
	s_ob_t *ob;
	int rc;

	/*
	 * Make sure alignment is correct.
	 */
	if (uva & (OBT_USIZE(type)-1))
		EXC_RETURN_INSANITY(NOT_ALIGNED);

	task = current_thread()->task_state.task;

	/*
	 * XXX ugh!  No task for thread, we must be in the bootstrap process.
	 * Skip address space lookups.
	 */
	if (task == 0) {
		u_pa = kvtophys(uva);
		rc = s_ob_create(type, u_pa, hash, &ob);
		assert(rc == 0);
#if FLASK
		ob->sid = SECINITSID_KERNEL;
#endif 
		*sobp = ob;
#ifdef CONFIG_FLOB_CHECK
		task_ob_user_set(ob, uva);
#endif
		return (0);
	}

	/*
	 * Make sure memory exists and allows required access.
	 * On successful return, the fragment covering the address is returned
	 * with that fragment's mapping unmap lock held.
	 */
	rc = task_lookup(current_thread(), 0, trunc_page(uva),
			 FLUKE_PROT_CREATE_OB|FLUKE_PROT_WRITE, &frag);
	if (rc)
		return (rc);

	assert(frag_mapping(frag) != 0);

	/*
	 * Translate to a physical address
	 */
	u_pa = frag_vtop(frag, uva);

	/*
	 * Now create a supervisor object.
	 * Supervisor object is returned locked.
	 */
	rc = s_ob_create(type, u_pa, hash, &ob);
	if (rc)
	{
		s_mapping_u_unlock(frag_mapping(frag));
		return (rc);
	}

#if FLASK
	ob->sid = frag->segment_sid;
#endif 

	/*
	 * Enter object in the task lookup cache.
	 *
	 * Once we have grabbed the task unmap lock and set the fragment
	 * obs_cached flag, we can safely drop the mapping unmap lock.
	 * Any attempt to destroy the fragment will see the obs_cached flag
	 * and block attempting to get the task unmap lock.
	 */
	s_task_u_lock(task);
	frag_set_obs_cached(frag);
	s_mapping_u_unlock(frag_mapping(frag));

	OBCACHE_STATS_INC(6);
	obtab = &task->obtab[OBTAB_HASH(uva)];
	if (obtab->uva != -1) {
		OBCACHE_STATS_MISSTYPE(task->obtab, 7);
		link_destroy(&obtab->link);
	}
	obtab->uva = uva;
	obtab->prot = frag_prot(frag);
	link_produce(&obtab->link, ob);

	s_task_u_unlock(task);
#ifdef DEBUG
	s_ob_store_trace(ob);
#endif
#ifdef CONFIG_FLOB_CHECK
	task_ob_user_set(ob, uva);
#endif

	*sobp = ob;
	return (0);
}

/*
 * Lookup an object at the indicated user address with the intention of
 * destroying it.  If the object is of the correct type, we flush it from
 * the task object cache and return a pointer to the locked supervisor object.
 */
int task_ob_destroy(s_thread_t *thread, oskit_addr_t uva, fluke_type_t type,
		    struct s_ob **out_sob)
{
	struct s_ob *sob;
	int rc;

	rc = task_ob_lookup(thread, uva, FLUKE_PROT_WRITE, 0, &sob);
	if (rc)
		return rc;

	if (sob->type != type) {
		s_ob_unlock(sob);
		exc_set_insanity(thread, FLUKE_INSANITY_WRONG_TYPE | type);
		return KR_USER_EXCEPTION;
	}

	*out_sob = sob;
	return 0;
}

/*
 * Lookup a supervisor object given a user virtual address, ensuring that
 * the specified thread has the indicated permission to that memory.
 * If cache is true, then the object is added to the task object cache.
 * If cache is false, then the object is removed from the cache.
 */
int task_ob_lookup(s_thread_t *thread, oskit_addr_t uva, fluke_prot_t prot,
		   int cache, struct s_ob **out_sob)
{
	struct obtab_entry *obtab;
	oskit_addr_t u_pa;
	s_task_t *task;
	s_ob_t *ob;
	frag_t *frag;
	int rc;

	/* XXX check alignment? */

	task = thread->task_state.task;

	/*
	 * XXX ugh!  No task for thread, we must be in the bootstrap process.
	 * Skip address space lookups.
	 */
	if (task == 0) {
		u_pa = kvtophys(uva);
		rc = s_ob_lookup(u_pa, &ob);
		assert(rc == 0);
#ifdef CONFIG_FLOB_CHECK
		if (task_ob_user_check(ob, uva)) {
			s_ob_unlock(ob);
			EXC_RETURN_INSANITY(INVALID_OBJECT);
		}
#endif
		*out_sob = ob;
		return (0);
	}

	OBCACHE_STATS_INC(0);
	if (cache)
		OBCACHE_STATS_INCUSE(OBTAB_HASH(uva));
	else
		OBCACHE_STATS_INC(14);

	/*
	 * Check lookup cache for the object.
	 */
	s_task_u_lock(task);
	obtab = &task->obtab[OBTAB_HASH(uva)];
	if (uva == obtab->uva) {
		/*
		 * Found a match.  Need to check that link is valid and that
		 * protection matches.  We need to copy the obtab link so that
		 * we can drop the task unmap lock before following the link.
		 * If we didn't do this, we might violate the locking hierarchy.
		 *
		 * We check the protection first while we have the task lock.
		 * If the protection doesn't match, it might be because the
		 * frag protection isn't up to date (due to lazy relaxing of
		 * protection) so we do a full search.
		 */
		if ((obtab->prot & prot) == prot) {
			link_t tmplink;

			link_create_copy(&obtab->link, &tmplink);
			if (cache == 0) {
				obtab->uva = -1;
				link_destroy(&obtab->link);
				OBCACHE_STATS_INC(13);
			}
			s_task_u_unlock(task);
			ob = link_consume(&tmplink);
			if (ob) {
#ifdef CONFIG_FLOB_CHECK
				if (task_ob_user_check(ob, uva)) {
					OBCACHE_STATS_INC(12);
					s_ob_unlock(ob);
					EXC_RETURN_INSANITY(INVALID_OBJECT);
				}
#endif
				OBCACHE_STATS_INC(1);
				*out_sob = ob;
				return (0);
			} else
				OBCACHE_STATS_INC(2);
		} else {
			OBCACHE_STATS_INC(3);
			s_task_u_unlock(task);
		}
	} else
		s_task_u_unlock(task);

	/*
	 * Not in the cache.  Simulate an access of the appropriate type.
	 * If successful, will return a pointer to the frag describing
	 * the physical memory for the object.
	 */
	rc = task_lookup(thread, 0, trunc_page(uva), prot, &frag);
	if (rc)
		return (rc);

	/*
	 * Make sure there is an actual object there by looking up the
	 * physical address of the object.
	 *
	 * XXX more locking problems: ob_lookup locks the target object
	 * which again may be out of order with respect to the mapping unmap
	 * lock we hold.
	 */
	assert(frag_mapping(frag) != 0);
	u_pa = frag_vtop(frag, uva);
	rc = s_ob_lookup(u_pa, &ob);
	if (rc) {
		s_mapping_u_unlock(frag_mapping(frag));
		return (rc);
	}

#ifdef CONFIG_FLOB_CHECK
	/*
	 * If the object is corrupt, we don't cache it.
	 */
	if (task_ob_user_check(ob, uva)) {
		s_mapping_u_unlock(frag_mapping(frag));
		s_ob_unlock(ob);
		EXC_RETURN_INSANITY(INVALID_OBJECT);
	}
#endif

	/*
	 * Once we have grabbed the task unmap lock and set the fragment
	 * obs_cached flag, we can safely drop the mapping unmap lock.
	 * Any attempt to destroy the fragment will see the obs_cached flag
	 * and block attempting to get the task unmap lock.
	 */
	s_task_u_lock(task);
	frag_set_obs_cached(frag);
	s_mapping_u_unlock(frag_mapping(frag));

	/*
	 * Now update the cached state.
	 */
	if (obtab->uva != -1) {
		OBCACHE_STATS_MISSTYPE(task->obtab, 4);
		link_destroy(&obtab->link);
	}
	if (cache) {
		obtab->uva = uva;
		obtab->prot = frag_prot(frag);
		link_produce(&obtab->link, ob);
	} else
		obtab->uva = -1;
	s_task_u_unlock(task);

	*out_sob = ob;
	return (0);
}

int task_ob_lookup_type(s_thread_t *thread, oskit_addr_t uva, fluke_prot_t prot,
			fluke_type_t type, struct s_ob **out_sob)
{
	struct s_ob *sob;
	int rc;

	rc = task_ob_lookup(thread, uva, prot, 1, &sob);
	if (rc)
		return rc;

	if (sob->type != type)
	{
		s_ob_unlock(sob);
		exc_set_insanity(thread, FLUKE_INSANITY_WRONG_TYPE | type);
		return KR_USER_EXCEPTION;
	}

	*out_sob = sob;
	return 0;
}

int task_ob_lookup_ref_type(s_thread_t *thread, oskit_addr_t uva,
			fluke_type_t type, struct s_ob **out_sob)
{
	s_ref_t *ref;
	s_ob_t *ob;
	int rc;

	rc = task_ob_lookup_type(thread, uva, FLUKE_PROT_READ, FLUKE_REF,
				 (s_ob_t**)&ref);
	if (rc)
		return rc;

	/* XXX major locking order violation? */
	ob = s_ref_follow(ref);
	s_ref_unlock(ref);

	/*
	 * If the reference points to the wrong type of object,
	 * act like the reference wasn't valid in the first place.
	 * We could generate an insanity trap,
	 * but then anyone receiving a capability for anything
	 * would have to type-check it manually before using it.
	 */
	if (ob && (ob->type != type))
	{
		s_ob_unlock(ob);
		ob = 0;
	}

	*out_sob = ob;
	return 0;
}

/*
 * Search an address range of a task for kernel objects.
 * On call, *obcountp contains the size of the uobs array.
 * On return, it is the number of objects placed in the array.
 * On call, *saddrp contains the start address of the search.
 * On return, it contains where the search terminated.  It will
 * be equal to eaddr if all objects were found, less than eaddr
 * if the array was not large enough to hold all objects.
 *
 * Routine return non-zero due to page faults or other memory faults.
 */
int task_ob_search(
	s_thread_t	*thread,
	s_task_t	*task,
	oskit_addr_t	*saddrp,
	oskit_addr_t	eaddr,
	fluke_obspec_t	*obs,
	unsigned int	*obcountp)
{
	oskit_addr_t saddr, pa, epa;
	frag_t *frag;
	int rc, moreobs, i;
	unsigned int obcount, lobcount;

	saddr = *saddrp;
	obcount = 0;
	while (saddr < eaddr) {
		/*
		 * Lookup pieces of the indicated range and locate all
		 * the objects contained therein.
		 */
		rc = task_lookup(thread, task, saddr, FLUKE_PROT_READ, &frag);
		if (rc)
			return (rc);

		pa = frag_vtop(frag, saddr);
		epa = frag_vtop(frag, eaddr);
		if (epa > (frag_start_pa(frag) + frag_size(frag)))
			epa = (frag_start_pa(frag) + frag_size(frag));

		do {
			/*
			 * Locate objects in this piece of physical memory
			 */
			lobcount = (*obcountp - obcount);
			moreobs = s_ob_search(&pa, epa, obs, &lobcount);
			if (lobcount == 0)
				break;

			/*
			 * Convert locations to virtual addresses
			 */
			for (i = 0; i < lobcount; i++) {
				s_ob_t *ob = (s_ob_t *)obs[i].ob;
				oskit_addr_t uva = frag_ptov(frag, ob->u_pa);
				obs[i].ob = (void *)uva;
				if (obinfo[ob->type].u_sanify)
					/*
					 * The fluke_region_search call
					 * is supposed to fix up mangled
					 * flobs so they behave properly
					 * again, so do that now if this
					 * type needs it.
					 */
					(*obinfo[ob->type].u_sanify)(ob);
#ifdef CONFIG_FLOB_CHECK

				/*
				 * If the user space of this object is
				 * trashed, we fix it.  This works in our
				 * implementation since we just have to
				 * write the magic pattern back out.
				 */
				if (task_ob_user_check(ob, uva)) {
					task_ob_user_set(ob, uva);
#ifdef DEBUG
					printf("%s: flob fixed: type=%d, uva=%08x\n",
					       "FLOB_CHECK", ob->type, uva);
#endif
				}
#endif
			}
			obs += lobcount;
			obcount += lobcount;
		} while (moreobs);

		assert(moreobs || pa == epa);

		saddr = frag_ptov(frag, pa);
		s_mapping_u_unlock(frag_mapping(frag));

		/*
		 * If there are still more objects but we have filled the
		 * array, then quit.  Otherwise keep going until the array
		 * is filled.  If the array is already filled, we must keep
		 * going anyway to determine if there are more objects and
		 * fulfill the requirement that *saddrp point to the end
		 * of the range if all objects were found.
		 */
		if (moreobs && *obcountp == obcount)
			break;
	}

	assert(saddr == eaddr || *obcountp == obcount);

	*obcountp = obcount;
	*saddrp = saddr;

	return (0);
}

/*
 * Move a kernel object from one user VA to another.
 * May return non-zero due to source or destination page faults,
 * other memory faults, or various object related insanity conditions.
 *
 * XXX size and alignment could be inferred from object type.
 */
int task_ob_move(
	s_thread_t	*thread,
	oskit_addr_t	old_uva,
	oskit_addr_t	new_uva,
	fluke_type_t	type,
	oskit_size_t	usize,
	oskit_size_t	copy_usize, /* bytes of user data to copy */
	struct s_ob	**out_sob)
{
	struct obtab_entry *obtab;
	oskit_addr_t old_pa, new_pa;
	s_task_t *task;
	s_ob_t *ob;
	frag_t *old_frag, *new_frag;
	int rc, drc = 0;

	assert(copy_usize <= usize);

	/*
	 * Make sure destination alignment is correct.
	 * XXX check source alignment too?
	 */
	if (new_uva & (usize-1))
		EXC_RETURN_INSANITY(NOT_ALIGNED);

	/*
	 * Lookup and lock target fragments in the task.
	 * Note that old_frag and new_frag may be the same.
	 */
	rc = task_lookup_two_frags(thread, thread, old_uva, new_uva,
				   FLUKE_PROT_WRITE,
				   FLUKE_PROT_WRITE|FLUKE_PROT_CREATE_OB,
				   &old_frag, &new_frag, &drc);
	if (rc) return rc;
	if (drc) return drc;

	task = frag_mapping(old_frag)->taskp;
	assert(task == frag_mapping(new_frag)->taskp);
	assert(task == thread->task_state.task);

	/*
	 * Make sure there was an object at the old location.
	 * XXX locks object out of order.
	 */
	old_pa = frag_vtop(old_frag, old_uva);
	rc = s_ob_lookup(old_pa, &ob);
	if (rc)
	{
		if (frag_mapping(old_frag) != frag_mapping(new_frag))
			s_mapping_u_unlock(frag_mapping(new_frag));
		s_mapping_u_unlock(frag_mapping(old_frag));
		return rc;
	}

	/*
	 * Make sure object is of correct type
	 */
	if (type && ob->type != type)
	{
		s_ob_unlock(ob);
		if (frag_mapping(old_frag) != frag_mapping(new_frag))
			s_mapping_u_unlock(frag_mapping(new_frag));
		s_mapping_u_unlock(frag_mapping(old_frag));
		exc_set_insanity(thread, FLUKE_INSANITY_WRONG_TYPE | type);
		return KR_USER_EXCEPTION;
	}

#ifdef CONFIG_FLOB_CHECK
	if (task_ob_user_check(ob, old_uva)) {
		s_ob_unlock(ob);
		if (frag_mapping(old_frag) != frag_mapping(new_frag))
			s_mapping_u_unlock(frag_mapping(new_frag));
		s_mapping_u_unlock(frag_mapping(old_frag));
		EXC_RETURN_INSANITY(INVALID_OBJECT);
	}
#endif

#if FLASK
	/*
	 * Ensure that flobs are not moved across SID boundaries.
	 */
	if (old_frag->segment_sid != new_frag->segment_sid)
	{
#ifdef DEBUG
	    printf("kernel: flobs may not be moved across SID boundaries.\n");
	    printf("type %d old segsid %ld -> new segsid %ld, obsid %ld\n",
		   ob->type, 
		   old_frag->segment_sid, 
		   new_frag->segment_sid, 
		   ob->sid);
#endif DEBUG
	    s_ob_unlock(ob);
	    if (frag_mapping(old_frag) != frag_mapping(new_frag))
		    s_mapping_u_unlock(frag_mapping(new_frag));
	    s_mapping_u_unlock(frag_mapping(old_frag));
	    EXC_RETURN_INSANITY(PERMISSION_DENIED);
	}
#endif 

	/*
	 * We have now identified that the source is an object of the
	 * correct type.  Check the trivial case of not moving at all.
	 * XXX should this be an error instead?
	 */
	new_pa = frag_vtop(new_frag, new_uva);
	if (old_pa == new_pa)
		goto done;

	/*
	 * Move the supervisor part of the object.
	 */
	rc = s_ob_move(ob, new_pa);
	if (rc)
	{
		s_ob_unlock(ob);
		if (frag_mapping(old_frag) != frag_mapping(new_frag))
			s_mapping_u_unlock(frag_mapping(new_frag));
		s_mapping_u_unlock(frag_mapping(old_frag));
		return rc;
	}

	/*
	 * Now copy the active data in user-space part of the object.
	 */
	memcpy((void *) phystokv(new_pa), (void *) phystokv(old_pa),
	       copy_usize);

	OBCACHE_STATS_INC(9);

	/*
	 * Now update the lookup cache, clear it from its old location
	 * and enter it at the new.
	 * XXX lock ordering, we have the object locked.
	 */
	s_task_u_lock(task);
	obtab = &task->obtab[OBTAB_HASH(old_uva)];
	if (old_uva == obtab->uva)
	{
		OBCACHE_STATS_INC(10);
		obtab->uva = -1;
		link_destroy(&obtab->link);
	}
	obtab = &task->obtab[OBTAB_HASH(new_uva)];
	if (obtab->uva != -1) {
		OBCACHE_STATS_INC(11);
		link_destroy(&obtab->link);
	}
	obtab->uva = new_uva;
	obtab->prot = frag_prot(new_frag);
	link_produce(&obtab->link, ob);
	s_task_u_unlock(task);

done:
	if (frag_mapping(old_frag) != frag_mapping(new_frag))
		s_mapping_u_unlock(frag_mapping(new_frag));
	s_mapping_u_unlock(frag_mapping(old_frag));
#ifdef CONFIG_FLOB_CHECK
	task_ob_user_set(ob, new_uva);
#endif


	*out_sob = ob;
	return (0);
}

/*
 * Special helper routine for space-to-space copy and fluke_*_move
 * functions.  Used to lookup and lock two fragments at the same time
 * while avoiding potential deadlock.
 */
int task_lookup_two_frags(
	s_thread_t		*src_thread,
	s_thread_t		*dest_thread,
	oskit_addr_t		src_uva,
	oskit_addr_t		dest_uva,
	fluke_prot_t		src_prot,
	fluke_prot_t		dest_prot,
	frag_t			**out_src_frag,
	frag_t			**out_dest_frag,
	int			*out_remote_rc)
{
	s_task_t *src_task = src_thread->task_state.task;
	s_task_t *dest_task = dest_thread->task_state.task;
	s_mapping_t *src_mapping = 0;
	s_mapping_t *dest_mapping = 0;
	frag_t *src_frag;
	frag_t *dest_frag;
	int local_rc = 0;

	assert(src_thread == current_thread());
	assert(dest_thread->sob.type == FLUKE_THREAD);
	assert(src_task != 0);
	assert(dest_task != 0);
	assert(*out_remote_rc == 0);

	INSTRUMENT_EVENT(src_thread /*==current_thread()*/,
			 INSTRUMENT_EVENT_MIGHT_FAULT, src_uva);
	/*
	 * We need to lock both the source and destination mappings;
	 * they might even end up being the same mapping.
	 * First lock the two tasks in the proper order,
	 * or deal with them both at once if necessary.
	 */
	if (src_task == dest_task)
	{
		s_mapping_t *mapping;

		s_task_lock(src_task);

		queue_iterate(&src_task->mappings, mapping,
			      s_mapping_t *, tlink)
		{
			if (src_uva >= mapping->start_va &&
			    src_uva < mapping->start_va + mapping->size)
			{
				assert(src_mapping == 0);
				src_mapping = mapping;
				if (dest_mapping != 0)
					break;
			}
			if (dest_uva >= mapping->start_va &&
			    dest_uva < mapping->start_va + mapping->size)
			{
				assert(dest_mapping == 0);
				dest_mapping = mapping;
				if (src_mapping != 0)
					break;
			}
		}
	}
	else /* src_task != dest_task */
	{
		s_mapping_t *mapping;

		/*
		 * Locking two different tasks,
		 * so be sure to lock them in the correct order.
		 */
		if (src_task < dest_task)
		{
			s_task_lock(src_task);
			s_task_lock(dest_task);
		}
		else
		{
			s_task_lock(dest_task);
			s_task_lock(src_task);
		}

		/*
		 * Look up the appropriate mapping in each one.
		 */
		queue_iterate(&src_task->mappings, mapping,
			      s_mapping_t *, tlink)
		{
			if (src_uva >= mapping->start_va &&
			    src_uva < mapping->start_va + mapping->size)
			{
				assert(src_mapping == 0);
				src_mapping = mapping;
				break;
			}
		}
		queue_iterate(&dest_task->mappings, mapping,
			      s_mapping_t *, tlink)
		{
			if (dest_uva >= mapping->start_va &&
			    dest_uva < mapping->start_va + mapping->size)
			{
				assert(dest_mapping == 0);
				dest_mapping = mapping;
				break;
			}
		}
	}

	if (src_mapping == 0)
	{
		exc_set_pagefault(src_thread, src_uva, src_prot);
		local_rc = KR_PAGE_FAULT;
		goto out1;
	}
	if (dest_mapping == 0)
	{
		exc_set_pagefault(dest_thread, dest_uva, dest_prot);
		*out_remote_rc = KR_PAGE_FAULT;
		goto out1;
	}

	/* Lock the mappings in the correct order.  */
	if (src_mapping < dest_mapping)
	{
		s_mapping_u_lock(src_mapping);
		s_mapping_u_lock(dest_mapping);
	}
	else if (src_mapping > dest_mapping)
	{
		s_mapping_u_lock(dest_mapping);
		s_mapping_u_lock(src_mapping);
	}
	else
		s_mapping_u_lock(src_mapping);

	/* Find an appropriate fragment in each mapping.  */
	src_frag = mapping_frag_locate(src_mapping,
				       src_uva - src_mapping->start_va, 1);
	if ((src_frag == 0) || (frag_prot(src_frag) & src_prot) != src_prot)
	{
		exc_set_pagefault(src_thread, src_uva, src_prot);
		local_rc = KR_PAGE_FAULT;
		goto out2;
	}

	dest_frag = mapping_frag_locate(dest_mapping,
					dest_uva - dest_mapping->start_va, 1);
	if ((dest_frag == 0) || (frag_prot(dest_frag) & dest_prot) != dest_prot)
	{
		exc_set_pagefault(dest_thread, dest_uva, dest_prot);
		*out_remote_rc = KR_PAGE_FAULT;
		goto out2;
	}

	/*
	 * We found fragments for both virtual addresses.
	 * Leave their mapping(s) locked on return,
	 * but unlock the task(s).
	 */
	assert(local_rc == 0);
	assert(*out_remote_rc == 0);
	*out_src_frag = src_frag;
	*out_dest_frag = dest_frag;
	goto out1;

	out2:
	s_mapping_u_unlock(src_mapping);
	if (dest_mapping != src_mapping)
		s_mapping_u_unlock(dest_mapping);
	out1:
	s_task_unlock(src_task);
	if (dest_task != src_task)
		s_task_unlock(dest_task);

#ifdef CONFIG_INSTRUMENT
	if (local_rc)
	{
		INSTRUMENT_EVENT(src_thread /* ==current_thread()*/,
				 local_rc, src_uva);
	}
#endif
	return local_rc;
}

void
obcache_stats_dump(void)
{
#ifdef OBCACHE_STATS
	int i;

	printf("cache: %d lookups (%d cached, %d uncached)\n",
	       ocachestats[0], ocachestats[0]-ocachestats[14], ocachestats[14]);
	printf("  %d hits (bad: %d gone, %d prot, %d corrupt, %d flushed)\n",
	       ocachestats[1], ocachestats[2], ocachestats[3], ocachestats[12],
	       ocachestats[13]);
	printf("  %d misses (%d capacity, %d conflict, %d noob/badprot)\n",
	       ocachestats[0]-ocachestats[1], ocachestats[4], ocachestats[5],
	       (ocachestats[0]-ocachestats[1])-(ocachestats[4]+ocachestats[5]));
	printf("  %d creates (miss: %d capacity, %d conflict)\n",
	       ocachestats[6], ocachestats[7], ocachestats[8]);
	printf("  %d moves (%d source hits, %d dest displacements)\n",
	       ocachestats[9], ocachestats[10], ocachestats[11]);
	for (i = 0; i < OBTAB_SIZE; i++)
		if (ocacheuse[i])
			printf("   %d: %d\n", i, ocacheuse[i]);
#endif
}

#ifdef DEBUG
void
s_task_dump(s_task_t *task, int indent)
{
	s_thread_t *th;
	int i;

	for (i = 0; i < indent; i++) putchar(' ');
	printf("T%p: keeper=%p, threads:\n", task, task->keeper.target);
	queue_iterate(&task->threads, th, s_thread_t *, task_state.task_chain) {
#ifndef VERBOSE_DUMP
		for (i = 0; i < indent+1; i++) putchar(' ');
		printf("thread=%p, ", th);
		if (th->sob.label)
			printf("label=`%s'", th->sob.label);
		else
			printf("no label");
		printf("\n");
#else
		s_debug_dump_flob(&th->sob, indent+1);
#endif
	}

#ifdef VERBOSE_DUMP
	{
		s_mapping_t *mp;
		s_region_t *reg;

		for (i = 0; i < indent; i++) putchar(' ');
		printf("mappings:\n");
		queue_iterate(&task->mappings, mp, s_mapping_t *, tlink)
			s_debug_dump_flob(&mp->sob, indent+1);

		for (i = 0; i < indent; i++) putchar(' ');
		printf("regions:\n");
		queue_iterate(&task->regions, reg, s_region_t *, tlink)
			s_debug_dump_flob(&reg->sob, indent+1);
	}
#endif
}
#endif
