/*
 * Copyright (c) 1996, 1997 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include <fluke/exception.h>
#include <fluke/ipc.h>
#include <fluke/machine/syscall.h>

#include "s_thread.h"
#include "s_pset.h"
#include "s_port.h"
#include "ipc.h"
#include "debug.h"

/*** Client-side system calls ***/

int sys_ipc_send()
{
#if FLASK
	return ipc_send(current_thread(), FALSE);
#else
	return ipc_send(current_thread());
#endif 
}

#if FLASK
int sys_ipc_send_secure()
{
	return ipc_send(current_thread(), TRUE);
}
#endif 

int sys_ipc_call()
{
#if FLASK
	return ipc_call(current_thread(), FALSE);
#else
	return ipc_call(current_thread());
#endif 
}

#if FLASK
int sys_ipc_call_secure()
{
	return ipc_call(current_thread(), TRUE);
}
#endif 

int sys_ipc_client_connect_send()
{
	struct s_thread * const client = current_thread();
	int rc;

#if FLASK
	rc = ipc_client_1_connect_send(client, FALSE, fluke_ipc_client_send__entry);
#else
	rc = ipc_client_1_connect_send(client, fluke_ipc_client_send__entry);
#endif 
	if (rc) return rc;

	thread_wait_other(client, WAIT_IPC_SRV_RECV, client->ipc_state.server);
	return 0;
}

#if FLASK
int sys_ipc_client_connect_send_secure()
{
	struct s_thread * const client = current_thread();
	int rc;

	rc = ipc_client_1_connect_send(client, TRUE, fluke_ipc_client_send__entry);
	if (rc) return rc;

	thread_wait_other(client, WAIT_IPC_SRV_RECV, client->ipc_state.server);
	return 0;
}
#endif 

int sys_ipc_client_ack_send()
{
	struct s_thread * const client = current_thread();
	int rc;

	rc = ipc_client_1_ack_send(client, fluke_ipc_client_send__entry);
	if (rc) return rc;

	thread_wait_other(client, WAIT_IPC_SRV_RECV, client->ipc_state.server);
	return 0;
}

int sys_ipc_client_send()
{
	struct s_thread * const client = current_thread();
	wait_val_t wval;
	int rc;

	rc = ipc_client_1_send(client, &wval);
	if (rc) return rc;

	thread_wait_other(client, wval, client->ipc_state.server);
	return 0;
}

int sys_ipc_client_connect_send_over_receive()
{
	struct s_thread * const client = current_thread();
	int rc;

#if FLASK
	rc = ipc_client_1_connect_send(client, FALSE,
				fluke_ipc_client_send_over_receive__entry);
#else
	rc = ipc_client_1_connect_send(client,
				fluke_ipc_client_send_over_receive__entry);
#endif 
	if (rc) return rc;

	return ipc_client_2_over_receive(client, WAIT_IPC_SRV_RECV);
}

#if FLASK
int sys_ipc_client_connect_send_over_receive_secure()
{
	struct s_thread * const client = current_thread();
	int rc;

	rc = ipc_client_1_connect_send(client, TRUE,
				       fluke_ipc_client_send_over_receive__entry);
	if (rc) return rc;

	return ipc_client_2_over_receive(client, WAIT_IPC_SRV_RECV);
}
#endif 

int sys_ipc_client_ack_send_over_receive()
{
	struct s_thread * const client = current_thread();
	int rc;

	rc = ipc_client_1_ack_send(client,
				fluke_ipc_client_send_over_receive__entry);
	if (rc) return rc;

	return ipc_client_2_over_receive(client, WAIT_IPC_SRV_RECV);
}

int sys_ipc_client_send_over_receive()
{
	struct s_thread * const client = current_thread();
	wait_val_t wval;
	int rc;

	rc = ipc_client_1_send(client, &wval);
	if (rc) return rc;

	return ipc_client_2_over_receive(client, wval);
}

int sys_ipc_client_receive()
{
	return ipc_client_12_receive(current_thread());
}

int sys_ipc_client_disconnect()
{
	return ipc_client_disconnect(current_thread());
}


/*** Server-side system calls ***/

int sys_ipc_wait_receive()
{
#if FLASK
	struct s_thread * const server = current_thread();

	server->ipc_state.server_wants_client_sid = FALSE;	
	return s_pset_wait_receive(server);
#else
	return s_pset_wait_receive(current_thread());
#endif
}

#if FLASK
int sys_ipc_wait_receive_secure()
{
	struct s_thread * const server = current_thread();

	server->ipc_state.server_wants_client_sid = TRUE;
	return s_pset_wait_receive(server);
}
#endif

int sys_ipc_reply()
{
	return ipc_reply(current_thread());
}

int sys_ipc_reply_wait_receive()
{
#if FLASK
	struct s_thread * const server = current_thread();
	int rc = ipc_reply(server);

	if (rc) return rc;
	server->ipc_state.server_wants_client_sid = FALSE;	
	return ipc_server_3_wait_receive(server,
					 fluke_ipc_wait_receive__entry);
#else
	int rc = ipc_reply(current_thread());
	if (rc) return rc;

	return ipc_server_3_wait_receive(current_thread(),
					 fluke_ipc_wait_receive__entry);
#endif
}


#if FLASK
int sys_ipc_reply_wait_receive_secure()
{
	struct s_thread * const server = current_thread();
	int rc;

	server->ipc_state.server_wants_client_sid = TRUE;

	rc = ipc_reply(server);
	if (rc) return rc;

	return ipc_server_3_wait_receive(server, 
					 fluke_ipc_wait_receive_secure__entry);
}
#endif


int sys_ipc_server_ack_send()
{
	struct s_thread * const server = current_thread();
	int rc;

	rc = ipc_server_1_ack_send(server, fluke_ipc_server_send__entry);
	if (rc) return rc;

	thread_wait_other(server, WAIT_IPC_CLI_RECV, server->ipc_state.client);
	return 0;
}

int sys_ipc_server_send()
{
	struct s_thread * const server = current_thread();
	wait_val_t wval;
	int rc;

	rc = ipc_server_1_send(server, &wval);
	if (rc) return rc;

	thread_wait_other(server, wval, server->ipc_state.client);
	return 0;
}

int sys_ipc_server_ack_send_wait_receive()
{
	struct s_thread * const server = current_thread();
	int rc;

#if FLASK
	server->ipc_state.server_wants_client_sid = FALSE;	
#endif

	rc = ipc_server_1_ack_send(server,
				fluke_ipc_server_send_wait_receive__entry);
	if (rc) return rc;

	return ipc_server_2_wait_receive(server, WAIT_IPC_CLI_RECV,
					 fluke_ipc_wait_receive__entry);
}

#if FLASK
int sys_ipc_server_ack_send_wait_receive_secure()
{
	struct s_thread * const server = current_thread();
	int rc;

	server->ipc_state.server_wants_client_sid = TRUE;	
	
	rc = ipc_server_1_ack_send(server,
				fluke_ipc_server_send_wait_receive_secure__entry);
	if (rc) return rc;

	return ipc_server_2_wait_receive(server, WAIT_IPC_CLI_RECV,
					 fluke_ipc_wait_receive_secure__entry);
}
#endif

int sys_ipc_server_ack_send_over_receive()
{
	struct s_thread * const server = current_thread();
	int rc;

	rc = ipc_server_1_ack_send(server,
				fluke_ipc_server_send_over_receive__entry);
	if (rc) return rc;

	return ipc_server_2_over_receive(server, WAIT_IPC_CLI_RECV);
}

int sys_ipc_server_send_wait_receive()
{
	struct s_thread * const server = current_thread();
	wait_val_t wval;
	int rc;

#if FLASK
	server->ipc_state.server_wants_client_sid = FALSE;
#endif

	rc = ipc_server_1_send(server, &wval);
	if (rc) return rc;

	return ipc_server_2_wait_receive(server, wval,
					 fluke_ipc_wait_receive__entry);
}

#if FLASK
int sys_ipc_server_send_wait_receive_secure()
{
	struct s_thread * const server = current_thread();
	wait_val_t wval;
	int rc;
	
	server->ipc_state.server_wants_client_sid = TRUE;

	rc = ipc_server_1_send(server, &wval);
	if (rc) return rc;

	return ipc_server_2_wait_receive(server, wval,
					 fluke_ipc_wait_receive_secure__entry);
}
#endif

int sys_ipc_server_send_over_receive()
{
	struct s_thread * const server = current_thread();
	wait_val_t wval;
	int rc;

	rc = ipc_server_1_send(server, &wval);
	if (rc) return rc;

	return ipc_server_2_over_receive(server, wval);
}

int sys_ipc_server_receive()
{
	return ipc_server_12_receive(current_thread());
}

int sys_ipc_server_disconnect()
{
	return ipc_server_disconnect(current_thread());
}


int sys_ipc_setup()
{
	s_thread_t *server = current_thread();

	server->ipc_state.server_pset_uva =
		EXC_SYSPARAM(server, ipc_setup, pset);
	server->ipc_state.server_esp =
		EXC_SYSPARAM(server, ipc_setup, esp);
	server->ipc_state.server_oneway_eip =
		EXC_SYSPARAM(server, ipc_setup, oneway_eip);
	server->ipc_state.server_idempotent_eip =
		EXC_SYSPARAM(server, ipc_setup, idempotent_eip);
	server->ipc_state.server_reliable_eip =
		EXC_SYSPARAM(server, ipc_setup, reliable_eip);

	return 0;
}

