/*
 * Copyright (c) 1996, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * System calls for manipulating pset objects.
 */

#include <fluke/pset.h>
#include <fluke/machine/syscall.h>

#include "s_thread.h"
#include "s_pset.h"
#include "xfer.h"

int sys_pset_create_hash()
{
	struct s_thread *cur_th = current_thread();
	s_pset_t *new_pset;
	int rc;

	rc = s_pset_create(EXC_SYSPARAM(cur_th, pset_create_hash, pset),
			   EXC_SYSPARAM(cur_th, pset_create_hash, hash),
			   &new_pset);
	if (rc)
		return rc; /* XXX no memory */

	EXC_SYSPARAM(cur_th, pset_create_hash, hash) =
		s_ob_gethash(&new_pset->sob);

	s_pset_unlock(new_pset);

	EXC_SYSPARAM(cur_th, pset_create_hash, rc) = 0;
	return 0;
}

int sys_pset_destroy()
{
	struct s_thread *cur_th = current_thread();
	s_pset_t *pset;
	int rc;

	/*
	 * Look up the pset
	 */
	rc = task_ob_destroy(cur_th, EXC_SYSPARAM(cur_th, pset_destroy, pset),
			     FLUKE_PSET, (s_ob_t **)&pset);
	if (rc) return rc;

	s_pset_destroy(pset);

	return 0;
}

int sys_pset_get_state()
{
	struct s_thread *cur_th = current_thread();
	s_pset_t *pset = 0;
	fluke_pset_state_t state, *statep = 0;
	int rc;

	/*
	 * Determine which values we want
	 */
	if (EXC_SYSPARAM(cur_th, pset_get_state, state))
		statep = &state;

	/*
	 * Lookup the pset
	 */
	rc = task_ob_lookup_type(cur_th,
				 EXC_SYSPARAM(cur_th, pset_get_state, pset),
				 FLUKE_PROT_READ, FLUKE_PSET,
				 (s_ob_t **)&pset);
	if (rc) return rc;

	/*
	 * Get its state
	 */
	rc = s_pset_get_state(pset, statep);
	s_pset_unlock(pset);
	if (rc) return rc;

	/*
	 * Copyout state
	 */
	if (statep)
		rc = copyout_local(statep,
				   EXC_SYSPARAM(cur_th, pset_get_state, state),
				   sizeof(state) / sizeof(oskit_reg_t));

	return rc;
}

int sys_pset_set_state()
{
	struct s_thread *cur_th = current_thread();
	s_pset_t *pset = 0;
	fluke_pset_state_t state, *statep = 0;
	int rc;


	/*
	 * Copy in the simple state structure if there is one.
	 */
	if (EXC_SYSPARAM(cur_th, pset_set_state, state)) {
		rc = copyin_local(EXC_SYSPARAM(cur_th, pset_set_state, state),
				  &state, sizeof(state) / sizeof(oskit_reg_t));
		if (rc) return rc;
		statep = &state;
	}

	/*
	 * Lookup and lock the pset
	 */
	rc = task_ob_lookup_type(cur_th,
				 EXC_SYSPARAM(cur_th, pset_set_state, pset),
				 FLUKE_PROT_WRITE, FLUKE_PSET,
				 (s_ob_t **)&pset);

	if (rc == 0) {
		rc = s_pset_set_state(pset, statep);
		s_pset_unlock(pset);
	}

	return rc;
}

int sys_pset_move()
{
	struct s_thread *cur_th = current_thread();
	s_pset_t *pset;
	int rc;

	rc = task_ob_move(cur_th, EXC_SYSPARAM(cur_th, pset_move, oaddr),
			  EXC_SYSPARAM(cur_th, pset_move, naddr),
			  FLUKE_PSET, FLUKE_PSET_SIZE, 0, (s_ob_t **)&pset);
	if (rc == 0)
		s_pset_unlock(pset);

	return rc;
}

int sys_pset_reference()
{
	struct s_thread *cur_th = current_thread();

	return kentry_ob_reference(cur_th,
				   EXC_SYSPARAM(cur_th, pset_reference, pset),
				   EXC_SYSPARAM(cur_th, pset_reference, ref),
				   FLUKE_PSET);
}
