/*
 * Copyright (c) 1996, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * System calls for manipulating region objects.
 */

#include <oskit/page.h>
#include <fluke/region.h>
#include <fluke/exception.h>
#include <fluke/machine/syscall.h>

#include "s_thread.h"
#include "s_region.h"
#include "s_ref.h"
#include "xfer.h"
#if FLASK
#include "avc.h"
#endif         

int sys_region_create_hash()
{
	struct s_thread *cur_th = current_thread();
	s_region_t *new_region;
	int rc;

	rc = s_region_create(EXC_SYSPARAM(cur_th, region_create_hash, region),
			     EXC_SYSPARAM(cur_th, region_create_hash, hash),
			     &new_region);
	if (rc)
		return rc; /* XXX no memory */

	EXC_SYSPARAM(cur_th, region_create_hash, hash) =
		s_ob_gethash(&new_region->sob);

	s_region_unlock(new_region);

	EXC_SYSPARAM(cur_th, region_create_hash, rc) = 0;
	return 0;
}

int sys_region_destroy()
{
	struct s_thread *cur_th = current_thread();
	s_region_t *region;
	int rc;

	/*
	 * Look up the region
	 */
	rc = task_ob_destroy(cur_th,
			     EXC_SYSPARAM(cur_th, region_destroy, region),
			     FLUKE_REGION, (s_ob_t **)&region);
	if (rc) return rc;

	s_region_destroy(region);

	return (0);
}

int sys_region_protect()
{
	struct s_thread *cur_th = current_thread();
	s_region_t *region;
	int rc;

	rc = task_ob_lookup_type(cur_th,
				 EXC_SYSPARAM(cur_th, region_protect, region),
				 FLUKE_PROT_WRITE, FLUKE_REGION,
				 (s_ob_t **)&region);
	if (rc) return rc;

	s_region_protect(region, EXC_SYSPARAM(cur_th, region_protect, prot));

	s_region_unlock(region);
	return 0;
}

int sys_region_get_state()
{
	struct s_thread *cur_th = current_thread();
	s_region_t *region = 0;
	fluke_region_state_t state, *statep = 0;
	struct link task_link, *task_linkp = 0;
	struct link keeper_link, *keeper_linkp = 0;
	int rc;

	/*
	 * Determine which values we want
	 */
	if (EXC_SYSPARAM(cur_th, region_get_state, state))
		statep = &state;
	if (EXC_SYSPARAM(cur_th, region_get_state, task))
		task_linkp = &task_link;
	if (EXC_SYSPARAM(cur_th, region_get_state, keeper))
		keeper_linkp = &keeper_link;

	/*
	 * Lookup the region
	 */
	rc = task_ob_lookup_type(cur_th,
				 EXC_SYSPARAM(cur_th, region_get_state, region),
				 FLUKE_PROT_READ, FLUKE_REGION,
				 (s_ob_t **)&region);
	if (rc) return rc;

	/*
	 * Get its state
	 */
	rc = s_region_get_state(region, statep, task_linkp, keeper_linkp);
	s_region_unlock(region);
	if (rc) return rc;

	/*
	 * Copyout state
	 */
	if (statep) {
		rc = copyout_local(statep,
				   EXC_SYSPARAM(cur_th, region_get_state, state),
				   sizeof(state) / sizeof(oskit_reg_t));
		if (rc) goto out;
	}
	if (task_linkp)
	{
		rc = kentry_ref_associate(cur_th,
				EXC_SYSPARAM(cur_th, region_get_state, task),
				task_linkp);
		task_linkp = 0;
		if (rc) goto out;
	}
	if (keeper_linkp)
	{
		rc = kentry_ref_associate(cur_th,
				EXC_SYSPARAM(cur_th, region_get_state, keeper),
				keeper_linkp);
		keeper_linkp = 0;
		if (rc) goto out;
	}
out:
	if (task_linkp)
		link_destroy(task_linkp);
	if (keeper_linkp)
		link_destroy(keeper_linkp);
	return rc;
}

int sys_region_set_state()
{
	struct s_thread *cur_th = current_thread();
	s_region_t *region = 0;
	fluke_region_state_t state, *statep = 0;
	struct link task_link, *task_linkp = 0;
	struct link keeper_link, *keeper_linkp = 0;
	int rc;


	/*
	 * Copy in the simple state structure if there is one.
	 */
	if (EXC_SYSPARAM(cur_th, region_set_state, state))
	{
		rc = copyin_local(EXC_SYSPARAM(cur_th, region_set_state, state),
				  &state, sizeof(state) / sizeof(oskit_reg_t));
		if (rc) return rc;
		statep = &state;
	}

	/*
	 * Look up the references to insert into the region.
	 */
	if (EXC_SYSPARAM(cur_th, region_set_state, task))
	{
		rc = kentry_ref_extract_type(cur_th,
				EXC_SYSPARAM(cur_th, region_set_state, task),
				FLUKE_TASK, &task_link);
		if (rc) goto out;
		task_linkp = &task_link;
	}
	if (EXC_SYSPARAM(cur_th, region_set_state, keeper))
	{
		rc = kentry_ref_extract_type(cur_th,
				EXC_SYSPARAM(cur_th, region_set_state, keeper),
				FLUKE_PORT, &keeper_link);
		if (rc) goto out;
		keeper_linkp = &keeper_link;
	}

	/*
	 * Lock the region and task (if any) in the proper order.
	 * XXX what is the proper order?
	 */
	rc = task_ob_lookup_type(cur_th,
				 EXC_SYSPARAM(cur_th, region_set_state, region),
				 FLUKE_PROT_WRITE, FLUKE_REGION,
				 (s_ob_t **)&region);
	if (rc) goto out;

#if FLASK
	if (task_linkp)
	{
	    struct s_ob *sob;


	    sob = link_follow(task_linkp);
	    if (sob)
	    {
		rc = avc_sid_equiv(FLUKE_REGION, 
				   region->sob.sid,
				   FLUKE_TASK, 
				   sob->sid);
		if (rc)
		{
		    s_ob_unlock(sob);
		    goto out;
		}
		s_ob_unlock(sob);
	    }
	}

	if (keeper_linkp)
	{
	    struct s_ob *sob;

	    sob = link_follow(keeper_linkp);
	    if (sob)
	    {
		rc = avc_sid_equiv(FLUKE_REGION,
				   region->sob.sid,
				   FLUKE_PORT,
				   sob->sid); 
		if (rc)
		{
		    s_ob_unlock(sob);
		    goto out;
		}
		s_ob_unlock(sob);
	    }	    
	}	
#endif 

	rc = s_region_set_state(region, statep, task_linkp, keeper_linkp);

	out:
	if (region)
		s_region_unlock(region);
	if (task_linkp)
		link_destroy(&task_link);
	if (keeper_linkp)
		link_destroy(&keeper_link);

	return rc;
}

int sys_region_move()
{
	struct s_thread *cur_th = current_thread();
	s_region_t *region;
	int rc;

	rc = task_ob_move(cur_th, EXC_SYSPARAM(cur_th, region_move, oaddr),
			  EXC_SYSPARAM(cur_th, region_move, naddr),
			  FLUKE_REGION, FLUKE_REGION_SIZE, 0,
			  (s_ob_t **)&region);
	if (rc == 0)
		s_region_unlock(region);
	return (rc);
}

int sys_region_reference()
{
	struct s_thread *cur_th = current_thread();

	return kentry_ob_reference(cur_th,
				   EXC_SYSPARAM(cur_th, region_reference, region),
				   EXC_SYSPARAM(cur_th, region_reference, ref),
				   FLUKE_REGION);
}

/*
 * Search (subset of) a region for kernel objects.
 * If provided array is not large enough for all objects in the range,
 * offset will be adjusted to reflect where the search should continue.
 *
 * XXX note that the target task is not locked for the complete operation.
 * There are a variety of problems that locking would cause.  One is that
 * we would have to release the lock before a copyin or copyout when
 * task == cur_thread->task since faults there will attempt to reacquire
 * the lock.
 */
int sys_region_search()
{
	s_thread_t *cur_th = current_thread();
	s_region_t *region;
	s_task_t *task;
	oskit_addr_t offset, saddr, eaddr;
	oskit_size_t size;
	int rc;
	fluke_obspec_t *uobs;
	unsigned int obcount, uobcount;
	fluke_obspec_t lobs[32];	/* XXX on kstack, keep this small */
	unsigned int lobcount;

	offset = EXC_SYSPARAM(cur_th, region_search, offset);
	size = EXC_SYSPARAM(cur_th, region_search, size);
	uobs = (fluke_obspec_t *)EXC_SYSPARAM(cur_th, region_search, obs);
	uobcount = EXC_SYSPARAM(cur_th, region_search, count);

	/*
	 * No array, zero length array, or no area to search--
	 * these are easy.
	 */
	if (uobs == 0 || uobcount == 0 || size == 0)
		return 0;

	/*
	 * Lookup the region.
	 */
	rc = task_ob_lookup_type(cur_th,
				 EXC_SYSPARAM(cur_th, region_search, region),
				 FLUKE_PROT_READ, FLUKE_REGION,
				 (s_ob_t **)&region);
	if (rc)
		return rc;

	/*
	 * Find the region's task to scan, returning immediately if no task.
	 * XXX insanity trap instead?
	 */
	task = (s_task_t *)link_follow(&region->task);
	if (task == 0) {
		s_region_unlock(region);
		return 0;
	}

	/*
	 * Convert region offset and size into task address range.
	 */
	rc = s_region_range(region, offset, size, &saddr, &eaddr);
	if (rc) {
		s_task_unlock(task);
		s_region_unlock(region);
		return rc;
	}
	s_task_unlock(task);
	s_region_unlock(region);

	/*
	 * Scan the address range.
	 *
	 * XXX we repeatedly fill in a small local object array and copy
	 * that out to user space.
	 */
	obcount = 0;
	while (saddr < eaddr && obcount != uobcount) {

		lobcount = (uobcount - obcount);
		if (lobcount > sizeof(lobs) / sizeof(lobs[0]))
			lobcount = sizeof(lobs) / sizeof(lobs[0]);

		rc = task_ob_search(cur_th, task, &saddr, eaddr,
				    lobs, &lobcount);
		if (rc)
			return rc;

		/*
		 * No more objects found, update the out parameters:
		 * Set user address pointer to end of requested search range.
		 * User count field is already correct.
		 */
		if (lobcount == 0) {
			EXC_SYSPARAM(cur_th, region_search, offset) =
				offset + size;
			EXC_SYSPARAM(cur_th, region_search, size) = 0;
			break;
		}

		/*
		 * Copyout to user
		 */
		rc = copyout_local(lobs, (oskit_addr_t)uobs,
				   lobcount *
				   sizeof(lobs[0]) / sizeof(oskit_reg_t));
		if (rc) {
			return rc;
		}

		uobs += lobcount;
		obcount += lobcount;

		/*
		 * Update out parameters to reflect transfer.
		 */
		EXC_SYSPARAM(cur_th, region_search, offset) =
			offset + size - (eaddr - saddr);
		EXC_SYSPARAM(cur_th, region_search, count) = uobcount - obcount;

		/*
		 * Update the parameters so that we don't rescan everything
		 * after page faults and cancelations.
		 */
		EXC_SYSPARAM(cur_th, region_search, obs) = (unsigned)uobs;
		EXC_SYSPARAM(cur_th, region_search, size) = eaddr - saddr;
	}

	return 0;
}

/*
 * Called from s_thread_exception.
 *
 * If we were doing a region_search and there was a non page fault exception
 * (KR_USER_EXCEPTION) we ignore it and move on.
 *
 * XXX gak, this doesn't work right if we get a KR_USER_EXCEPTION while
 * looking up the region object.
 */
int
sys_region_search_hack(s_thread_t *cur_thread)
{
	oskit_size_t size;

	if (EXC_GET_IP(cur_thread) != fluke_region_search__entry)
		return 1;

	if (cur_thread->exc_state.code == ((FLUKE_EXC_INSANITY << 8) | 0xff))
		return 1;

	size = EXC_SYSPARAM(cur_thread, region_search, size);
	if (size > PAGE_SIZE)
		size = PAGE_SIZE;

	EXC_SYSPARAM(cur_thread, region_search, offset) += size;
	EXC_SYSPARAM(cur_thread, region_search, size) -= size;

	return 0;
}
