/*
 * Copyright (c) 1996, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * System calls for manipulating task objects.
 */

#include <fluke/task.h>
#include <fluke/machine/syscall.h>

#include "s_thread.h"
#include "s_task.h"
#include "s_ref.h"
#include "xfer.h"

#if FLASK
#include "avc.h"
#endif         

int sys_task_create_hash()
{
	struct s_thread *cur_th = current_thread();
	s_task_t *new_task;
	int rc;

	rc = s_task_create(EXC_SYSPARAM(cur_th, task_create_hash, task),
			   EXC_SYSPARAM(cur_th, task_create_hash, hash),
			   &new_task);
	if (rc)
		return rc; /* XXX no memory */

	EXC_SYSPARAM(cur_th, task_create_hash, hash) =
		s_ob_gethash(&new_task->sob);

	s_task_unlock(new_task);

	EXC_SYSPARAM(cur_th, task_create_hash, rc) = 0;
	return 0;
}

int sys_task_destroy()
{
	struct s_thread *cur_th = current_thread();
	s_task_t *task;
	int rc;

	/*
	 * Look up the task
	 */
	rc = task_ob_destroy(cur_th, EXC_SYSPARAM(cur_th, task_destroy, task),
			     FLUKE_TASK, (s_ob_t **)&task);
	if (rc) return rc;

	return s_task_destroy(task);
}

int sys_task_get_state()
{
	struct s_thread *cur_th = current_thread();
	s_task_t *task = 0;
	fluke_task_state_t state, *statep = 0;
	struct link keeper_link, *keeper_linkp = 0;
	int rc;

	/*
	 * Determine which values we want
	 */
	if (EXC_SYSPARAM(cur_th, task_get_state, state))
		statep = &state;
	if (EXC_SYSPARAM(cur_th, task_get_state, keeper))
		keeper_linkp = &keeper_link;

	/*
	 * Lookup the task
	 */
	rc = task_ob_lookup_type(cur_th,
				 EXC_SYSPARAM(cur_th, task_get_state, task),
				 FLUKE_PROT_READ, FLUKE_TASK, (s_ob_t **)&task);
	if (rc) return rc;

	/*
	 * Get its state
	 */
	rc = s_task_get_state(task, statep, keeper_linkp);
	s_task_unlock(task);
	if (rc) return rc;

	/*
	 * Copyout state
	 */
	if (statep) {
		rc = copyout_local(statep,
				   EXC_SYSPARAM(cur_th, task_get_state, state),
				   sizeof(state) / sizeof(oskit_reg_t));
		if (rc) goto out;
	}
	if (keeper_linkp)
	{
		rc = kentry_ref_associate(cur_th,
				EXC_SYSPARAM(cur_th, task_get_state, keeper),
				keeper_linkp);
		keeper_linkp = 0;
		if (rc) goto out;
	}
out:
	if (keeper_linkp)
		link_destroy(keeper_linkp);
	return rc;
}

int sys_task_set_state()
{
	struct s_thread *cur_th = current_thread();
	s_task_t *task = 0;
	fluke_task_state_t state, *statep = 0;
	struct link keeper_link, *keeper_linkp = 0;
	int rc;

	/*
	 * Copy in the simple state structure if there is one.
	 */
	if (EXC_SYSPARAM(cur_th, task_set_state, state))
	{
		rc = copyin_local(EXC_SYSPARAM(cur_th, task_set_state, state),
				  &state, sizeof(state) / sizeof(oskit_reg_t));
		if (rc) return rc;
		statep = &state;
	}

	/*
	 * Look up the references to insert into the task.
	 */
	if (EXC_SYSPARAM(cur_th, task_set_state, keeper))
	{
		rc = kentry_ref_extract_type(cur_th,
				EXC_SYSPARAM(cur_th, task_set_state, keeper),
				FLUKE_PORT, &keeper_link);
		if (rc) goto out;
		keeper_linkp = &keeper_link;
	}

	/*
	 * Lookup and lock the task.
	 */
	rc = task_ob_lookup_type(cur_th,
				 EXC_SYSPARAM(cur_th, task_set_state, task),
				 FLUKE_PROT_WRITE, FLUKE_TASK,
				 (s_ob_t **)&task);
	if (rc) goto out;

#if FLASK
	if (keeper_linkp)
	{
	    struct s_ob *sob;


	    sob = link_follow(keeper_linkp);
	    if (sob)
	    {
		rc = avc_has_permission(SECCLASS_SUBJECT, 
					task->sob.sid, 
					SECCLASS_SUBJECT,
					sob->sid, 
					SUBJECT_SUBJECT__TASK_KEEPER);
		if (rc)
		{
		    s_ob_unlock(sob);
		    goto out;
		}
		s_ob_unlock(sob);
	    }
	}
#endif 

	rc = s_task_set_state(task, statep, keeper_linkp);

	out:
	if (task)
		s_task_unlock(task);
	if (keeper_linkp)
		link_destroy(&keeper_link);

	return rc;
}

int sys_task_move()
{
	struct s_thread *cur_th = current_thread();
	s_task_t *task;
	int rc;

	rc = task_ob_move(cur_th, EXC_SYSPARAM(cur_th, task_move, oaddr),
			  EXC_SYSPARAM(cur_th, task_move, naddr),
			  FLUKE_TASK, FLUKE_TASK_SIZE, 0, (s_ob_t **)&task);
	if (rc == 0)
		s_task_unlock(task);
	return (rc);
}

int sys_task_reference()
{
	struct s_thread *cur_th = current_thread();

	return kentry_ob_reference(cur_th,
				   EXC_SYSPARAM(cur_th, task_reference, task),
				   EXC_SYSPARAM(cur_th, task_reference, ref),
				   FLUKE_TASK);
}
