/*
 * Copyright (c) 1996, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include "dispatch.h"
#include "csw.h"
#include "exception.h"
#include "per_cpu.h"
#include "s_thread.h"
#include "s_pset.h"
#include "s_port.h"
#include "debug.h"
#include "s_instrument.h"
#include <fluke/thread.h>
#include <fluke/ipc.h>
#include <fluke/machine/syscall.h>
#include <stdio.h>
#include <stddef.h>

#include "config_smp.h"
#include "config_instrument.h"
#include "config_fdev.h"


/*#define BASE_DEBUG*/
#ifdef BASE_DEBUG
#define BD_BEGIN() DEBUG_BASE_BEGIN()
#define BD_END() DEBUG_BASE_END()
#define BD() DEBUG_BASE()
#define BD_ARGS(fmt, args...) DEBUG_BASE_ARGS(fmt,args)
#else
#define BD_BEGIN()
#define BD_END()
#define BD()
#define BD_ARGS(args...)
#endif

#ifdef CONFIG_SMP
#define do_smp(foo)	(foo)
#else
#define do_smp(foo)	(void)0
#endif


static void disp_yield(struct preempter *p);


/*
 * Change the server thread's state to indicate its wait-receive
 * completed with a "schedule-me" message on the given port alias.
 */
static inline void
sched_ipc_finish(struct s_thread *server, fluke_alias_t alias)
{
	EXC_SYSPARAM(server, ipc, port_alias) = alias;
	exc_set_min_msg(server, msg);
	IPC_RECV_BUF_WORDS(server) -= FLUKE_MIN_MSG_WORDS;
	IPC_STATUS(server) = FLUKE_IPC_WAIT_ONEWAY;
	EXC_SET_IP(server, fluke_nop__entry);
}


/*
 * First half of readying a thread: if it's donating its CPU,
 * revoke the donation.  Returns nonzero iff readying is complete.
 * On zero return, an IPC to the thread's scheduler port is required.
 */
static int
pd_thread_ready(struct s_thread *thread)
{
	spin_lock(&thread->disp_state.lock);
	do_smp(assert(!thread->disp_state.cpu)); /* no CPU already assigned */

	if (thread->disp_state.donating_to) {
		/*
		 * This thread has control of a real CPU but is
		 * donating his time on that CPU to somebody else.
		 * Tell his inheritor to give the CPU back to him.
		 */

		s_thread_t *revoking, *inheritor;

		revoking = thread;
		goto revoke;	/* see below */

		while (inheritor->disp_state.donating_to) {
			/*
			 * This inheritor thread is donating his
			 * CPU to yet another thread.  We will
			 * cut him out of the inheritance chain
			 * entirely, and have the thread actually
			 * using the CPU dispatch directly back to
			 * the ready thread whose CPU he's using.
			 * This intermediate inheritor thread remains
			 * blocked because he is still donating his
			 * CPU; he will be readied by cond_wakeup
			 * when his donation is revoked by an IPC.
			 */
			do_smp(assert(!inheritor->disp_state.cpu));
			inheritor->disp_state.provider = 0;
			spin_unlock(&revoking->disp_state.lock);
			revoking = inheritor;

		revoke:		/* entry point for first iteration */
			/*
			 * Clear the revoking thread's donation link since
			 * we will be telling his inheritor to yield back
			 * to him.
			 */
			inheritor = revoking->disp_state.donating_to;
			revoking->disp_state.donating_to = 0;

			/*
			 * Iterate on this inheritor to chase down the CPU,
			 * in case he himself donated it to another thread.
			 */
			spin_lock(&inheritor->disp_state.lock);
		}


		/*
		 * Now we have followed the CPU donation chain down to an
		 * inheriting thread that is actually using his CPU right
		 * now (i.e. either the current thread or one running on
		 * another CPU).  (Each thread involved in all but the last
		 * iteration of the loop is still blocked in his own
		 * schedule-wait-receive system call.)
		 *
		 * Tell him to give it back to his highest-up ready provider,
		 * who is the thread_ready argument thread.
		 */
		inheritor->disp_state.provider = thread;

		/*
		 * Tell the CPU to preempt its current thread.
		 */
		register_preempter(&inheritor->disp_state.preempter,
				   disp_yield);

#ifdef CONFIG_SMP
		if (inheritor != current_thread()) {
			/*
			 * This thread is running on another
			 * CPU, so we need to send an IPI.
			 */
			smp_thread_cancel(inheritor->disp_state.cpu);
		}
		else
#endif
		{
#ifndef CONFIG_SMP
			assert(inheritor == current_thread());
#endif
			/*
			 * It's us!  We're using the CPU this thread
			 * donated to us, and now he wants it back!
			 * The pending preemption will make us
			 * switch back to our provider before going
			 * back to user mode.
			 */
		}

		spin_unlock(&inheritor->disp_state.lock);

		return 1;
	}

	assert(!thread->disp_state.provider); /* can't already be running */

	/*
	 * This thread is ready to run but has no CPU to run on.
	 * It will do an IPC on its scheduler port to request one.
	 */

	spin_unlock(&thread->disp_state.lock);

	return 0;
}


/*
 * This thread has unblocked from a wait, so make it solicit a CPU to run it.
 */
void
thread_ready(struct s_thread *thread)
{
	int kpreempt;

	save_preemption_enable(kpreempt);
	disable_preemption();

	if (! pd_thread_ready(thread)) {
		/*
		 * We can't follow thread's scheduler port link here to
		 * complete the IPC, because we might deadlock if this
		 * thread is being woken by its own scheduler port's pset.
		 * Before we return to user mode, run the function to
		 * make this thread do its scheduler-port IPC.
		 */
		static void disp_scheduler_ipc(struct preempter *p);
		register_preempter(&thread->disp_state.preempter,
				   disp_scheduler_ipc);
	}

	restore_preemption_enable(kpreempt);
}

/*
 * This is called from run_preempters to cause a readied thread
 * to do an IPC on its scheduler port.
 */
static void
disp_scheduler_ipc(struct preempter *p)
{
	s_thread_t *thread = (void *)
		((char *) p - offsetof(s_thread_t, disp_state.preempter));

	s_port_t *sched_port;
	s_pset_t *pset;
	fluke_alias_t alias;
	s_thread_t *server;
 ipc:				/* entry point for tail recursion */

	/*
	 * Extract the port reference.
	 */
	sched_port = (s_port_t *) link_follow(thread->sched_state.sched_link);

	if (sched_port) {
		/* Find and lock the port set this port is connected to.  */
		pset = (s_pset_t*)link_follow(&sched_port->pset_link);

		alias = sched_port->alias;
		s_port_unlock(sched_port);
	}
	else {			/* no port same as port with no pset, below */
		return;
	}

	if (pset == 0) {
		/*
		 * This thread has noone to request CPU from.
		 * He'll just sit in limbo until someone decides
		 * to schedule him or give him a scheduler port ref.
		 * Note that this is the normal way for root threads
		 * of non-boot processors to start life.
		 */
		return;
	}

	/* First see if there's already a server thread available.  */
	server = pd_cond_wakeup(&pset->server_cond, 0);
	if (server) {
		s_pset_unlock(pset);

		/*
		 * We found a server thread waiting to hear about sched faults.
		 * Complete its IPC with our CPU request message.
		 */

		sched_ipc_finish(server, alias);

		/*
		 * Now ready the server thread.
		 */
		if (pd_thread_ready(server))
			return;	/* he's ready, we're ready */

		/*
		 * The server needs to do an IPC to his own scheduler!
		 */
		thread = server;
		goto ipc;	/* approximate tail recursion */
	}
	else {
		/*
		 * There is no scheduler available to talk to us.
		 * Put this thread on the wait queue for a scheduler.
		 *
		 * Cache the scheduler port alias to be picked
		 * up in s_thread_schedule_wait_receive by the
		 * server thread that takes us out of scheduler wait.
		 * We don't need to lock this member because it's
		 * only accessed in that one place, that we know
		 * can only happen after this code completes.
		 */
		thread->disp_state.alias = alias;

		spin_lock(&thread->wait_state.lock);

		assert(thread->wait_state.val == WAIT_NONE);

		/*
		 * We always go on the tail of the queue so that
		 * any clients who already have a CPU with which
		 * to deliver their IPC get readied first.
		 * (Since readying us again just gets back here.)
		 */
		wait_queue_add_tail(&pset->client_cond.queue, thread);
		thread->wait_state.waiting_on = &pset->client_cond;
		thread->wait_state.val = WAIT_SCHEDULER;

		thread->wait_state.resume_rc = KR_RESTART;

		spin_unlock(&thread->wait_state.lock);

		s_pset_unlock(pset);
	}
}


/*
 * Final stage of dispatching to a new thread, both threads are locked.
 */
static struct s_thread *
pd_thread_dispatch_finish_locked(
	struct s_thread *cur_thread,
	struct s_thread *next_thread)
{
	/*
	 * The next thread will release the lock on our own thread.
	 * We can't release it ourselves because we'll still be using our stack
	 * clear up until the time when the switch is actually completed.
	 */

	/*
	 * Assign the CPU to the new thread.  We must do this before we
	 * switch into it so that `my_cpu' is defined for interrupt
	 * handlers.  For the same reason, we must leave the old thread's
	 * pointer intact until after we have switched.
	 */
	do_smp(next_thread->disp_state.cpu = cur_thread->disp_state.cpu);

	INSTRUMENT_EVENT(cur_thread,
			 INSTRUMENT_EVENT_SWITCH_OUT, 0);

	/* Switch to the selected next thread. */
	cur_thread = csw_switch(cur_thread, next_thread);

	INSTRUMENT_EVENT(cur_thread,
			 INSTRUMENT_EVENT_SWITCH_IN, 0);
	/*
	 * We were woken up by switching from another thread.
	 * Both threads are still locked.
	 * Just unlock both threads and return.
	 */
	if (cur_thread->disp_state.donating_to) {
		/*
		 * We were donating to a thread and revoked the donation,
		 * so we are switching in from that thread.
		 */
		do_smp(cur_thread->disp_state.donating_to->disp_state.cpu = 0);
		spin_unlock(&cur_thread->disp_state.
			    donating_to->disp_state.lock);
		cur_thread->disp_state.donating_to = 0;
	}
	else {
		/*
		 * We have just been granted control of the processor,
		 * so we are switching in from our CPU provider thread.
		 */
		do_smp(cur_thread->disp_state.provider->disp_state.cpu = 0);
		spin_unlock(&cur_thread->disp_state.provider->disp_state.lock);
	}

	spin_unlock(&cur_thread->disp_state.lock);

	enable_preemption();

	return cur_thread;
}


/*
 * This is the dispatcher, which saves the cur_thread state and
 * runs the next thread on the run queue.
 */

s_thread_t *pd_thread_dispatch(
	struct s_thread		*cur_thread,
	spin_lock_t		*unlock_lock,
	struct s_thread		*inheritor,
	struct link		*inheritor_link)
{
	s_thread_t *provider;
	fluke_error_t rc;

	assert(cur_thread == current_thread());
	assert(inheritor != cur_thread);
	BD_ARGS("cur_thread=%p",cur_thread);

	spin_lock(&cur_thread->disp_state.lock);

	/*
	 * It's now safe to unlock the specified spin lock;
	 * anyone who locks it and tries to wake us up
	 * will encounter our thread's disp_state.lock instead.
	 */
	spin_unlock(unlock_lock);

	if (inheritor_link) {
		assert(!inheritor);
		inheritor = (s_thread_t *) link_consume(inheritor_link);
		if (inheritor) {
			spin_lock(&inheritor->disp_state.lock);
			s_thread_unlock(inheritor);
			goto donate;
		}
		else
			rc = FLUKE_SCHEDULE_BLOCKED;
	}
	else if (inheritor) {
		/*
		 * Donate cur_thread's CPU to inheritor thread.
		 */
		spin_lock(&inheritor->disp_state.lock);
	donate:
		if (inheritor->disp_state.provider) {
			/*
			 * He already has control of a CPU!
			 * Tell our provider we donated in vain.
			 */
			do_smp(assert(inheritor->disp_state.cpu ||
				      inheritor->disp_state.donating_to));
			spin_unlock(&inheritor->disp_state.lock);
			rc = FLUKE_SCHEDULE_INHERITOR_RUNNING;
		}
		else {
			/*
			 * Point us at him and him at us and off he goes.
			 */
			cur_thread->disp_state.donating_to = inheritor;
			inheritor->disp_state.provider = cur_thread;
			return pd_thread_dispatch_finish_locked(cur_thread,
								inheritor);
		}
	}
	else
		rc = FLUKE_SCHEDULE_BLOCKED;


	/*
	 * Yield the CPU back to our provider thread.
	 * We clear our provider link to indicate we
	 * have yielded control of the processor.
	 */
	provider = cur_thread->disp_state.provider;
	cur_thread->disp_state.provider = 0;

	if (!provider) {
	no_provider:
		panic("root thread %p yields CPU %d to the void!",
		      cur_thread,
#ifdef CONFIG_SMP
		      cur_thread->disp_state.cpu->number
#else
		      0
#endif
			);
		return 0;
	}

 yield:				/* to thread in variable `provider' */

	/*
	 * We know this thread cannot be running, because just a moment
	 * ago it was donating its CPU to us; it could not have woken
	 * up and stopped donating to us without taking our disp_state.lock,
	 * which we still hold.  It cannot start running on another CPU
	 * because no thread can donate to it without taking the same lock.
	 */
	do_smp(assert(!provider->disp_state.cpu));

	if (EXC_GET_IP(provider) == fluke_ipc_schedule_wait_receive__entry) {
		/*
		 * This thread explicitly donated its CPU with a system call.
		 * Take it off the pset wait queue it's on.
		 */
		spin_lock(&provider->wait_state.lock);
		pd_thread_wakeup_locked(provider, WAIT_QUEUED_ON_COND);
		spin_unlock(&provider->wait_state.lock);

		/*
		 * Make its syscall return with "inheritor blocked" status.
		 */
		IPC_STATUS(provider) = rc;
		EXC_SET_IP(provider, fluke_nop__entry);
	}
	else {
		/*
		 * This was an implicit CPU donation (e.g. mutex block).
		 * We just restart the donating thread where it left off.
		 */
		wait_val_t wv;
		spin_lock(&provider->wait_state.lock);
		wv = provider->wait_state.val;
		spin_unlock(&provider->wait_state.lock);
		if (wv != WAIT_NONE) {
			/*
			 * The thread donating to us is still blocked,
			 * so there's no point in switching to it;
			 * instead, we'll iterate on to its provider.
			 */
			s_thread_t *const blocked = provider;
			spin_lock(&blocked->disp_state.lock);
			provider = blocked->disp_state.provider;
			blocked->disp_state.provider = 0; /* loses control */
			if (!provider)
				goto no_provider;
			spin_lock(&provider->disp_state.lock);
			spin_unlock(&blocked->disp_state.lock);
			assert(provider->disp_state.donating_to == blocked);
			provider->disp_state.donating_to = cur_thread;
			spin_unlock(&provider->disp_state.lock);

			/*
			 * Yield to this still-blocked thread's provider
			 * and tell him that the thread he chose to schedule
			 * has donated its priority to a blocked thread.
			 */
			rc = FLUKE_SCHEDULE_INHERITOR_BLOCKED;
			goto yield;
		}
	}

	spin_lock(&provider->disp_state.lock);
	assert(provider->disp_state.donating_to == cur_thread);

	return pd_thread_dispatch_finish_locked(cur_thread, provider);
}


/*
 * This is called from run_preempters to cause the current
 * thread to yield the CPU back to its provider thread.
 */
void
disp_yield(struct preempter *p)
{
	spin_lock_t lock = SPIN_LOCK_INITIALIZER; /* XXX ? */
	spin_lock(&lock);
	(void) pd_thread_dispatch(current_thread(), &lock, 0, 0);
}


/*
 * This is the explicit schedule-until-IPC system call.
 * All other types of thread switches are done via pd_thread_dispatch.
 */
int
s_thread_schedule_ipc_wait_receive(
	struct s_thread *cur_th,
	struct s_thread *inheritor)
{
	int rc;
	s_pset_t *pset;
	s_thread_t *client;
	wait_val_t wv;

	/*
	 * Set ourselves up to receive from this pset.
	 * This returns pset locked.
	 */
	rc = s_pset_wait_receive_1(cur_th, &pset);
	if (rc) {
		s_thread_unlock(inheritor);
		return rc;
	}

	disable_preemption();

	/*
	 * Put ourselves on the wait queue for the pset.
	 * We must continue to hold our wait_state lock until
	 * we switch, because client threads might do
	 * cond_wakeup(&pset->client_cond) and try to ready us.
	 */
	spin_lock(&cur_th->wait_state.lock);
	pd_cond_wait_other_locked(&pset->server_cond, cur_th);

	/*
	 * Wake up all client threads waiting for service.
	 */
	while ((client = pd_cond_wakeup(&pset->client_cond, &wv)) != 0) {
		if (wv == WAIT_SCHEDULER) {
			/*
			 * We found a thread blocked on its scheduler
			 * port trying to send us a "schedule me" message.
			 * This thread can't be readied in the normal way,
			 * because that just ends him up right where he is now.
			 *
			 * What we want to do is simply complete our IPC
			 * receive of this thread's "schedule me" message.
			 * We know he wants to talk to us, because
			 * changing his scheduler port or changing
			 * that port's pset would have woken him up.
			 * We know his cached alias is valid, because
			 * changing a port alias wakes up clients waiting
			 * on the port's pset (for exactly this reason).
			 */
#if 1
			s_thread_unlock(inheritor); /* we won't run him */

			sched_ipc_finish(cur_th, client->disp_state.alias);
		unwait:
			/* clear our pset wait */
			pd_thread_wakeup_locked(cur_th, WAIT_QUEUED_ON_COND);
			spin_unlock(&cur_th->wait_state.lock);
			enable_preemption();
			return KR_RESTART;
#else
			 /* Unfortunately, we have to go through all
			 * this rigamarole again to find the port alias.
			 * XXX stash it in new disp_state member??
			 * but then what about changing the alias value??
			 * Perhaps because of this we should declare in
			 * the API that alias values might be sticky as
			 * until you reset the pset; conversely, could
			 * just s_pset_cancel_clients in that case too,
			 * rare enough it should be.
			 */
			link_t pset_link;
			s_port_t *sched_port;
			fluke_alias_t alias;

			/*
			 * Save our pset in a link, since we can't
			 * keep it locked below.
			 */
			link_produce(&pset_link, &pset->sob);
			s_pset_unlock(pset);

			/*
			 * Extract this thread's scheduler port reference.
			 */
			sched_port = (s_port_t *)
				link_follow(client->sched_state.sched_link);

			if (sched_port) {
				/*
				 * Find and lock the port set this port is
				 * connected to.
				 */
				pset = (s_pset_t*)
					link_follow(&sched_port->pset_link);
				alias = sched_port->alias;
				s_port_unlock(sched_port);

				if (link_compare_ptr(&pset_link, &pset->sob)) {
					/*
					 * Lo and behold, he really is
					 * talking to us.  And now we have
					 * the alias value, so we can
					 * complete the IPC.
					 */
					link_destroy(&pset_link);
					s_pset_unlock(pset);

					sched_ipc_finish(cur_th, alias);

				unwait:
					/* clear our pset wait */
					pd_thread_wakeup_locked
						(cur_th, WAIT_QUEUED_ON_COND);
					spin_unlock(&cur_th->wait_state.lock);
					enable_preemption();
					return KR_RESTART;
				}
			}
#endif
		}

		/*
		 * Got a client thread in normal IPC wait (or in scheduler
		 * wait but not on this pset).  Ready him normally.  Note
		 * that if we are indeed servicing his scheduler port, this
		 * will just put him into scheduler wait on our pset, so he
		 * may be chosen by a later iteration of this very loop.
		 */
		thread_ready(client);
	}

	/*
	 * At this point we are enqueued on the pset's server_cond,
	 * and there are no clients waiting on the pset right now.
	 * Now we are ready to dispatch to the requested thread.
	 * We still hold our wait_state lock, and preemption is disabled.
	 */

	spin_lock(&inheritor->wait_state.lock);
	s_thread_unlock(inheritor); /* now that we have the wait_state lock */
	pd_thread_wakeup_locked(inheritor, WAIT_ANYTHING); /* unblock */
	/* XXX this could just check and return FLUKE_SCHEDULE_BLOCKED,
	   instead of restarting him from user mode */
	spin_lock(&inheritor->disp_state.lock);
	spin_unlock(&inheritor->wait_state.lock);
	if (inheritor->disp_state.provider) {
		/*
		 * He already has control of a CPU!
		 * You can't run one thread on two CPUs at once.
		 */
		spin_unlock(&inheritor->disp_state.lock);
		IPC_STATUS(cur_th) = FLUKE_SCHEDULE_RUNNING;
		goto unwait;
	}

	spin_lock(&cur_th->disp_state.lock);

	/*
	 * Now it's finally safe to unlock our wait_state lock,
	 * because any thread trying to ready us will encounter
	 * our disp_state lock.
	 */
	spin_unlock(&cur_th->wait_state.lock);

	/*
	 * Point us at him and him at us and off he goes.
	 */
	cur_th->disp_state.donating_to = inheritor;
	inheritor->disp_state.provider = cur_th;

	(void) pd_thread_dispatch_finish_locked(cur_th, inheritor);

	return KR_RESTART;
}



void thread_disp_state_init(
	struct s_thread		*thread)
{
	BD_ARGS("thread=%p",thread);
	spin_lock_init(&thread->disp_state.lock);
	thread->disp_state.donating_to = 0;
	thread->disp_state.provider = 0;
	do_smp(thread->disp_state.cpu = 0);
	link_init(&thread->disp_state.sched_link);
}




#ifndef CONFIG_NONBLOCKING
static int thread_disp_init_func(void)
{
	s_thread_t *cur_th = current_thread()

	/*
	 * Unlock the current and previous threads
	 * before returning to the actual startup function.
	 */
	BD();
	spin_unlock(&cur_th->disp_state.provider->disp_state.lock);
	spin_unlock(&cur_th->disp_state.lock);

	return 0;
}

void thread_disp_state_setup_stack(
	struct s_thread		*thread,
	void			(*start_func)(void),
	void			*start_sp)
{
	/*
	 * Set up the thread's context switch state
	 * so that the first time we context switch to it,
	 * it'll run thread_disp_init_func()
	 * and then jump to the specified start_func.
	 */
	BD_ARGS("thread=%p",thread);
	thread_csw_state_setup_stack(thread, thread_disp_init_func,
				     start_func, start_sp);
}
#endif /* CONFIG_NONBLOCKING */


void thread_disp_state_destroy(
	struct s_thread		*thread)
{
	/*
	 * We don't actually have to do anything here,
	 * because the thread is being held captive
	 * and is therefore not on any wait queue.
	 */
#ifdef DEBUG
	spin_lock(&thread->disp_state.lock);
	if (thread->disp_state.provider) {
		assert(thread->disp_state.provider == thread);
#ifdef CONFIG_SMP
		panic("destroyed root thread for CPU %d",
		      thread->disp_state.cpu->number);
#else
		otsan();	/* impossible */
#endif
	}
	spin_unlock(&thread->disp_state.lock);
#endif
}

void disp_boot(void)
{

	BD();
}

/* Start a processor running real threads.  Never returns.  */
void
disp_start(struct s_thread *cur_th)
{
	BD();
	assert(cur_th == current_thread());
#ifndef CONFIG_SMP
	assert(cur_th == &startup_thread);
#endif

	/*
	 * Lock against the master CPU still running on this thread.
	 * We know that it held this lock before we were called,
	 * and won't unlock it until it's no longer running cur_th.
	 */
	spin_lock(&cur_th->disp_state.lock);

	/*
	 * The thread has already been assigned a real CPU to run on.
	 * He will be his own CPU provider!
	 */
	cur_th->disp_state.provider = cur_th;

	spin_unlock(&cur_th->disp_state.lock);

	/*
	 * From here to user mode.
	 */
	csw_start(cur_th);

	/* Should never reach here... */
	otsan();
}


#ifdef CONFIG_SMP
/*
 * This is the first thing run on each additional processor's root thread.
 * The user-mode state of the thread is set up already, all we need to
 * do is start the show.
 */
void
smp_disp_boot(struct s_thread *cur_th)
{
	/* This shouldn't return. */
	disp_start(cur_th);
}
#endif CONFIG_SMP
