/*
 * Copyright (c) 1996, 1997 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Simple atomic thread queues.
 * Three atomic operations are supported:
 * add a thread to a queue, 
 * remove the first thread from a queue and return it,
 * and conditionally remove a specific thread from a queue.
 * A thread can only be on one queue at a time.
 */
#ifndef _FLUKE_KERNEL_WAIT_QUEUE_H_
#define _FLUKE_KERNEL_WAIT_QUEUE_H_

#include "spin_lock.h"

#undef LIFO_QUEUEING	/* Define this for LIFO rather than FIFO queueing */

#ifdef LIFO_QUEUEING
#define wait_queue_add		wait_queue_add_head
#define wait_queue_add_locked	wait_queue_add_head_locked
#else
#define wait_queue_add		wait_queue_add_tail
#define wait_queue_add_locked	wait_queue_add_tail_locked
#endif

typedef struct wait_queue
{
	spin_lock_t		lock;
	struct s_thread		*head;
	struct s_thread		**tail;
} wait_queue;


/* Initialize a wait queue to the empty, unlocked state.  */
void wait_queue_init(
	struct wait_queue	*queue);

/* Deinitialize a wait queue.
   Caller must guarantee that no threads are waiting on the wait queue,
   and that the wait queue is unlocked.  */
void wait_queue_destroy(
	struct wait_queue	*queue);

/* Lock a wait queue.
   Caller must worry about locking order dependencies,
   and ensure that interrupts are disabled while any queue is locked.  */
#define wait_queue_lock(queue)		spin_lock(&(queue)->lock)

/* Unlock a wait queue.  */
#define wait_queue_unlock(queue)	spin_unlock(&(queue)->lock)

/*
 * Enqueue a thread onto the head of the specified wait queue.
 * Assumes interrupts are disabled.
 */
void wait_queue_add_head(
	struct wait_queue	*queue,
	struct s_thread		*thread);
void wait_queue_add_head_locked(
	struct wait_queue	*queue,
	struct s_thread		*thread);

/*
 * Enqueue a thread onto the tail of the specified wait queue.
 * Assumes interrupts are disabled.
 */
void wait_queue_add_tail(
	struct wait_queue	*queue,
	struct s_thread		*thread);
void wait_queue_add_tail_locked(
	struct wait_queue	*queue,
	struct s_thread		*thread);

/*
 * Remove and return one thread from the specified wait queue.
 * The returned thread is exclusively owned by the caller
 * until it places the thread on another wait queue
 * or makes it runnable by calling thread_ready() on it.
 * Assumes interrupts are disabled and the queue is locked.
 * Returns NULL if the wait queue is empty.
 */
struct s_thread *wait_queue_remove(
	struct wait_queue	*queue);
struct s_thread *wait_queue_remove_locked(
	struct wait_queue	*queue);

/*
 * Remove a specific thread from the specified wait queue and return the thread.
 * Returns NULL and otherwise does nothing if the thread isn't on this queue.
 * Assumes interrupts are disabled and the queue is locked.
 */
struct s_thread *wait_queue_remove_specific(
	struct wait_queue	*queue,
	struct s_thread		*thread);

#endif /* _FLUKE_KERNEL_WAIT_QUEUE_H_ */

