/*
 * Copyright (c) 1996, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <string.h>

#include <oskit/x86/trap.h>
#include <oskit/x86/eflags.h>
#include <oskit/x86/proc_reg.h>
#include <fluke/exception.h>
#include <fluke/thread.h>

#include "gdt.h"
#include "exception.h"
#include "s_thread.h"

#define EFL_USER_SET	(EFL_IF)
#define EFL_USER_CLEAR	(EFL_IOPL|EFL_NT|EFL_RF)

void exc_set_insanity(s_thread_t *th, int code)
{
	 th->exc_state.code = (FLUKE_EXC_INSANITY << 8) | 0xff; /*XXX*/
	 th->exc_state.subcode = code;
}

void exc_set_pagefault(s_thread_t *th, int fva, fluke_prot_t prot)
{
	 th->exc_state.code =
		 (prot << 16) | (FLUKE_EXC_PAGE_FAULT << 8) | T_PAGE_FAULT;
	 th->exc_state.subcode = fva;
}

void exc_state_dump(struct exc_state *st)
{
	int from_user = (st->cs & 3) || (st->eflags & EFL_VM);

	printf("Dump of trap_state at %p:\n", st);
	printf("EAX %08x EBX %08x ECX %08x EDX %08x\n",
		st->eax, st->ebx, st->ecx, st->edx);
	printf("ESI %08x EDI %08x EBP %08x ESP %08x\n",
		st->esi, st->edi, st->ebp,
		from_user ? st->esp : (unsigned)&st->esp);
	printf("EIP %08x EFLAGS %08x\n", st->eip, st->eflags);
	printf("CS %04x SS %04x\n",
		st->cs & 0xffff, from_user ? st->ss & 0xffff : get_ss());
	printf("exception code %08x, subcode %08x, from %s mode\n",
		st->code, st->subcode, from_user ? "user" : "kernel");

	/* Dump the top of the stack too.  */
	if (!from_user)
		hexdumpw(&st->esp, &st->esp, 128);
}

void thread_exc_state_init(s_thread_t *th)
{
	memset(&th->exc_state, 0, sizeof(th->exc_state));
	th->exc_state.eflags = EFL_USER_SET;
	th->exc_state.cs = USER_CS;
	th->exc_state.ss = USER_DS;

#ifdef CONFIG_KSTACKCHECK
	th->minksp = (unsigned)&th->exc_state;
#endif
}

void thread_exc_get_state(s_thread_t *th, fluke_thread_state_t *state)
{
	state->edi = th->exc_state.edi;
	state->esi = th->exc_state.esi;
	state->ebp = th->exc_state.ebp;
	state->ebx = th->exc_state.ebx;
	state->edx = th->exc_state.edx;
	state->ecx = th->exc_state.ecx;
	state->eax = th->exc_state.eax;
	state->eip = th->exc_state.eip;
	state->eflags = th->exc_state.eflags;
	state->esp = th->exc_state.esp;
}

void thread_exc_set_state(s_thread_t *th, fluke_thread_state_t *state)
{
	th->exc_state.edi = state->edi;
	th->exc_state.esi = state->esi;
	th->exc_state.ebp = state->ebp;
	th->exc_state.ebx = state->ebx;
	th->exc_state.edx = state->edx;
	th->exc_state.ecx = state->ecx;
	th->exc_state.eax = state->eax;
	th->exc_state.eip = state->eip;
	th->exc_state.esp = state->esp;
	th->exc_state.eflags = (state->eflags | EFL_USER_SET) & ~EFL_USER_CLEAR;

#ifndef IOPL0
	{
		/*
		 * Threads running in the kernel task (i.e. kserver threads)
		 * get IOPL 3, so they have full io access.
		 */
#if 0
		extern link_t kernel_task_link;
		if (link_compare(&th->task_state.task_link, &kernel_task_link))
#endif
			th->exc_state.eflags |= EFL_IOPL_USER;
	}
#endif
}
