/*
 * Copyright (c) 1996 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#ifndef _FLUKE_KERNEL_X86_EXCEPTION_H_
#define _FLUKE_KERNEL_X86_EXCEPTION_H_

#include "config_nonblocking.h"
#include "config_kstackcheck.h"

/* Nonblocking and kstackcheck don't work together. */
#if defined(CONFIG_NONBLOCKING) && defined(CONFIG_KSTACKCHECK)
#error --enable-kstackcheck and --enable-nonblocking do not work together.
#endif

/*
 * This structure describes the state of user registers
 * as saved upon kernel trap/interrupt entry.
 * Note that we don't need segment registers or v86-mode stuff
 * because none of that is supported in the standard Fluke environment;
 * emulating other OS's is done in a separate "emulation" environment.
 */
struct exc_state
{
	/* PUSHA register state frame */
	unsigned int	edi;
	unsigned int	esi;
	unsigned int	ebp;
	unsigned int	ebx;
	unsigned int	edx;
	unsigned int	ecx;
	unsigned int	eax;

	/* Fluke exception code and subcode */
	unsigned int	code;
	unsigned int	subcode;

	/* Processor state frame */
	unsigned int	eip;
	unsigned int	cs;
	unsigned int	eflags;
	unsigned int	esp;
	unsigned int	ss;
};

/*
 * The actual exc_state frame pushed by the processor
 * varies in size depending on where the trap came from.
 */
#define EXC_KSIZE	((int)&((struct exc_state*)0)->esp)
#define EXC_USIZE	sizeof(struct exc_state)

/*
 * Macro to allow machine-independent code to access system call parameters.
 * We use preprocessor magic to build the appropriate symbol name
 * defined in <fluke/x86/syscall.h> to evaluate to a register name.
 */
#define EXC_SYSPARAM(cur_thread, call, param)	\
	((cur_thread)->exc_state.fluke_##call##__##param)

/*
 * Machine-independent code uses this macro to set a thread's eip
 * to point to a particular Fluke entrypoint defined in <fluke/x86/syscall.h>.
 */
#define EXC_SET_IP(th, val)	((th)->exc_state.eip = (val))
#define EXC_GET_IP(th)		((oskit_addr_t)(th)->exc_state.eip)


/*
 * The DECLARE_EXC_STATE macro is used to declare the exc_state
 * in the machine-independent s_thread structure.
 * It is needed because, on the x86 and some other architectures,
 * it is very desirable to have the kernel stack right next to the exc_state.
 */
#ifndef CONFIG_NONBLOCKING
/* Process-model (blocking) kernels need a kernel stack per thread.  */

#ifdef CONFIG_KSTACKCHECK

/*
 * The remote debugger is sort of a stack hog, so be conservative.
 * We can reduce this later when we know what the requirements are.
 */
#define THREAD_OVERFLOW_SIZE	500
#define THREAD_STACK_SIZE	3000

/* Stack checking enabled: add a slop zone before the beginning of the stack. */
#define DECLARE_EXC_STATE						\
	unsigned		minksp;					\
	unsigned		stackof[THREAD_OVERFLOW_SIZE/4-4];	\
	unsigned		stack[THREAD_STACK_SIZE/4];		\
	struct exc_state	exc_state;

#define KSTACK_CHECK(th) { \
	unsigned sp; \
	asm volatile("	movl %%esp,%0" : "=r" (sp)); \
	if (sp <= (unsigned)(th)->stack) \
		panic("%s: %d: k-stack overflow 0x%x <= 0x%x", \
		      __FILE__, __LINE__, sp, (th)->stack); \
	if (sp < (th)->minksp) (th)->minksp = sp; \
}

#else /* not CONFIG_KSTACKCHECK */

#define THREAD_STACK_SIZE	3500

/* No stack checking: just declare the stack itself followed by exc_state. */
#define DECLARE_EXC_STATE						\
	unsigned		stack[THREAD_STACK_SIZE/4];		\
	struct exc_state	exc_state;

#define KSTACK_CHECK(th)	/* do nothing */

#endif /* not CONFIG_KSTACKCHECK */

#else /* not CONFIG_NONBLOCKING */

/* Nonblocking model: no per-thread stack needed. */
#define DECLARE_EXC_STATE						\
	struct exc_state	exc_state;

/*
 * Instead, there's a naturally-aligned per-processor stack of this size,
 * with the current thread's address at the high end of the stack (offset 0);
 * see the current_thread() macro in x86/s_thread.h.
 */
#define PROCESSOR_STACK_SHIFT	12
#define PROCESSOR_STACK_SIZE	(1<<PROCESSOR_STACK_SHIFT)

#endif /* not CONFIG_NONBLOCKING */


#include_next "exception.h"

#endif /* _FLUKE_KERNEL_X86_EXCEPTION_H_ */
