/*
 * Copyright (c) 1997, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * This `registers' a client thread for a hardware interrupt.
 */

#include <fluke/ref.h>
#include <fluke/thread.h>
#include "s_thread.h"
#include "s_ref.h"
#include <oskit/dev/dev.h>
#include <oskit/x86/pc/pic.h>	/* pic_disable_irq() */
#include <stdio.h>

#include "prog.h"
#include "intr.h"
#include "debug.h"


void pd_thread_cancel_locked(struct s_thread *thread);


/* Note: interrupt handlers run in supervisor mode */
static void *
kernel_intr_handler(void *data)
{
	s_thread_t *client = data;	/* passed in s_ob... */
	static int count = 0;

	assert(client);
	count++;

	/*
	 * XXX: race condition possible; we don't lock higher-level lock
	 * (client->sob)
	 */
	/* from sys_thread_interrupt */
	if (!(client->gflags & FLUKE_THREAD_INTERRUPT_PENDING)) {

		/*
		 * Flag the pending interrupt and wake up the thread.
		 */
		client->gflags |= FLUKE_THREAD_INTERRUPT_PENDING;

		/*
		 * I need to do a thread_cancel, w/o preemption being
		 * enabled, but only if interrupts are currently enabled.
		 * Note: return from exception & enable exceptions
		 * should do a thread_cancel, as well as set_state
		 * (if necessary).
		 */
		if (client->th_lflags & FLUKE_THREAD_EXCEPTION_ENABLE) {
			spin_lock(&client->wait_state.lock);
			pd_thread_cancel_locked(client);
			spin_unlock(&client->wait_state.lock);
		}

	} else {
		/* The interrupt is going to get lost --
		 * thread hasn't replied to last one yet!
		 */
#if 0
		putchar('!');
#endif
	}
	return NULL;
}

struct my_register_args {
	unsigned irq;
	unsigned flags;
	fluke_ref_t *client_ref;
};


static int
do_my_fdev_register(void *a)
{
	struct my_register_args *args = a;
	s_ref_t *sref;
	struct s_ob *client_thread;
	int rc;

	/*
	 * Now, find the thread object associated by that reference.
	 * this could be done using a region search...
	 * but we're in the kernel now, so I am cheating.
	 */

	/* convert fluke_ref_t to s_ref_t... inspired by kmodule code */
	rc = task_ob_lookup_type(current_thread(),
				 (oskit_addr_t)args->client_ref,
				 FLUKE_PROT_READ|FLUKE_PROT_WRITE,
				 FLUKE_REF, (struct s_ob **)&sref);
	assert(rc == 0);

	/*
	 * XXX egregious hack, take a naked reference to the thread object.
	 * We can't take a link since the interrupt handler runs at too
	 * low of a level. 
 	 */
	client_thread = s_ref_follow(sref);
	assert(client_thread);
	s_ob_incrref(client_thread);
	s_ob_unlock(client_thread);
	s_ob_unlock(&sref->sob);

	irq_alloc(args->irq, kernel_intr_handler, client_thread,
		  1 /* sharable */);
	/* XXX: return code! */

	/* XXX XXX HACK!!! */
	pic_disable_irq(args->irq);

	return 0;
}


int
my_fdev_irq_register(fluke_ref_t *client_thread, unsigned irq, unsigned flags)
{
	struct my_register_args args;

	args.flags = flags & OSENV_IRQ_SHAREABLE;
	args.irq = irq;
	args.client_ref = client_thread;

	do_in_supervisor_mode(do_my_fdev_register, &args);
	/* XXX: return code! */

	return 0;
}


/* ============= */

struct my_unregister_args {
	unsigned irq;
	unsigned flags;
	fluke_ref_t *client_ref;
};

static int
do_my_fdev_unregister(void *a)
{
	struct my_unregister_args *args = a;
	s_ref_t *sref;
	struct s_ob *client_thread;
	int rc;

	rc = task_ob_lookup_type(current_thread(),
				 (oskit_addr_t)args->client_ref,
				 FLUKE_PROT_READ|FLUKE_PROT_WRITE,
				 FLUKE_REF, (struct s_ob **)&sref);
	assert(rc == 0);

	/*
	 * XXX egregious hack phase II.
	 * Blindly dereference the thread, it had better be one of ours!
 	 */
	client_thread = s_ref_follow(sref);
	assert(client_thread);
	s_ob_unlock(client_thread);
	s_ob_decrref(client_thread);
	s_ob_unlock(&sref->sob);

	/* XXX: return code! */
	irq_free(args->irq, kernel_intr_handler, client_thread);

	return 0;
}


int
my_fdev_irq_unregister(fluke_ref_t *client_thread, unsigned irq)
{
	struct my_unregister_args args;

	args.irq = irq;
	args.client_ref = client_thread;

	do_in_supervisor_mode(do_my_fdev_unregister, &args);
	/* XXX: return code! */

	return 0;
}
