/*
 * Copyright (c) 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * This file overrides the -loskit_kern gdb_copyin/gdb_copyout functions
 * with ones that compensate for executing within the kernel segment,
 * so gdb accesses the memory through the user-mode mappings instead.
 * All the physical memory is mapped at the same one-to-one addresses in
 * both segments, but user-mode kserver threads make additional user-only
 * mappings that we'd like to access from gdb.
 */

#include "config_kmodule.h"

#ifndef CONFIG_KMODULE

#include <oskit/types.h>
#include <oskit/gdb_serial.h>
#include "pmap.h"

/*
 * Compensate for the kernel segment's offset of the address,
 * so the user-mode mappings are what gdb sees.
 */
static inline oskit_addr_t
gdb_map_child_address(oskit_addr_t dest, oskit_size_t size)
{
	if (dest >= LIN_KERNEL_MIN)
		dest -= LIN_KERNEL_MIN;

	return dest;
}

/*
 * Called by the GDB stub to copy
 * data into the child's address space.
 */
int
gdb_copyout(const void *src, oskit_addr_t dest, oskit_size_t size)
{
	int err;

	asm volatile ("movl $1f,gdb_trap_recover;"
		      "movl %%ecx,%%eax;"
		      "shrl $2,%%ecx;"
		      "cld;"
		      "rep; movsl;"
		      "movl %%eax,%%ecx;"
		      "andl $3,%%ecx;"
		      "rep; movsb;"
		      "xorl %%eax,%%eax;"
		      "jmp  2f;"
		      "1:"
		      "movl $1,%%eax;"
		      "2:"
		      "movl $0,gdb_trap_recover"
		      : "=a" (err)
		      : "c" (size),
		      "D" (gdb_map_child_address (dest, size)), "S" (src)
		      : "ax", "cx", "di", "si", "memory");

	return (err);
}

/*
 * Called by the GDB stub to copy
 * data from the child task's address space.
 */
int
gdb_copyin(oskit_addr_t src, void *dest, oskit_size_t size)
{
	int err;

	asm volatile ("movl $1f,gdb_trap_recover;"
		      "movl %%ecx,%%eax;"
		      "shrl $2,%%ecx;"
		      "cld;"
		      "rep; movsl;"
		      "movl %%eax,%%ecx;"
		      "andl $3,%%ecx;"
		      "rep; movsb;"
		      "xorl %%eax,%%eax;"
		      "jmp  2f;"
		      "1:"
		      "movl $1,%%eax;"
		      "2:"
		      "movl $0,gdb_trap_recover"
		      : "=a" (err)
		      : "c" (size), "D" (dest),
		      "S" (gdb_map_child_address (src, size))
		      : "ax", "cx", "di", "si", "memory");

	return (err);
}

#endif
