/*
 * Copyright (c) 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * The PC RTC (Real Time Clock) is used for the gprof clock.
 */

#include <stdio.h>
#include <oskit/c/sys/time.h>

#include <oskit/x86/pio.h>
#include <oskit/x86/seg.h>
#include <oskit/x86/pc/pic.h>
#include <oskit/x86/pc/pit.h>
#include <oskit/x86/pc/rtc.h>
#include <oskit/x86/base_trap.h>
#include <oskit/x86/eflags.h>
#include <sys/types.h>

#include "config_profiling.h"
#include "config_user_gprof.h"
#include "per_cpu.h"

#include <sys/gmon.h>
#include "s_profil.h"

#include "spin_lock.h"
#include "exception.h"
#include "intr.h"
#include "preempt.h"
#include "debug.h"
#include "dispatch.h"

#ifdef CONFIG_PROFILING
extern struct link kernel_task_link;
#endif

extern int stathz;
extern int profhz;

extern volatile struct timeval time;

#define TS_USERMODE(ts) USERMODE((ts)->cs, (ts)->eflags)


#include "frag.h"
#include "s_mapping.h"
#include "s_task.h"
#include "s_thread.h"
#include <oskit/page.h>


#ifdef CONFIG_USER_GPROF
/*
 * Note: this assumes that 32-bit counters are used.  Change to ~1 for
 * 16-bit counters.
 */
#define PC_TO_INDEX(pc, prof) \
	((int)(((long long)((pc) - (prof)->pr_off) * \
		(long long)((prof)->pr_scale)) >> 16) & ~3)



void
addupc_task(struct s_task *p, oskit_addr_t pc)
{
	struct uprof *prof;
	oskit_addr_t addr;
	unsigned i;

	prof = p->p_prof;
	assert(prof);

	/* XXX first 16k map to the Fluke entry points */
	if ((pc >= 64*1024*1024) && (pc < 64*1024*1024 + 4*4096)) {
		pc += prof->pr_off - 64*1024*1024;
	}
	if (pc < prof->pr_off ||
	    (i = PC_TO_INDEX(pc, prof)) >= prof->pr_size)
		return;		 /* out of range; ignore */

	addr = prof->buffer + i;
	*(unsigned*)(addr) = *(unsigned*)(addr) + 1;
}
#endif CONFIG_USER_GPROF


#ifdef CONFIG_PROFILING
void
addupc_kernel(oskit_addr_t pc)
{
	int i;
	struct gmonparam *g;

	/*
	 * Kernel statistics are just like addupc_intr, only easier.
	 */
	g = &_gmonparam;
	if (g->state == GMON_PROF_ON) {
		i = pc - g->lowpc;
		if (i < g->textsize) {
			i /= HISTFRACTION * sizeof(*g->kcount);
			g->kcount[i]++;
		}
	}
}
#endif CONFIG_PROFILING

static int
u_clock(void *data)
{
	unsigned char save;
	struct s_task *p;
	struct s_thread *cur_th = current_thread();
	p = cur_th->task_state.task;

	/* This came from user mode, so we can do any kernel calls */
	assert_preemption_enabled();
	assert_interrupts_disabled();

#if (defined CONFIG_PROFILING)
	/* is this in the kserver? */
	if (link_compare(&cur_th->task_state.task_link, &kernel_task_link)) {
		addupc_kernel(cur_th->exc_state.eip);

		/* Kernel and kserver are profiled together */
		p = 0;
	}
#endif

#if (defined CONFIG_USER_GPROF)
	/* task can be NULL for a short window... */
	if (p && (p->p_flag & P_PROFIL)) {
		addupc_task(p, cur_th->exc_state.eip);
	}
#endif

	pic_enable_irq(8);

	save = inb(0x70); inb(0x84);
	/* ACK RTC interrupt */
	outb(0x70, 0x0c); inb(0x84);
	inb(0x71); inb(0x84);
	/* Restore RTC state */
	outb(0x70, save); inb(0x84);

	return 0;
}


static void
k_clock(void *data, struct trap_state *ts)
{
	unsigned char save;

	assert_interrupts_disabled();

#ifdef CONFIG_PROFILING
	assert(!TS_USERMODE(ts));
	addupc_kernel(ts->eip);
#endif
	/* XXX: clock ticks get counted twice for kserver system calls... */
#ifdef CONFIG_USER_GPROF
	{
		static struct preempter gprof_preepter;
		struct s_task *p;
		struct s_thread *cur_th = current_thread();
		p = cur_th->task_state.task;

		if (p && (p->p_flag & P_PROFIL)) {
			addupc_task(p, cur_th->exc_state.eip);
		}
	}
#endif CONFIG_USER_GPROF

	pic_enable_irq(8);

	save = inb(0x70); inb(0x84);
	/* ACK RTC interrupt */
	outb(0x70, 0x0c); inb(0x84);
	inb(0x71); inb(0x84);
	/* Restore RTC state */
	outb(0x70, save); inb(0x84);
}

void
prof_clock_init(void)
{
	static int init = 0;

	if (init) {
		printf("re-initing prof clock\n");
		pic_enable_irq(8);
		enable_interrupts();
	}
	init = 1;
	disable_interrupts();

	irq_alloc_ku(8, k_clock, u_clock, NULL);

	disable_interrupts();

	/*
	 * Control Register B determines the power-of-2 frequency
	 * at which the RTC interrupts.
	 */
	outb(0x70, 0x0a); inb(0x84);
	outb(0x71, 0x23); inb(0x84);	/* 29 = 128 Hz; 23 = 8192 */
	outb(0x70, 0x0b); inb(0x84);
	outb(0x71, 0x42); inb(0x84);	/* periodic interrupt */
	outb(0x70, 0x0c); inb(0x84);
	inb(0x71);			/* clear interrupt flag */

	/* irq_alloc_ku already did this ... */
	enable_interrupts();
}

