/*
 * Copyright (c) 1996 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Hardware interrupt management for uniprocessor x86 PC's.
 */

#include "dispatch.h"
#include "cancel.h"
#include "intr.h"

#define MIN_VECT	32
#define MAX_VECT	255
#define NUM_VECTS	(MAX_VECT - MIN_VECT + 1)

struct intr
{
	struct wait_queue queue;

	/* True if an interrupt is pending.
	   This flag is protected by the queue's lock.  */
	int pending;
};
static struct intr intr[NUM_VECTS];

/*
 * This routine is called whenever a hardware interrupt occurs,
 * either from user mode or from supervisor mode.
 * It disables the interrupt in the PIC,
 * then wakes up the thread waiting for the appropriate interrupt.
 * Assumes the processor interrupt flag is disabled.
 */
void interrupt(unsigned irq)
{
	struct intr *i = &intr[irq - MIN_VECT];
	assert(irq >= MIN_VECT);
	assert(irq <= MAX_VECT);

	/* Acknowledge the interrupt.
	   This must be done now rather than later in interrupt_wait(),
	   because we may not actually switch immediately
	   to the thread that woke up to handle the interrupt.
	   XXX problem: what about level-triggered interrupts?  */
	apic_eoi(...);

	/* Wake up any thread waiting for the interrupt.  */
	wait_queue_lock(&i->queue);
	wait_queue_wakeup(&i->queue);
	wait_queue_unlock(&i->queue);
}

int interrupt_wait(unsigned irq)
{
	struct thread *th = current_thread();
	struct intr *i = &intr[irq - MIN_VECT];
	int rc;

	assert(irq >= MIN_VECT);
	assert(irq <= MAX_VECT);

	disable_interrupts();

	/* Assert that we are about to do a cancelable wait
	   on the hardware interrupt's wait queue.  */
	assert(irq >= MIN_VECT);
	assert(irq <= MAX_VECT);
	rc = assert_cancelable_wait(th, &i->queue);
	if (rc)
	{
		enable_interrupts();
		return rc;
	}

	/* See if an interrupt is already pending.
	   If so, return immediately.  */
	if (i->pending)
	{
		wait_queue_unlock(&i->queue);
		XXX undo_cancelable_wait
		return 0;
	}

	/* Go to sleep waiting on the appropriate condition.  */
	rc = do_cancelable_wait(th, &i->queue);

	enable_interrupts();

	return rc;
}

