/*
 * Copyright (c) 1996 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include <oskit/x86/asm.h>
#include <oskit/x86/eflags.h>
#include <oskit/x86/seg.h>
#include <oskit/x86/gate_init.h>
#include <oskit/x86/base_gdt.h>
#include <oskit/x86/anno.h>

#include "x86.h"
#include "intr.h"
#include "idt.h"
#include "s_thread_asm.h"
#include "config_gdb_serial.h"
#include "config_opt_threads.h"
#include "debug.h"

#include "config_nonblocking.h"
	
	.text

/*
 * Macro to load the standard segment registers
 * on entry to a hardware interrupt handler.
 * The specified scratch register will be trashed.
 */
#define LOAD_SEGS(scratch)						\
	movl	$KERNEL_DS,%e##scratch					;\
	movw	%scratch,%ds						;\
	movw	%scratch,%es


#ifdef CONFIG_NONBLOCKING

/*
 * Interrupt entry template for the nonblocking kernel.
 * On interrupts coming in from user mode,
 * we have to save our state in the thread control block,
 * which is separate from the kernel stack.
 */
#define INTERRUPT(n)							\
	GATE_ENTRY(n + IDT_IRQ_BASE, 0f, ACC_PL_K|ACC_INTR_GATE)	;\
	P2ALIGN(TEXT_ALIGN)						;\
0:	testb	$3,4(%esp)						;\
	jz	1f							;\
	pushl	%ebx							;\
	LOAD_SEGS(bx)							;\
	movl	24-PROCESSOR_STACK_SIZE(%esp),%ebx /* cur thread */	;\
	movl	%eax,TH_EXC_STATE+EXC_EAX(%ebx)				;\
	movl	$(n),%eax						;\
	jmp	all_u_intrs						;\
1:	pushl	$(n)							;\
	pushl	$(n) + IDT_IRQ_BASE					;\
	pushl   %eax							;\
	movl    $(n),%eax						;\
	jmp	all_k_intrs

#else /* not CONFIG_NONBLOCKING */

/*
 * Interrupt entry template for the normal blocking kernel.
 * On interrupts coming in from user mode,
 * we need to push the standard exc_state frame.
 * On interrupts from kernel mode,
 * we push the oskit's trap_state structure instead.
 */
#define INTERRUPT(n)							\
	GATE_ENTRY(n + IDT_IRQ_BASE, 0f, ACC_PL_K|ACC_INTR_GATE)	;\
	P2ALIGN(TEXT_ALIGN)						;\
0:	pushl	$(n)							;\
	pushl	$-1							;\
	pushl   %eax							;\
	movl    $(n),%eax						;\
	testb	$3,16(%esp)						;\
	jnz	all_u_intrs						;\
	jmp	all_k_intrs

#endif /* not CONFIG_NONBLOCKING */

GATE_INITTAB_BEGIN(base_int_inittab)

INTERRUPT(0)
INTERRUPT(1)
INTERRUPT(2)
INTERRUPT(3)
INTERRUPT(4)
INTERRUPT(5)
INTERRUPT(6)
INTERRUPT(7)
INTERRUPT(8)
INTERRUPT(9)
INTERRUPT(10)
INTERRUPT(11)
INTERRUPT(12)
INTERRUPT(13)
INTERRUPT(14)
INTERRUPT(15)

GATE_INITTAB_END




/* XXX handle NMI - at least print a warning like Linux does.  */


/*
 * All normal interrupts from supervisor mode enter here.
 * We push the standard oskit trap_state frame,
 * so that things like kernel debugging and profiling can be done easily.
 * Don't have to push eax; it's already been done for us.
 * (Currently all interrupt handlers are "normal" interrupt handlers,
 * but in the future there may be special "quick" interrupt handlers
 * that use their own custom streamlined entry code.)
 * Note that we have to save and restore all registers including cr2,
 * because, for example, we might preempt the first part of t_page_fault.
 */
	ALIGN_TEXT
all_k_intrs:
	pushl   %ecx
	pushl   %edx
	pushl   %ebx
	movl	%cr2,%edx
	pushl	%edx
	pushl   %ebp
	pushl   %esi
	pushl   %edi
	pushl   %ds
	pushl   %es
	pushl   %fs
	pushl   %gs

	LOAD_SEGS(dx)
	CLEAR_EBP

	/* Call the kernel interrupt handler */
	pushl 	%eax
	call	EXT(k_interrupt)
	addl	$4,%esp

	jmp	EXT(return_from_k_trap)


/*
 * All normal interrupts from user mode enter here.
 * Don't have to save eax; it's already been done for us.
 */
	ALIGN_TEXT
all_u_intrs:

#ifdef CONFIG_NONBLOCKING

	/* Save the general registers in the thread state */
	movl	%ecx,TH_EXC_STATE+EXC_ECX(%ebx)
	movl	%edx,TH_EXC_STATE+EXC_EDX(%ebx)
	popl	TH_EXC_STATE+EXC_EBX(%ebx)
	movl	%ebp,TH_EXC_STATE+EXC_EBP(%ebx)
	movl	%esi,TH_EXC_STATE+EXC_ESI(%ebx)
	movl	%edi,TH_EXC_STATE+EXC_EDI(%ebx)

	/* Move important stuff from the kernel stack to the thread state */
	movl	(%esp),%esi
	movl	8(%esp),%ecx
	movl	12(%esp),%edx
	movl	%esi,TH_EXC_STATE+EXC_EIP(%ebx)
	movl	%ecx,TH_EXC_STATE+EXC_EFLAGS(%ebx)
	movl	%edx,TH_EXC_STATE+EXC_ESP(%ebx)

#else /* not CONFIG_NONBLOCKING */

	pushl   %ecx
	pushl   %edx
	pushl   %ebx
	pushl   %ebp
	pushl   %esi
	pushl   %edi

	LOAD_SEGS(dx)

#endif /* not CONFIG_NONBLOCKING */

	CLEAR_EBP

	/* push interrupt number! */
	pushl 	%eax
	call	EXT(u_interrupt)
	addl	$4,%esp

	sti	/* return_to_user assumes that interrupts are enabled */
	jmp	EXT(return_to_user)

