/*
 * Copyright (c) 1996 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include <fluke/ipc.h>
#include <fluke/thread.h>
#include <fluke/exception.h>
#include <fluke/machine/syscall.h>

#include "exception.h"
#include "xfer.h"
#include "s_port.h"
#include "ipc.h"
#include "s_thread.h"
#include "debug.h"

#if FLASK
#include "avc.h"
#endif 

void thread_ipc_get_state_md(s_thread_t *th, fluke_thread_state_t *state)
{
	state->min_msg[0] = th->ipc_state.md.min_msg[0];
	state->min_msg[1] = th->ipc_state.md.min_msg[1];
}

void thread_ipc_set_state_md(s_thread_t *th, fluke_thread_state_t *state)
{
	th->ipc_state.md.min_msg[0] = state->min_msg[0];
	th->ipc_state.md.min_msg[1] = state->min_msg[1];
}

#if FLASK
int ipc_send(struct s_thread *const client, oskit_bool_t secure)
#else
int ipc_send(struct s_thread *const client)
#endif 
{
	struct s_thread *server;
	fluke_ipc_params_t params;
	fluke_ipc_word_t min_msg[FLUKE_MIN_MSG_WORDS];
	int rc;
#if FLASK
	security_id_t client_sid;
#endif 


	/*
	 * Read the IPC send parameters and the minimum message words.
	 */
	rc = copyin_local(EXC_SYSPARAM(client, ipc, params),
			  &params, sizeof(params) / sizeof(fluke_ipc_word_t));
	if (rc) return rc;
	if (params.send_buf.words < FLUKE_MIN_MSG_WORDS)
		EXC_RETURN_INSANITY(BUFFER_TOO_SMALL);

	rc = copyin_local((oskit_addr_t)params.send_buf.buf,
			  min_msg, FLUKE_MIN_MSG_WORDS);
	if (rc) return rc;
	params.send_buf.words -= FLUKE_MIN_MSG_WORDS;
	params.send_buf.buf += FLUKE_MIN_MSG_SIZE;

#if FLASK
	if (secure)
		client_sid = EXC_SYSPARAM(client, ipc, client_sid);
	else
		client_sid = client->sob.sid;

	/* Capture a server thread. */
	rc = s_port_capture(client, SUBJECT_SUBJECT__SEND, 
			    client_sid, &server);
#else
	/* Capture a server thread. */
	rc = s_port_capture(client, &server);
#endif
	if (rc)
		return rc;

	/* Extract the receive parameters from the server thread state. */
	params.recv_buf.words =	IPC_RECV_BUF_WORDS(server) -
							FLUKE_MIN_MSG_WORDS;
	params.recv_buf.buf =	(fluke_ipc_word_t*)IPC_RECV_BUF_BUF(server) +
							FLUKE_MIN_MSG_WORDS;
	params.recv_buf_count =	IPC_RECV_BUF_COUNT(server);
	params.recv_buf_tab =	(fluke_ipc_buf_t*)IPC_RECV_BUF_TAB(server);
	params.recv_ref_count =	IPC_RECV_REF_COUNT(server);
	params.recv_ref_tab =	(struct fluke_ref**)IPC_RECV_REF_TAB(server);

#if FLASK
	if (server->ipc_state.server_wants_client_sid)
	{
	    min_msg[0] = client_sid;
	    min_msg[1] = 0xdeadbeef;
	    params.send_buf.words += FLUKE_MIN_MSG_WORDS;
	    params.send_buf.buf -= FLUKE_MIN_MSG_SIZE;
	    params.recv_buf.words += FLUKE_MIN_MSG_WORDS;
	    params.recv_buf.buf -= FLUKE_MIN_MSG_SIZE;
	}
#endif

	/* Speculatively transfer our request message to the server. */
	rc = ipc_speculative_transfer(client, server, &params, &params);
	if (rc)
		return rc;

	/* The transfer succeeded, so update the receiver's register state. */
	IPC_RECV_BUF_WORDS(server) = params.recv_buf.words;
	IPC_RECV_BUF_COUNT(server) = params.recv_buf_count;
	IPC_RECV_REF_COUNT(server) = params.recv_ref_count;
	server->ipc_state.md.min_msg[0] = min_msg[0];
	server->ipc_state.md.min_msg[1] = min_msg[1];
	assert(FLUKE_MIN_MSG_WORDS == 2);
	IPC_FINISH_RECEIVE(server, FLUKE_IPC_WAIT_ONEWAY);
	thread_ready(server);

	IPC_STATUS(client) = 0;
	return 0;
}


#if FLASK
int ipc_call(struct s_thread *const client, oskit_bool_t secure)
#else
int ipc_call(struct s_thread *const client)
#endif 
{
	struct s_thread *server;
	fluke_ipc_params_t sparams;
	fluke_ipc_word_t min_msg[FLUKE_MIN_MSG_WORDS];
	int rc;
#if FLASK
	security_id_t client_sid;
#endif 


	/* Break any existing idempotent connection. */
	rc = ipc_client_idemp_disconnect(client);
	if (rc)
		return rc;
	assert(client->ipc_state.idemp_server == 0);

	/*
	 * Read the IPC send and recv parameters and the minimum message words.
	 */
	rc = copyin_local(EXC_SYSPARAM(client, ipc, params),
			  &client->ipc_state.md.params,
			  sizeof(fluke_ipc_params_t) / sizeof(fluke_ipc_word_t));
	if (rc) return rc;
	if (client->ipc_state.md.params.send_buf.words < FLUKE_MIN_MSG_WORDS)
		EXC_RETURN_INSANITY(BUFFER_TOO_SMALL);

	rc = copyin_local((oskit_addr_t)client->ipc_state.md.params.send_buf.buf,
			  min_msg, FLUKE_MIN_MSG_WORDS);
	if (rc) return rc;
	client->ipc_state.md.params.send_buf.words -= FLUKE_MIN_MSG_WORDS;
	client->ipc_state.md.params.send_buf.buf += FLUKE_MIN_MSG_SIZE;

#if FLASK
	if (secure)
		client_sid = EXC_SYSPARAM(client, ipc, client_sid);
	else
		client_sid = client->sob.sid;

	/* Capture a server thread. */
	rc = s_port_capture(client, SUBJECT_SUBJECT__CALL, 
			    client_sid, &server);
#else
	/* Capture a server thread. */
	rc = s_port_capture(client, &server);
#endif
	if (rc)
		return rc;

	/* Extract the receive parameters from the server thread state. */
	sparams.recv_buf.words = IPC_RECV_BUF_WORDS(server) -
							FLUKE_MIN_MSG_WORDS;
	sparams.recv_buf.buf = (fluke_ipc_word_t*)IPC_RECV_BUF_BUF(server) +
							FLUKE_MIN_MSG_WORDS;
	sparams.recv_buf_count = IPC_RECV_BUF_COUNT(server);
	sparams.recv_buf_tab = (fluke_ipc_buf_t*)IPC_RECV_BUF_TAB(server);
	sparams.recv_ref_count = IPC_RECV_REF_COUNT(server);
	sparams.recv_ref_tab = (struct fluke_ref**)IPC_RECV_REF_TAB(server);

#if FLASK
	if (server->ipc_state.server_wants_client_sid)
	{
	    min_msg[0] = client_sid;
	    min_msg[1] = 0xdeadbeef;
	    client->ipc_state.md.params.send_buf.words += FLUKE_MIN_MSG_WORDS;
	    client->ipc_state.md.params.send_buf.buf -= FLUKE_MIN_MSG_SIZE;
	    sparams.recv_buf.words += FLUKE_MIN_MSG_WORDS;
	    sparams.recv_buf.buf -= FLUKE_MIN_MSG_SIZE;
	}
#endif

	/* Speculatively transfer our request message to the server. */
	rc = ipc_speculative_transfer(client, server,
				      &client->ipc_state.md.params, &sparams);
	if (rc)
		return rc;

	/* The transfer succeeded, so update the receiver's register state. */
	client->ipc_state.idemp_server = server;
	server->ipc_state.idemp_client = client;
	IPC_RECV_BUF_WORDS(server) = sparams.recv_buf.words;
	IPC_RECV_BUF_COUNT(server) = sparams.recv_buf_count;
	IPC_RECV_REF_COUNT(server) = sparams.recv_ref_count;
	server->ipc_state.md.min_msg[0] = min_msg[0];
	server->ipc_state.md.min_msg[1] = min_msg[1];
	assert(FLUKE_MIN_MSG_WORDS == 2);
	IPC_FINISH_RECEIVE(server, FLUKE_IPC_WAIT_IDEMPOTENT);
	return thread_handoff(client, WAIT_IPC_CALL, server);
}


int ipc_reply(struct s_thread *const server)
{
	struct s_thread *client = server->ipc_state.idemp_client;
	wait_val_t wval;
	fluke_ipc_params_t sparams;
	fluke_ipc_word_t min_msg[FLUKE_MIN_MSG_WORDS];
	int rc;
#if FLASK
#ifdef DEBUG
	static fluke_ipc_word_t security_fault_reply_buf[301];
#else
	fluke_ipc_word_t security_fault_reply_buf[4];
#endif 
#define SFRB_SZ (sizeof(security_fault_reply_buf)/sizeof(fluke_ipc_word_t))
#endif 

	assert(server == current_thread());
	assert(client == 0 || client->ipc_state.idemp_server == server);

	IPC_STATUS(server) = 0;
	if (client == 0) {
		/*
		 * The idempotent IPC was probably interrupted for some reason,
		 * so just throw away the reply and return success.
		 */
		return 0;
	}

	/* Capture the client thread. */
	wval = thread_wakeup(client, WAIT_IPC_IDEMPOTENT);
	if (wval == 0) {
		/*
		 * Client thread wasn't still waiting on the call,
		 * so it was probably interrupted somehow.
		 * Again, just discard the reply;
		 * we will disconnect at the next wait_receive().
		 */
		return 0;
	}
	if (wval == WAIT_IPC_EXCEPTION) {
		/*
		 * This call was caused by an exception;
		 * Need to read the minimum message to see if the exception
		 * handler succeeded.  If not we arrange to return a user
		 * exception.  In either case we restart the client thread.
		 */
		rc = copyin_local(IPC_SEND_BUF_BUF(server), min_msg,
				  FLUKE_MIN_MSG_WORDS);
		if (rc) {
			thread_ready(client);
			return rc;
		}
		if (min_msg[0])
			client->wait_state.resume_rc = KR_USER_EXCEPTION;
		client->ipc_state.idemp_server = 0;
		server->ipc_state.idemp_client = 0;
		thread_ready(client);
		return 0;
	}
#if FLASK
	if (wval == WAIT_IPC_SECURITY) {
		/*
		 * This call was caused by a security fault;
		 * Need to read the minimum message to see if the exception
		 * handler succeeded.  If not we arrange to return a user
		 * exception.  In either case we restart the client thread.
		 */
		rc = copyin_local(IPC_SEND_BUF_BUF(server), min_msg,
				  FLUKE_MIN_MSG_WORDS);
		if (rc) {
			thread_ready(client);
			return rc;
		}
		assert(min_msg[0] == 0);
		if (min_msg[0])
		{
		    exc_set_insanity(client, FLUKE_INSANITY_PERMISSION_DENIED);
		    client->wait_state.resume_rc = KR_USER_EXCEPTION;
		}
		else
		{
		    /* Read the rest of the reply */
		    security_fault_reply_buf[0] = min_msg[1];
		    assert(FLUKE_MIN_MSG_WORDS == 2);

		    sparams.send_buf.words = IPC_SEND_BUF_WORDS(server) -
			    FLUKE_MIN_MSG_WORDS;
		    sparams.send_buf.buf = (fluke_ipc_word_t*)IPC_SEND_BUF_BUF(server) +
			    FLUKE_MIN_MSG_WORDS;

		    if (sparams.send_buf.words > SFRB_SZ-1)
		    {
			/* truncate reply if necessary */
			printf("ipc_reply():  truncating security reply\n");
			sparams.send_buf.words = SFRB_SZ-1;
		    }

		    rc = copyin_local((oskit_addr_t)sparams.send_buf.buf, 
				      &security_fault_reply_buf[1], 
				      sparams.send_buf.words);
		    if (rc) {
			thread_ready(client);
			return rc;
		    }
		    rc = avc_update_cache(client->sclass, 
					  client->ssid,	
					  client->tclass,
					  client->tsid,
					  security_fault_reply_buf,
					  sparams.send_buf.words+1);
		    assert(rc == 0);
		    if (rc)
		    {
			exc_set_insanity(client, FLUKE_INSANITY_PERMISSION_DENIED);
			client->wait_state.resume_rc = KR_USER_EXCEPTION;
		    }
		}
		
		client->ipc_state.idemp_server = 0;
		server->ipc_state.idemp_client = 0;
		thread_ready(client);
		return 0;
	}
#endif 
	assert(wval == WAIT_IPC_CALL);

	/* Adjust the client's message params to skip the minimum message. */
	client->ipc_state.md.params.recv_buf.words -= FLUKE_MIN_MSG_WORDS;
	client->ipc_state.md.params.recv_buf.buf += FLUKE_MIN_MSG_SIZE;

	/* Extract the send parameters from the server's register state.  */
	sparams.send_buf.words = IPC_SEND_BUF_WORDS(server) -
							FLUKE_MIN_MSG_WORDS;
	sparams.send_buf.buf = (fluke_ipc_word_t*)IPC_SEND_BUF_BUF(server) +
							FLUKE_MIN_MSG_WORDS;
	sparams.send_buf_count = IPC_SEND_BUF_COUNT(server);
	sparams.send_buf_tab = (fluke_ipc_buf_t*)IPC_SEND_BUF_TAB(server);
	sparams.send_ref_count = IPC_SEND_REF_COUNT(server);
	sparams.send_ref_tab = (struct fluke_ref**)IPC_SEND_REF_TAB(server);

	/* Read the server's minimum message to send. */
	rc = copyin_local(IPC_SEND_BUF_BUF(server), min_msg, FLUKE_MIN_MSG_WORDS);
	if (rc) {
		thread_ready(client);
		return rc;
	}

	/*
	 * Speculatively transfer our reply message to the client.
	 * If this gets interrupted, the client will restart from scratch.
	 */
	rc = ipc_speculative_transfer(server, client, &sparams,
				      &client->ipc_state.md.params);
	if (rc)
		return rc;

	/* The transfer succeeded, so disconnect and release the client. */
	client->ipc_state.idemp_server = 0;
	server->ipc_state.idemp_client = 0;
	IPC_RECV_BUF_WORDS(client) = client->ipc_state.md.params.recv_buf.words;
	IPC_RECV_BUF_COUNT(client) = client->ipc_state.md.params.recv_buf_count;
	IPC_RECV_REF_COUNT(client) = client->ipc_state.md.params.recv_ref_count;
	client->ipc_state.md.min_msg[0] = min_msg[0];
	client->ipc_state.md.min_msg[1] = min_msg[1];
	assert(FLUKE_MIN_MSG_WORDS == 2);
	IPC_FINISH_RECEIVE(client, 0);
	thread_ready(client);
	return 0;
}

#if FLASK
int ipc_client_1_connect_send(struct s_thread *const client,
			      oskit_bool_t secure,
			      oskit_addr_t connected_ip)
#else
int ipc_client_1_connect_send(struct s_thread *const client,
			      oskit_addr_t connected_ip)
#endif 
{
	struct s_thread *server;
	fluke_ipc_params_t client_ipc;
	fluke_ipc_word_t min_msg[FLUKE_MIN_MSG_WORDS];
	int rc;
#if FLASK
	security_id_t client_sid;
#endif 


	rc = ipc_client_disconnect(client);
	if (rc) return rc;
	assert(client->ipc_state.server == 0);

	/*
	 * Read the IPC send parameters structure from user space now.
	 * We have to do this before building the connection,
	 * because if we did it afterwards
	 * and took a page fault during the copyin
	 * after the connection had been established,
	 * we would be unable to get the registers
	 * into a valid send state.
	 */
	rc = copyin_local(EXC_SYSPARAM(client, ipc, params),
			  &client_ipc, sizeof(client_ipc) / sizeof(fluke_ipc_word_t));
	if (rc) return rc;

	/*
	 * Also read the minimum message words now.
	 * It's OK to decrease client_ipc.send_buf.words here,
	 * because they won't get written into the client's registers
	 * until the connect actually succeeds.
	 */
	if (client_ipc.send_buf.words < FLUKE_MIN_MSG_WORDS)
		EXC_RETURN_INSANITY(BUFFER_TOO_SMALL);

	rc = copyin_local((oskit_addr_t)client_ipc.send_buf.buf,
			  min_msg, FLUKE_MIN_MSG_WORDS);
	if (rc) return rc;
	client_ipc.send_buf.words -= FLUKE_MIN_MSG_WORDS;
	client_ipc.send_buf.buf += FLUKE_MIN_MSG_SIZE;

#if FLASK
	if (secure)
		client_sid = EXC_SYSPARAM(client, ipc, client_sid);
	else
		client_sid = client->sob.sid;

	/* Establish a connection to a new server thread. */
	rc = s_port_capture(client, SUBJECT_SUBJECT__CONNECT, 
			    client_sid, &server);
#else
	/* Establish a connection to a new server thread. */
	rc = s_port_capture(client, &server);
#endif
	if (rc)
		return rc;

	client->ipc_state.server = server;
	server->ipc_state.client = client;
	client->ipc_state.flags |= FLUKE_THREAD_CLIENT_SENDER;
	server->ipc_state.flags &= ~FLUKE_THREAD_SERVER_SENDER;

#if FLASK
        client->ipc_state.effective_client_sid = client_sid;
	server->ipc_state.client_effective_client_sid = client_sid;
	client->ipc_state.server_effective_server_sid =
		server->ipc_state.effective_server_sid;
#endif 

	/* Update the client's register state with the send parameters. */
	EXC_SYSPARAM(client, ipc, counts) =
		*((unsigned short*)&client_ipc.send_buf_count);
	IPC_SEND_BUF_BUF(client) = (oskit_addr_t)client_ipc.send_buf.buf;
	IPC_SEND_BUF_WORDS(client) = (oskit_addr_t)client_ipc.send_buf.words;
	IPC_SEND_BUF_TAB(client) = (oskit_addr_t)client_ipc.send_buf_tab;
	IPC_SEND_REF_TAB(client) = (oskit_addr_t)client_ipc.send_ref_tab;
	EXC_SET_IP(client, connected_ip);

	/* Update the server's state. */
	IPC_RECV_BUF_BUF(server) += FLUKE_MIN_MSG_SIZE;
	IPC_RECV_BUF_WORDS(server) -= FLUKE_MIN_MSG_WORDS;
	server->ipc_state.md.min_msg[0] = min_msg[0];
	server->ipc_state.md.min_msg[1] = min_msg[1];
	assert(FLUKE_MIN_MSG_WORDS == 2);
	EXC_SET_IP(server, fluke_ipc_server_receive__entry);

#if FLASK
        if (server->ipc_state.server_wants_client_sid) {
                server->ipc_state.md.min_msg[0] = client_sid;
                server->ipc_state.md.min_msg[1] = 0xdeadbeef;
                IPC_SEND_BUF_BUF(client) -= FLUKE_MIN_MSG_SIZE;
                IPC_SEND_BUF_WORDS(client) += FLUKE_MIN_MSG_WORDS;
                IPC_RECV_BUF_BUF(server) -= FLUKE_MIN_MSG_SIZE;
                IPC_RECV_BUF_WORDS(server) += FLUKE_MIN_MSG_WORDS;
        }
#endif

	/* Perform the send operation. */
	return ipc_reliable_transfer(client, server);
}

int ipc_client_1_ack_send(struct s_thread *const client, oskit_addr_t acked_ip)
{
	struct s_thread *server;
	fluke_ipc_word_t min_msg[FLUKE_MIN_MSG_WORDS];
	wait_val_t wval;
	int rc;

	rc = ipc_client_find_server(client, FLUKE_IPC_ACK_DISCONNECTED);
	if (rc) return rc;

	server = client->ipc_state.server;
	assert(server != 0);

	/* Read the minimum message words from user space.  */
	if (IPC_SEND_BUF_WORDS(client) < FLUKE_MIN_MSG_WORDS)
		EXC_RETURN_INSANITY(BUFFER_TOO_SMALL);
	rc = copyin_local(IPC_SEND_BUF_BUF(client), min_msg, FLUKE_MIN_MSG_WORDS);
	if (rc) return rc;

	/* Capture the server if it's waiting to send. */
	if ((wval = thread_wakeup(server, WAIT_IPC_SRV_SENDER)) == 0) {

		/*
		 * Server's not in the proper wait state.
		 * Check the current connection direction.
		 */
		if (client->ipc_state.flags & FLUKE_THREAD_CLIENT_SENDER)
			EXC_RETURN_INSANITY(NOT_RECEIVER);

		/*
		 * It's going the right direction,
		 * but the server just isn't ready - wait for it.
		 */
		return thread_wait(client, WAIT_IPC_CLI_ASEND,
				   server, WAIT_IPC_SRV_SENDER);
	}
	if (wval != WAIT_IPC_SRV_ORECV) {
		assert(wval == WAIT_IPC_SRV_SEND);

		/*
		 * The server is still trying to send data to us.
		 * Just handoff to it and let it drain its message.
		 */
		return thread_handoff(client, WAIT_IPC_CLI_ASEND, server);
	}

	/* Both the client and the server are ready to reverse, so do so. */
	client->ipc_state.flags |= FLUKE_THREAD_CLIENT_SENDER;
	server->ipc_state.flags &= ~FLUKE_THREAD_SERVER_SENDER;

	/* Update the client's register state. */
	IPC_SEND_BUF_BUF(client) += FLUKE_MIN_MSG_SIZE;
	IPC_SEND_BUF_WORDS(client) -= FLUKE_MIN_MSG_WORDS;
	EXC_SET_IP(client, acked_ip);

	/* Update the server's state and transfer the min_msg. */
	IPC_RECV_BUF_BUF(server) = (unsigned)
		server->ipc_state.md.params.recv_buf.buf + FLUKE_MIN_MSG_SIZE;
	IPC_RECV_BUF_WORDS(server) = 
		server->ipc_state.md.params.recv_buf.words
							- FLUKE_MIN_MSG_WORDS;
	IPC_RECV_BUF_COUNT(server) =
		server->ipc_state.md.params.recv_buf_count;
	IPC_RECV_BUF_TAB(server) = (unsigned)
		server->ipc_state.md.params.recv_buf_tab;
	IPC_RECV_REF_COUNT(server) =
		server->ipc_state.md.params.recv_ref_count;
	IPC_RECV_REF_TAB(server) = (unsigned)
		server->ipc_state.md.params.recv_ref_tab;
	server->ipc_state.md.min_msg[0] = min_msg[0];
	server->ipc_state.md.min_msg[1] = min_msg[1];
	assert(FLUKE_MIN_MSG_WORDS == 2);
	EXC_SET_IP(server, fluke_ipc_server_receive__entry);

	return ipc_reliable_transfer(client, server);
}

int ipc_client_2_over_receive(struct s_thread *const client, wait_val_t wval)
{
	struct s_thread * const server = client->ipc_state.server;
	int rc;

	assert(server != 0);

	/*
	 * Read the IPC send parameters structure from user space.
	 * XXX on connect_send_over_receive, they get copied in twice!
	 */
	rc = copyin_local(EXC_SYSPARAM(client, ipc, params),
			  &client->ipc_state.md.params,
			  sizeof(fluke_ipc_params_t) / sizeof(fluke_ipc_word_t));
	if (rc) {
		thread_ready(server);
		return rc;
	}

	/*
	 * Since the server still thinks it's receiving data,
	 * notify it that the message has ended.
	 */
	if (wval == WAIT_IPC_SRV_RECV)
		IPC_FINISH_RECEIVE(server, 0);
	else
		assert(wval == WAIT_IPC_SRV_ASEND);

	/* Hand control to the server. */
	return thread_handoff(client, WAIT_IPC_CLI_ORECV, server);
}

int ipc_server_1_ack_send(struct s_thread *const server, oskit_addr_t acked_ip)
{
	struct s_thread *client;
	fluke_ipc_word_t min_msg[FLUKE_MIN_MSG_WORDS];
	wait_val_t wval;
	int rc;

	rc = ipc_server_find_client(server, FLUKE_IPC_ACK_DISCONNECTED);
	if (rc) return rc;

	client = server->ipc_state.client;
	assert(client != 0);

	/* Read the minimum message words from user space.  */
	if (IPC_SEND_BUF_WORDS(server) < FLUKE_MIN_MSG_WORDS)
		EXC_RETURN_INSANITY(BUFFER_TOO_SMALL);
	rc = copyin_local(IPC_SEND_BUF_BUF(server), min_msg, FLUKE_MIN_MSG_WORDS);
	if (rc) return rc;

	/* Capture the client if it's waiting to send. */
	if ((wval = thread_wakeup(client, WAIT_IPC_CLI_SENDER)) == 0) {

		/*
		 * Client's not in the proper wait state.
		 * Check the current connection direction.
		 */
		if (server->ipc_state.flags & FLUKE_THREAD_SERVER_SENDER)
			EXC_RETURN_INSANITY(NOT_RECEIVER);

		/*
		 * It's going the right direction,
		 * but the client just isn't ready - wait for it.
		 */
		return thread_wait(server, WAIT_IPC_SRV_ASEND,
				   client, WAIT_IPC_CLI_SENDER);
	}
	if (wval != WAIT_IPC_CLI_ORECV) {
		assert(wval == WAIT_IPC_CLI_SEND);

		/*
		 * The client is still trying to send data to us.
		 * Just handoff to it and let it drain its message.
		 */
		return thread_handoff(server, WAIT_IPC_SRV_ASEND, client);
	}

	/* Both the server and the client are ready to reverse, so do so. */
	server->ipc_state.flags |= FLUKE_THREAD_SERVER_SENDER;
	client->ipc_state.flags &= ~FLUKE_THREAD_CLIENT_SENDER;

	/* Update the server's register state. */
	IPC_SEND_BUF_BUF(server) += FLUKE_MIN_MSG_SIZE;
	IPC_SEND_BUF_WORDS(server) -= FLUKE_MIN_MSG_WORDS;
	EXC_SET_IP(server, acked_ip);

	/* Update the client's state and transfer the min_msg. */
	IPC_RECV_BUF_BUF(client) = (unsigned)
		client->ipc_state.md.params.recv_buf.buf + FLUKE_MIN_MSG_SIZE;
	IPC_RECV_BUF_WORDS(client) = 
		client->ipc_state.md.params.recv_buf.words
							- FLUKE_MIN_MSG_WORDS;
	IPC_RECV_BUF_COUNT(client) =
		client->ipc_state.md.params.recv_buf_count;
	IPC_RECV_BUF_TAB(client) = (unsigned)
		client->ipc_state.md.params.recv_buf_tab;
	IPC_RECV_REF_COUNT(client) =
		client->ipc_state.md.params.recv_ref_count;
	IPC_RECV_REF_TAB(client) = (unsigned)
		client->ipc_state.md.params.recv_ref_tab;
	client->ipc_state.md.min_msg[0] = min_msg[0];
	client->ipc_state.md.min_msg[1] = min_msg[1];
	assert(FLUKE_MIN_MSG_WORDS == 2);
	EXC_SET_IP(client, fluke_ipc_client_receive__entry);

	return ipc_reliable_transfer(server, client);
}

/*
 * This function is called to do a wait_receive after a send.
 * The current client thread is still captured and connected.
 * On the x86, we need to read the receive parameters from user space now
 * since those registers were occupied by the send parameters.
 */
int ipc_server_2_wait_receive(struct s_thread *const server, wait_val_t wval,
			      oskit_addr_t wait_receive_ip)
{
	struct s_thread * const client = server->ipc_state.client;
	fluke_ipc_params_t server_ipc;
	int rc;

	/*
	 * Read the server receive parameters structure from user space now.
	 * If this fails, we'll just leave the server in a send_wait_receive,
	 * with all the send parameters set to zero.
	 */
	rc = copyin_local(EXC_SYSPARAM(server, ipc, params),
			  &server_ipc, sizeof(server_ipc) / sizeof(fluke_ipc_word_t));
	if (rc) {
		thread_ready(client);
		return rc;
	}
	EXC_SYSPARAM(server, ipc, counts) =
		*((unsigned short*)&server_ipc.recv_buf_count);
	IPC_RECV_BUF_BUF(server) = (oskit_addr_t)server_ipc.recv_buf.buf;
	IPC_RECV_BUF_WORDS(server) = (oskit_addr_t)server_ipc.recv_buf.words;
	IPC_RECV_BUF_TAB(server) = (oskit_addr_t)server_ipc.recv_buf_tab;
	IPC_RECV_REF_TAB(server) = (oskit_addr_t)server_ipc.recv_ref_tab;
	EXC_SET_IP(server, wait_receive_ip);

	/*
	 * Break the connection with the client thread and release it.
	 * It's much more efficient to do this manually here
	 * than to call ipc_server_disconnect() after releasing the client,
	 * because right now we know the client is passive (captured),
	 * whereas ipc_server_disconnect() would see it as "not waiting"
	 * and cause us to sleep ourselves, producing extra context switches.
	 */
	server->ipc_state.client = 0;
	client->ipc_state.server = 0;
#if FLASK
	FLASK_IPC_STATE_DISCONNECT(client,server);
#endif 
	IPC_FINISH_RECEIVE(client, FLUKE_IPC_RECV_DISCONNECTED);
	thread_ready(client);

	return s_pset_wait_receive(server);
}

int ipc_server_2_over_receive(struct s_thread *const server, wait_val_t wval)
{
	struct s_thread * const client = server->ipc_state.client;
	int rc;

	assert(client != 0);

	/*
	 * Read the IPC send parameters structure from user space.
	 * XXX on connect_send_over_receive, they get copied in twice!
	 */
	rc = copyin_local(EXC_SYSPARAM(server, ipc, params),
			  &server->ipc_state.md.params,
			  sizeof(fluke_ipc_params_t) / sizeof(oskit_reg_t));
	if (rc) {
		thread_ready(client);
		return rc;
	}

	/*
	 * Since the client still thinks it's receiving data,
	 * notify it that the message has ended.
	 */
	if (wval == WAIT_IPC_CLI_RECV)
		IPC_FINISH_RECEIVE(client, 0);
	else
		assert(wval == WAIT_IPC_CLI_ASEND);

	/* Hand control to the client. */
	return thread_handoff(server, WAIT_IPC_SRV_ORECV, client);
}

/*
 * Called by ipc_reply_wait_receive.
 * Similar to ipc_server_2_wait_receive execept that client has been
 * readied and disconnected.
 */
int ipc_server_3_wait_receive(struct s_thread *const server,
			      oskit_addr_t wait_receive_ip)
{
	fluke_ipc_params_t sparams;
	int rc;

	/*
	 * Copyin the server receive parameters structure.
	 */
	rc = copyin_local(EXC_SYSPARAM(server, ipc, params),
			  &sparams, sizeof(sparams)/sizeof(oskit_reg_t));
	if (rc)
		return rc;

	EXC_SYSPARAM(server, ipc, counts) =
		*((unsigned short*)&sparams.recv_buf_count);
	IPC_RECV_BUF_BUF(server) = (oskit_addr_t)sparams.recv_buf.buf;
	IPC_RECV_BUF_WORDS(server) =
		(oskit_addr_t)sparams.recv_buf.words;
	IPC_RECV_BUF_TAB(server) = (oskit_addr_t)sparams.recv_buf_tab;
	IPC_RECV_REF_TAB(server) = (oskit_addr_t)sparams.recv_ref_tab;
	EXC_SET_IP(server, wait_receive_ip);

	return s_pset_wait_receive(server);
}
