/*
 * Copyright (c) 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#include <fluke/exception.h>
#include <fluke/machine/syscall.h>

#include <oskit/machine/asm.h>

#include "s_mutex.h"
#include "s_thread.h"
#include "exception.h"

#include "smp.h"
#include "config_kpreempt.h"
#include <assert.h>

#include "debug.h"

int s_mutex_lockop(
	s_mutex_t		*mutex,
	s_thread_t		*cur_th)
{
	u_mutex_t *const um = mutex->u_mutex;
	int ostate;
	int rc;

	assert_s_mutex_locked(mutex);

	/*
	 * The mutex->sob.lock protects us against other kernel threads here,
	 * so we are only competing against other processors in user mode.
	 */

	asm("xchgl %0,%1 # %2=%0" /* xchg by itself is always atomic */
	    : "=r" (ostate), "=m" (um->state)
	    : "0" (U_MUTEX_CONTESTED));
	if (ostate == U_MUTEX_AVAILABLE) {
		/*
		 * The lock became available since the user-mode code tried.
		 * Now we have the lock and can return success.
		 * Since any user threads in other processors will trap
		 * into the kernel and take mutex->sob.lock as long as
		 * the state is not AVAILABLE, there is no race here
		 * and we can safely change it to HELD so there will
		 * be no extraneous need to enter the kernel to unlock it.
		 */
		um->state = U_MUTEX_HELD;
		s_mutex_unlock(mutex); /* caller expects us to unlock it */
		return 0;
	}
#if DEBUG
	if (ostate != U_MUTEX_HELD && ostate != U_MUTEX_CONTESTED) {
		s_mutex_unlock(mutex); /* caller expects us to unlock it */
		EXC_SET_IP(cur_th, fluke_mutex_lock__entry);
		exc_set_insanity(cur_th, FLUKE_INSANITY_INVALID_OBJECT);
		return KR_USER_EXCEPTION;
	}
#endif

	/*
	 * The user-mode entrypoint code executed some instructions
	 * before trapping into the kernel, so we need to roll back
	 * the PC to the official entrypoint location.  It is important
	 * that we do this before blocking so that our exc_state shows
	 * the "atomically correct" user thread state.
	 */
	EXC_SET_IP(cur_th, fluke_mutex_lock__entry);

	/*
	 * The lock state now properly indicates CONTESTED so
	 * a user-mode thread unlocking the mutex will trap into
	 * the kernel and do cond_signal.  Since he will have to
	 * take the mutex->sob.lock around cond_signal, we can't
	 * lose the wakeup.
	 */

	rc = cond_wait(&mutex->cond, &mutex->sob.lock, 0/*XXX*/);
	assert(rc != 0);

	return rc;
}

int s_mutex_unlockop(
	s_mutex_t		*mutex,
	s_thread_t		*cur_th)
{
	int ostate;
	int rc;

	assert_s_mutex_locked(mutex);

	asm("xchgl %0,%1 # %2=%0" /* xchg is atomic by itself */
	    : "=r" (ostate), "=m" (mutex->u_mutex->state)
	    : "0" (U_MUTEX_AVAILABLE));

#if ! DEBUG
	if (ostate > 0)		/* HELD */
		return 0;	/* call from sys_cond_wait */
	else
		rc = 0;		/* normal case */
#else
	switch (ostate) {
	case U_MUTEX_CONTESTED:
		/*
		 * This is the normal case.  If the lock was not contested,
		 * the user-mode entry point code doesn't trap into the kernel.
		 */
		rc = 0;
		break;

	case U_MUTEX_AVAILABLE:
		exc_set_insanity(cur_th, FLUKE_INSANITY_MUTEX_NOT_LOCKED);
		EXC_SET_IP(cur_th, fluke_mutex_unlock__entry);
		rc = KR_USER_EXCEPTION;
		break;

	case U_MUTEX_HELD:
		/*
		 * This should only happen in calls from sys_cond_wait.
		 * There is no need to wake anyone up.
		 */
		return 0;

	default:
		/*
		 * Other cases should not happen.  Assume the user clobbered
		 * the user-space flob and berate him.
		 */
		exc_set_insanity(cur_th, FLUKE_INSANITY_INVALID_OBJECT);
		EXC_SET_IP(cur_th, fluke_mutex_unlock__entry);
		rc = KR_USER_EXCEPTION;
		break;
	}
#endif

	/*
	 * Wake up other threads blocked on this mutex.
	 *
	 * We need to wake up all threads (broadcast not signal) because we
	 * don't know when the thread we wake up will run.  Since the lock
	 * is now available, it needs to go to the first locking thread
	 * that gets scheduled.  Now all the blocked threads will return to
	 * user mode and compete anew for the lock.
	 */
	cond_broadcast(&mutex->cond);

	return rc;
}

void
s_mutex_md_sanify_flob(struct s_ob *sob)
{
	s_mutex_t *const mutex = (s_mutex_t *)sob;
	int ostate;

	s_mutex_lock(mutex);

	/*
	 * Sanify the state word so it's either AVAILABLE or HELD.
	 */
	asm("xchgl %0,%1 # %2=%0" /* xchg is atomic by itself */
	    : "=r" (ostate), "=m" (mutex->u_mutex->state)
	    : "0" (U_MUTEX_HELD));

	/*
	 * If the old state was AVAILABLE, we declare it to have
	 * been truly unlocked, and now we have the lock, so we release it.
	 * Otherwise, some thread better believe it has the lock.
	 */
	if (ostate == U_MUTEX_AVAILABLE)
		mutex->u_mutex->state = U_MUTEX_AVAILABLE;

	/*
	 * Wake up everybody blocked on this mutex, in case anyone was
	 * blocked and the lockholder never woke them up (and since we've
	 * now clobbered the state to HELD if it was CONTESTED).  If the
	 * lock is still held by another thread now, the contesters will
	 * all restart in user mode and block again on this same cond.
	 */
	cond_broadcast(&mutex->cond);

	s_mutex_unlock(mutex);
}
