/*
 * Copyright (c) 1996, 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

#ifndef _FLUKE_KERNEL_X86_PMAP_H_
#define _FLUKE_KERNEL_X86_PMAP_H_

#include "mutex.h"

#include "config_opt_xfer.h"
#include "config_opt_threads.h"

/* One of these is embedded in each s_task structure,
   representing the machine-dependent page table structures.  */
struct task_pmap_state
{
	/* Lock protecting access to pmap data structures.  */
	struct mutex lock;

	/* Physical address of the page directory.
	   Equal to kernel_pmap.phys_pdir if there currently is none.  */
	oskit_addr_t phys_pdir;

#ifdef CONFIG_OPT_THREADS
	/* Page directory to contain the special GS page mappings */
	oskit_addr_t phys_gs_ptab;
#endif
};


#include_next "pmap.h"

/*
 * Symbols describing the layout of linear memory in a user task.
 *
 * The kernel is mapped at [LIN_KERNEL_MIN - LIN_KERNEL_MAX] but
 * not accessible from user mode.
 *
 * The user application is mapped at [LIN_USER_MIN - LIN_USER_MAX]
 * and again at [LIN_USER_MIN+LIN_USERX_OFF - LIN_USER_MAX+LIN_USERX_OFF].
 * The former maps all user memory (text and data) and is mapped by the
 * DS segment.  The latter maps all user memory which includes execute
 * permission and is mapped by the CS segment.  This allows us to enforce
 * read without execute in memory; i.e., if a region includes execute
 * permission it will be mapped in both CS and DS (and SS) allowing
 * "traditional" access.  If it doesn't include execute, it will be mapped
 * only by DS (and SS) meaning that it cannot be executed.
 *
 * Note that the converse (execute without read) cannot be done by mapping
 * a region in just the CS and not the DS because segment override can be
 * used to read from CS, e.g., "ldw cs:foo,eax".
 */

#define LIN_USER_MIN	((oskit_addr_t) 0x00000000)
#define LIN_USER_MAX	((oskit_addr_t) 0x3fffffff)

#define LIN_USERX_OFF	((oskit_addr_t) 0x40000000)

/* XXX beware: pmap.c code assumes that LIN_ENTRY_MIN is a multiple of 4MB */
#define LIN_ENTRY_MIN	(LIN_USER_MIN+FLUKE_ENTRY_MIN)
#define LIN_ENTRY_MAX	(LIN_USER_MIN+FLUKE_ENTRY_MAX)

#define LIN_KERNEL_MIN	((oskit_addr_t) 0x80000000)
#define LIN_KERNEL_MAX	((oskit_addr_t) 0xffffffff)

/*
 * Linear space region used for optimized cross-address-space copying.
 * Each processor needs 16MB of virtual address space -
 * 8MB for the copying window, and an unmapped red zone beyond that.
 *
 * XXX currently setup for only a single processor due to LIN_GSPACE below.
 */
#ifdef CONFIG_OPT_XFER
#define LIN_XFER_MIN	((oskit_addr_t) 0xfe000000)
#endif

/*
 * Linear space region used for the per-processor GS segments.
 */
#ifdef CONFIG_OPT_THREADS
#define LIN_GSPAGE_MIN	((oskit_addr_t) 0xff000000)

struct fluke_cpu_info;		/* per_cpu.h */
void gspage_init_ptab(void);	/* initialize for GSPAGE */
void gspage_setup_cpu(struct fluke_cpu_info *cpu); /* allocate the page */
#endif

/*
 * Function to map physical memory into the kernel virtual address space.
 */
oskit_addr_t pmap_map_phys_mem(oskit_addr_t start_pa, oskit_size_t size,
			       oskit_bool_t writable);

#endif /* _FLUKE_KERNEL_X86_PMAP_H_ */
