/*
 * Copyright (c) 1995-1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * This is the pseudo-kernel-module that loads the first user-level task
 * and thereafter acts as the parent for that task.
 */

#include <pthread.h>
#include <string.h>
#include <malloc.h>
#include <errno.h>
#include <oskit/page.h>
#include <oskit/exec/exec.h>
#include <oskit/x86/base_vm.h>
#include <oskit/x86/multiboot.h>
#include <oskit/x86/pc/base_multiboot.h>
#include <fluke/task.h>
#include <fluke/thread.h>
#include <fluke/region.h>
#include <fluke/mapping.h>
#include <fluke/port.h>
#include <fluke/cp/startup.h>
#include <fluke/debug.h>

#include "prog.h"
#include "kserver/kserver.h"
#include "debug.h"
#include "config_smp.h"
#include "config_user_sched.h"

#define CTRL(c) ((c) & 0x1f)

#define PROG_ARGBUF_SIZE	256
#define PROG_ARGS_MAX		64
#define PROG_NAME_MAX		64

static char prog_name[PROG_NAME_MAX] = PROG_NAME;
int askname;

extern void *kserver_smp(void *);
extern void idle_boot(void);

/*
 * This is the first function run in user mode.
 * We are running on startup_thread, but it is now functioning
 * as a totally normal user-mode thread flob.
 */
void
prog_start(void)
{
	char **argv, **envp;
	extern char **global_argv, **environ;

	if (global_argv && global_argv[1] && askname == 0) {
		argv = &global_argv[1];	/* XXX skip Image name */
		envp = environ;
	} else
		prog_boot_parse_cmdline(&argv, &envp);

#ifndef CONFIG_USER_SCHED
	/*
	 * Startup an idle thread for this processor.
	 * Then startup the (other) CPUs and have another idle thread
	 * created for each additional processor started up.
	 */
	idle_boot();

#ifdef CONFIG_SMP
	{
		extern int smp_start(void *); /* smp_boot.c */
		do_in_supervisor_mode(smp_start, 0);
	}
#endif CONFIG_SMP
#endif

	/*
	 * Start the kserver up and load the first program.
	 */
	kserver(argv, envp);

#ifdef CONFIG_USER_SCHED
	/*
	 * Launch into the kserver-provided user-mode scheduler.
	 * This is responsible for starting up other CPUs.
	 * It should never return.
	 */
	ks_sched_start();
#else
	/*
	 * This loader thread's job is done.  Let it die.
	 */
	pthread_exit(0);
#endif

	otsan();
}

static char argbuf[PROG_ARGBUF_SIZE];
static char *argvec[PROG_ARGS_MAX];
int getchar(void);
int putchar(int);

/*
 * Prompt the console for command line arguments for the initial task.
 * Copy the strings into argbuf, returns argvec, the 'argv-style' array
 * of pointers to strings (in argbuf).
 *
 * XXX this could be changed to parse environment strings as well.
 * XXX if the boot loader would pass through arguments we wouldn't need this.
 */
void
prog_boot_parse_cmdline(char ***argvpp, char ***envpp)
{
	char *cp;
	char **argv;
	int inarg, narg, strsize;

	/*
	 * Push program name on the stack
	 */
	strsize = strlen(prog_name) + 1;
	memcpy(argbuf, prog_name, strsize);
	narg = 1;
	cp = argbuf + strsize;
	argv = argvec;
	*argv++ = argbuf;

	if (askname == 0)
		goto done;

again:
	inarg = 0;

	/* XXX editing, etc. */
	printf("task0 arguments [^C restarts]: %s ", argbuf);
	for ( ; cp < (char *)argv; cp++) {
		*cp = getchar();
		putchar(*cp);
		switch (*cp) {
		case CTRL('c'):
			puts("^C\r\n");
			narg = strsize = 0;
			argv = argvec;
			cp = argbuf;
			*cp = 0;
			goto again;
		case -1:
		case '\r':
			putchar('\n');
		case '\n':
			if (inarg) {
				*cp = 0;
				strsize++;
			}
			break;
		case ' ':
		case '\t':
			if (inarg) {
				*cp = 0;
				strsize++;
				inarg = 0;
			} else
				--cp;
			continue;
		default:
			if (!inarg) {
				*argv++ = cp;
				inarg = 1;
			}
			strsize++;
			continue;
		}
		break;
	}

	*argv = 0;
	assert(argv < &argvec[PROG_ARGS_MAX]);

done:
	strncpy(prog_name, argbuf, PROG_NAME_MAX); /* maybe they changed it */

	*argvpp = argvec;
	*envpp = 0;	/* XXX we could parse these too */
}
