/*
 * Copyright (c) 1996 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Fluke Operating System.  The Fluke Operating System
 * is free software, also known as "open source;" you can redistribute it
 * and/or modify it under the terms of the GNU General Public License (GPL),
 * version 2, as published by the Free Software Foundation (FSF).  To explore
 * alternate licensing terms, contact the University of Utah at
 * csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The Fluke Operating System is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GPL for more
 * details.  You should have received a copy of the GPL along with the Fluke
 * Operating System; see the file COPYING.  If not, write to the FSF, 59 Temple
 * Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Default implementation of user-space memory transfer routines.
 * These routines always just look up the corresponding physical addresses
 * and use them with the phystokv macros to copy the data
 * using the kernel's direct-mapped physical mappings.
 * On most if not all architectures
 * there are much more efficient ways of doing these things,
 * so this implementation should be replaced if good performance is desired.
 */

#include <string.h>
#include <fluke/thread.h>
#include <oskit/machine/base_vm.h>

#include "exception.h"
#include "frag.h"
#include "s_task.h"
#include "s_mapping.h"
#include "xfer.h"
#include "s_thread.h"
#include "debug.h"

/*
 * Copy data from the user address space the specified thread is running in
 * into the kernel's address space.
 */
int copyin_remote(
	s_thread_t		*src_thread,
	oskit_addr_t		src_uva,
	void			*dest_kva,
	oskit_size_t		nwords)
{
	oskit_size_t nbytes = nwords * sizeof(fluke_ipc_word_t);

	assert (nbytes > 0);
	do
	{
		oskit_addr_t frag_start_va;
		oskit_size_t frag_ofs;
		oskit_size_t copy_size;
		frag_t *frag;
		int rc;

		/*
		 * Find the fragment covering this address in the thread's task.
		 * If none exists, we will record a page fault and return.
		 */
		rc = task_lookup(src_thread, 0, src_uva, FLUKE_PROT_READ, &frag);
		if (rc)
			return rc;

		assert(frag);
		assert_mutex_locked(&frag_mapping(frag)->ulock);

		frag_start_va = frag_mapping(frag)->start_va + frag_offset(frag);
		assert(src_uva >= frag_start_va);
		assert(src_uva < frag_start_va + frag_size(frag));

		/*
		 * Copy as much data as possible using this frag.
		 * Keep the frag locked during the copy
		 * to prevent it from going away or changing.
		 */
		frag_ofs = src_uva - frag_start_va;
		copy_size = frag_size(frag) - frag_ofs;
		if (copy_size > nbytes)
			copy_size = nbytes;
		memcpy(dest_kva,
		       (void *)phystokv(frag_start_pa(frag) + frag_ofs),
		       copy_size);

		s_mapping_u_unlock(frag_mapping(frag));

		src_uva += copy_size;
		dest_kva += copy_size;
		nbytes -= copy_size;
	}
	while (nbytes > 0);

	return 0;
}

#ifndef MD_COPYIN_LOCAL
/*
 * In this implementation,
 * we can trivially implement copyin_local based on copyin_remote.
 */
int copyin_local(
	oskit_addr_t		src_uva,
	void			*dest_kva,
	oskit_size_t		nwords)
{
	return copyin_remote(current_thread(), src_uva, dest_kva, nwords);
}
#endif /* MD_COPYIN_LOCAL */

/*
 * Copy data from the kernel's address space
 * into the user address space the specified thread is running in.
 */
int copyout_remote(
	s_thread_t		*dest_thread,
	void			*src_kva,
	oskit_addr_t		dest_uva,
	oskit_size_t		nwords)
{
	oskit_size_t nbytes = nwords * sizeof(fluke_ipc_word_t);

	assert (nbytes > 0);
	do
	{
		oskit_addr_t frag_start_va;
		oskit_size_t frag_ofs;
		oskit_size_t copy_size;
		frag_t *frag;
		int rc;

		/*
		 * Find the fragment covering this address in the thread's task.
		 * If none exists, we will record a page fault and return.
		 */
		rc = task_lookup(dest_thread, 0, dest_uva, FLUKE_PROT_WRITE,
				 &frag);
		if (rc) {
#if 0
			printf("%s:%d: dest_uva 0x%x, rc 0x%x\n",
			       __FILE__, __LINE__, dest_uva, rc);
#endif
			return rc;
		}
		assert(frag);
		assert_mutex_locked(&frag_mapping(frag)->ulock);

		frag_start_va = frag_mapping(frag)->start_va + frag_offset(frag);
		assert(dest_uva >= frag_start_va);
		assert(dest_uva < frag_start_va + frag_size(frag));

		/*
		 * Copy as much data as possible using this frag.
		 * Keep the frag locked during the copy
		 * to prevent it from going away or changing.
		 */
		frag_ofs = dest_uva - frag_start_va;
		copy_size = frag_size(frag) - frag_ofs;
		if (copy_size > nbytes)
			copy_size = nbytes;
		memcpy((void*)phystokv(frag_start_pa(frag) + frag_ofs),
		       src_kva, copy_size);

		s_mapping_u_unlock(frag_mapping(frag));

		src_kva += copy_size;
		dest_uva += copy_size;
		nbytes -= copy_size;
	}
	while (nbytes > 0);

	return 0;
}

#ifndef MD_COPYOUT_LOCAL
/*
 * In this implementation,
 * we can trivially implement copyout_local based on copyout_remote.
 */
int copyout_local(
	void			*src_kva,
	oskit_addr_t		dest_uva,
	oskit_size_t		nwords)
{
	return copyout_remote(current_thread(), src_kva, dest_uva, nwords);
}
#endif /* MD_COPYOUT_LOCAL */

#ifndef MD_COPYTO_REMOTE
int copyto_remote(
	s_thread_t		*dest_thread,
	oskit_addr_t		src_uva,
	oskit_addr_t		dest_uva,
	oskit_size_t		*inout_nwords,
	int			*out_remote_rc)
{
	s_thread_t *src_thread = current_thread();
	oskit_size_t nbytes = *inout_nwords * sizeof(fluke_ipc_word_t);
	oskit_size_t actual = 0;

	assert(*out_remote_rc == 0);

	assert (nbytes > 0);

	do
	{
		oskit_size_t copy_size;
		frag_t *src_frag, *dest_frag;
		oskit_addr_t src_frag_va, dest_frag_va;
		oskit_size_t src_frag_ofs, dest_frag_ofs;
		int rc;

		/*
		 * Find the appropriate fragment in each the thread's task.
		 * If none exists, record a page fault and return.
		 */
		rc = task_lookup_two_frags(
			src_thread,	dest_thread,
			src_uva,	dest_uva,
			FLUKE_PROT_READ,FLUKE_PROT_WRITE,
			&src_frag,	&dest_frag,
			out_remote_rc);
		if (rc || *out_remote_rc)
		{
			/*
			 * Return the number of words successfully copied,
			 * rounding 'actual' down to the nearest word.
			 */
			*inout_nwords = actual / sizeof(fluke_ipc_word_t);
			return rc;
		}

		assert(src_frag);
		assert_mutex_locked(&frag_mapping(src_frag)->ulock);

		src_frag_va = frag_mapping(src_frag)->start_va +
			frag_offset(src_frag);
		assert(src_uva >= src_frag_va);
		assert(src_uva < src_frag_va + frag_size(src_frag));

		assert(dest_frag);
		assert_mutex_locked(&frag_mapping(dest_frag)->ulock);

		dest_frag_va = frag_mapping(dest_frag)->start_va +
			frag_offset(dest_frag);
		assert(dest_uva >= dest_frag_va);
		assert(dest_uva < dest_frag_va + frag_size(dest_frag));

		/*
		 * Copy as much data as possible using these frags.
		 * Keep the frag locked during the copy
		 * to prevent it from going away or changing.
		 */
		src_frag_ofs = src_uva - src_frag_va;
		dest_frag_ofs = dest_uva - dest_frag_va;
		copy_size = frag_size(src_frag) - src_frag_ofs;
		if (copy_size > frag_size(dest_frag) - dest_frag_ofs)
			copy_size = frag_size(dest_frag) - dest_frag_ofs;
		if (copy_size > nbytes)
			copy_size = nbytes;
		memcpy((void*)phystokv(frag_start_pa(dest_frag)+dest_frag_ofs),
		       (void*)phystokv(frag_start_pa(src_frag)+src_frag_ofs),
		       copy_size);

		/*
		 * Unlock the frags' mapping(s).
		 * (They may be the same mapping.)
		 */
		if (frag_mapping(dest_frag) != frag_mapping(src_frag))
			s_mapping_u_unlock(frag_mapping(dest_frag));
		s_mapping_u_unlock(frag_mapping(src_frag));

		/*
		 * Proceed with the next chunk of data.
		 */
		src_uva += copy_size;
		dest_uva += copy_size;
		actual += copy_size;
		nbytes -= copy_size;
	}
	while (nbytes > 0);

	return 0;
}
#endif /* MD_COPYTO_REMOTE */

